unit umeteoblue;
{$mode delphi}

interface

uses
  Classes, SysUtils,
  {$IFDEF upgsql}
  upgsql,udbconpg,
  {$ELSE}
  sqldb, udbconpg,
  {$ENDIF}
  xmatdefs;

type
  TEnergySimpleEncabRec = class
    nid: integer; // nid en la db nuestra
    nidCentral: integer; // identificador de la central generadora.
    dtPrimeraMuestra: TDateTime; // Fecha de la primer muestra

    //**** info de mb.
    id: string;  // este es el ID de mb ... por ahora biene vacío y no lo guardo
    name: string;
    longitude: NReal;
    latitude: NReal;
    height:NREal; // (m.asl.)
    country: string;
    timezone: string;
    utc_timedifference: string;
    sunrise: string;
    sunset: string;
    constructor CreateFromStr( nidCentral: integer; r: string );
    constructor CreateFromDBRec( r: TSQLQuery );
    procedure InsertIntoDB(db: TDBPqCon);

    // solo para llamar la primera vez
class procedure CrearTabla(db: TDBPqCon);
  end;

  TEnergySimpleDataRec = class
    nidPadre: integer;
    dt: TDateTime; // UTC0
 //   local date;
 //   weekday;
 //   local time;
    temperature_grC: NReal;
    winddirection_degr: NReal;
    windspeed_km_p_h: NReal;
    windspeed_variation_km_p_h: NReal;
    precipitation_rawModel_mm: NReal;
    solar_radiation_Wp_m2: NReal;
    air_density_kg_p_m3: NReal;
    rel_hum_p100: NReal;
    extraterestrial_Solar_Radiation_W_p_m2: NReal;
    snowFraction: NReal;
    WindSpeed_rawModel_km_p_h: NReal;
    winddirection_80m_degr: NReal;
    WindSpeed_80m_km_p_h: NReal;
    temperature_rawModel_grC: NReal;
    mean_sea_level_Pressure_Pa: NReal;
    Pressure_Pa: NReal;

    constructor CreateFromStr( r: string );
    constructor CreateFromDBRec( r: TSQLQuery );
    procedure InsertIntoDB(db: TDBPqCon; nidPadre: integer);


    // velocidad a 80m en m/s
    function v_mps: NReal;

    class procedure CrearTabla(db: TDBPqCon);


  end;

  TEnergySimpleRec = class
    encab: TEnergySimpleEncabRec;
    data: TList;
    constructor CreateFrom(nidCentral: integer; archi: string);
    constructor CreateFromDB(db: TDBPqCon; nidProno: integer);
    procedure InsertIntoDB(db: TDBPqCon);
    procedure Free;
    function NHoras: integer;
    function dtPrimerMuestra: TDateTime;
    function rec( k: integer ): TEnergySimpleDataRec;
  end;


implementation

function nextpal( var r: string ): string;
var
  i: integer;
  res: string;

begin
  i:= pos(';', r);
  if i = 0 then
  begin
    result:= r;
    r:= '';
    exit;
  end;
  res:= copy( r, 1, i-1 );
  delete( r, 1, i );
  r:= trim( r );
  result:= trim( res );
end;

function nextFloat( var r: string): double;
begin
  result:= StrToFloat( nextpal( r ) );
end;

function nextInt( var r: string ): integer;
begin
  result:= StrToInt( nextPal( r ) );
end;



function TEnergySimpleRec.NHoras: integer;
begin
  result:= data.Count;
end;

function TEnergySimpleRec.dtPrimerMuestra: TDateTime;
begin
  result:= TEnergySimpleDataRec( data.items[0] ).dt;
end;


function TEnergySimpleRec.rec( k: integer ): TEnergySimpleDataRec;
begin
  result:= TEnergySimpleDataRec( data.items[k] );
end;



constructor TEnergySimpleRec.CreateFrom( nidCentral: integer; archi: string );
var
  f: textfile;
  r: string;
  buscando: boolean;
  aDataRec: TEnergySimpleDataRec;
  mb_metadata1, mb_metadata2: TStringList;
  kLin: integer;

{$IFDEF DBG_METEOBLUE_EnergySimple}
procedure readln( var f: textfile; var r: string );
begin
  inc( klin );
  system.readln( f, r );
  writeln( kLin ,': ', r );
end;
{$ENDIF}

begin
  inherited Create;
  mb_metadata1:= TStringList.Create;
  mb_metadata2:= TStringlist.Create;
  data:= TList.Create;

  // abrir archivo
  assignfile( f , archi );
  reset( f );
  klin:= 0;

  // leer/verificar matadata
  readln( f, r );
  if r <> '<mb_metadata>' then
    raise Exception.Create( 'Error - no inicio con <mb_metadata> ' );
  readln( f, r );
  if r<>'id;name;longitude;latitude;height (m.asl.);country;timezone;utc-timedifference;sunrise;sunset;' then
    raise Exception.Create( 'Cambió encab1' );

  while r <> '' do
    mb_metadata1.Add( nextpal( r ) );

  readln( f, r );
  if r <> 'local date;weekday;local time;temperature(C);winddirection(degr);windspeed(km/h);windspeed variation(km/h);precipitation rawModel (mm);solar radiation(W/m2);air density(kg/m3);rel hum (%);extraterestrial Solar Radiation(W/m2);snowFraction;WindSpeed rawModel (km/h);winddirection (80m) (degr);WindSpeed (80m) (km/h);temperature rawModel (C);mean sea level Pressure (Pa);Pressure (Pa);' then
    raise Exception.Create( 'Cambió encab1' );

   while r <> '' do
     mb_metadata2.Add( nextpal( r ) );

   readln( f, r );
   if r <> '</mb_metadata>' then
      raise Exception.Create( 'Se esperaba: </mb_metadata>' );

   readln( f, r );
   if r <> '<station>' then
      raise Exception.Create( 'Se esperaba: <station>' );

   readln( f, r );
   encab:= TEnergySimpleEncabRec.createFromStr(nidCentral, r );

  buscando:= true;
  while not eof( f ) and buscando do
  begin
    readln( f, r );
    if r ='</station>' then
      buscando:= false
    else
    begin
      aDataRec:= TEnergySimpleDataRec.createFromStr( r );
      data.add( aDataRec );
    end
  end;

  if buscando then
    raise Exception.Create( 'Fin de archivo sin encontrar: </station>' );

  closefile( f );

  mb_metadata1.Free;
  mb_metadata2.Free;

  encab.dtPrimeraMuestra:= TEnergySimpleDataRec( data[0] ).dt;

end;

constructor TEnergySimpleRec.CreateFromDB(db:  TDBPqCon; nidProno: integer );
var
  r: TSQLQuery;
  aRec: TEnergySimpleDataRec;
begin
  data:= TList.Create;
  r:= db.query( 'SELECT * FROM mb_energysimple WHERE nid = '+IntToStr( nidProno ));
  encab:= TEnergySimpleEncabRec.CreateFromDBRec( r );
  r.Free;
  r:= db.query( 'SELECT * FROM mb_energysimpledata WHERE nidpadre= '+IntToStr( encab.nid )+' ORDER BY nid ' );
  while not r.EOF do
  begin
    aRec:= TEnergySimpleDataRec.CreateFromDBRec( r );
    data.Add( aRec );
    r.next;
  end;
  r.Free;
end;

procedure TEnergySimpleRec.InsertIntoDB( db: TDBPqCon );
var
  aRec: TEnergySimpleDataRec;
  k: integer;
begin
  encab.InsertIntoDB( db );
  for k:= 0 to data.count-1 do
  begin
writeln( 'Procesando: ', k );
    aRec:= data[k];
    aRec.InsertIntoDB( db, encab.nid );
  end;
end;

procedure TEnergySimpleRec.Free;
var
  k: integer;

begin
  for k:= 0 to data.count-1 do
    TEnergySimpleDataRec( data[k] ).Free;
  data.Free;
  inherited Free;
end;


constructor TEnergySimpleEncabRec.createFromStr(nidCentral: integer; r: string );
var
  pal: string;
  i: integer;
  c: char;
begin
  inherited Create;
  nid:= -1; // no identificado en la db
  self.nidCentral:= nidCentral;

  id:= nextpal( r );
  name:= nextpal( r );
  pal:= nextPal( r );
  c:= pal[length( pal )];
  delete( pal, length( pal ) - 2, 3 );
  if c = 'W' then
    longitude:= -StrToFloat( pal )
  else
    longitude:= StrToFloat( pal );

  pal:= nextPal( r );
  c:= pal[length( pal )];
  delete( pal, length( pal ) - 2, 3 );
  if c = 'S' then
    latitude:= -StrToFloat( pal )
  else
    latitude:= StrToFloat( pal );

  height:= NextFloat( r );
  country:= NextPal( r );
  timezone:= NextPal( r );
  utc_timedifference:= NextPal( r );
  sunrise:= NextPal( r );
  sunset:= NextPal( r );
end;

constructor TEnergySimpleEncabRec.CreateFromDBRec( r: TSQLQuery );
begin
  nid:= r.FieldByName( 'nid' ).AsInteger;
//  id:= r.FieldByName( 'id' ).AsString;
  id:= '';
  dtPrimeraMuestra:= r.FieldByName( 'dtPrimeraMuestra' ).AsDateTime;
  nidCentral:= r.FieldByName( 'nidcentral' ).AsInteger;
  longitude:= r.FieldByName( 'longitude' ).AsFloat;
  latitude:= r.FieldByName( 'latitude' ).AsFloat;
  height:= r.FieldByName( 'height' ).AsFloat;
  country:= r.FieldByName( 'country' ).AsString;
  timezone:= r.FieldByName( 'timezone' ).AsString;
  utc_timedifference:= r.FieldByName( 'utc_timedifference' ).AsString;
  sunrise:= r.FieldByName( 'sunrise' ).AsString;
  sunset:= r.FieldByName( 'sunset' ).AsString;
end;


class procedure TEnergySimpleEncabRec.CrearTabla( db: TDBPqCon );
var
  res: string;
begin
  db.exec( 'DROP TABLE IF EXISTS mb_energysimple; ' );
  res:= 'CREATE TABLE mb_energysimple ( ';
  res:= res + ' nid serial primary key, ';
  res:= res + ' nidCentral integer, ';
  res:= res + ' dtPrimeraMuestra timestamp, ';
  res:= res + ' longitude double precision, ';
  res:= res + ' latitude double precision, ';
  res:= res + ' height double precision, ';
  res:= res + ' country varchar, ';
  res:= res + ' timezone varchar, ';
  res:= res + ' utc_timedifference varchar, ';
  res:= res + ' sunrise varchar, ';
  res:= res + ' sunset varchar ';
  res:= res + ' ); ';
  db.exec( res );
end;

procedure TEnergySimpleEncabRec.InsertIntoDB( db: TDBPqCon );
var
  sql: string;
begin
  nid:= StrToInt( db.func( 'SELECT nextval( ''mb_energysimple_nid_seq'' ); ', '-1' ) );
  if nid < 0 then raise Exception.Create( 'No logré nid para TEnergySimpleRec.InsertIntoDB ' );

  sql:= 'INSERT INTO mb_energysimple ( ';
  sql:= sql+ ' nid, nidCentral, dtPrimeraMuestra,  longitude, latitude, height, country, timezone, utc_timedifference,';
  sql:= sql+ ' sunrise, sunset ) VALUES ( ';
  sql:= sql+ IntToStr(nid)+', '+IntToStr( nidCentral );
  sql:= sql + ', '''+FormatDateTime( 'yyyy-mm-dd hh:ss', self.dtPrimeraMuestra )+''' ';
  sql:= sql +', '+FloatToStr( longitude )+', '+FloatToStr( latitude )+', ';
  sql:= sql+ FloatToStr( height )+', '''+country+''', '''+timezone+''', ''';
  sql:= sql+ utc_timedifference+ ''', '''+sunrise+''', '''+sunset+''' );';

  db.exec( sql );
end;



function getint( r: string; kIni, nChars: integer ): integer;
var
  pal: string;
begin
  pal:= copy( r , kIni, nChars );
  while (length( pal ) > 1) and (pal[1] = '0') do delete( pal, 1, 1 );
  result:= StrToInt( pal );
end;

constructor TEnergySimpleDataRec.createFromStr( r: string );
var
  local_date, weekday, local_time: string;
  dia, mes, anio: word;
  hora: word;
  pal: string;

begin
  self.nidPadre:= -1;
  local_date:= nextpal( r );
  weekday:= nextpal( r );
  local_time:= nextpal( r );
 // 123456789A
 // 12.10.2015
  dia:= getint( local_date, 1, 2 );
  mes:= getint( local_date, 4, 2 );
  anio:= getint( local_date, 7, 4 );
  hora:= getint( local_time, 1, 2 );

  dt:= EncodeDate( anio, mes, dia );
  dt:= dt + hora / 24.0;

//   local date;
//   weekday;
//   local time;
  temperature_grC:= nextFloat( r );
  winddirection_degr:= nextFloat( r );
  windspeed_km_p_h:= NextFloat( r );
  windspeed_variation_km_p_h:= NextFloat( r );
  precipitation_rawModel_mm:= NextFloat( r );
  solar_radiation_Wp_m2:= NextFloat( r );
  air_density_kg_p_m3:= NextFloat( r );
  rel_hum_p100:= NextFloat( r );
  extraterestrial_Solar_Radiation_W_p_m2:= NextFloat( r );
  snowFraction:= NextFloat( r );
  WindSpeed_rawModel_km_p_h:= NextFloat( r );
  winddirection_80m_degr:= NextFloat( r );
  WindSpeed_80m_km_p_h:= NExtFloat( r );
  temperature_rawModel_grC:= NExtFloat( r );
  mean_sea_level_Pressure_Pa:= NextFloat( r );
  Pressure_Pa:= NextFloat( r );
end;




constructor TEnergySimpleDataRec.CreateFromDBRec( r: TSQLQuery );
begin
  nidPadre:= r.FieldByName( 'nidPadre' ).AsInteger;
  dt:= r.FieldByName( 'dt' ).AsDateTime;
  temperature_grC:= r.FieldByName('temperature_grC').AsFloat;
  winddirection_degr:= r.FieldByName('winddirection_degr').AsFloat;
  windspeed_km_p_h:= r.FieldByName('windspeed_km_p_h').AsFloat;
  windspeed_variation_km_p_h:= r.FieldByName('windspeed_variation_km_p_h').AsFloat;
  precipitation_rawModel_mm:= r.FieldByName('precipitation_rawModel_mm').AsFloat;
  solar_radiation_Wp_m2:= r.FieldByName('solar_radiation_Wp_m2').AsFloat;
  air_density_kg_p_m3:= r.FieldByName('air_density_kg_p_m3').AsFloat;
  rel_hum_p100:= r.FieldByName('rel_hum_p100').AsFloat;
  extraterestrial_Solar_Radiation_W_p_m2:= r.FieldByName('extraterestrial_Solar_Radiation_W_p_m2').AsFloat;
  snowFraction:= r.FieldByName('snowFraction').AsFloat;
  WindSpeed_rawModel_km_p_h:= r.FieldByName('WindSpeed_rawModel_km_p_h').AsFloat;
  winddirection_80m_degr:= r.FieldByName('winddirection_80m_degr').AsFloat;
  WindSpeed_80m_km_p_h:= r.FieldByName('WindSpeed_80m_km_p_h').AsFloat;
  temperature_rawModel_grC:= r.FieldByName('temperature_rawModel_grC').AsFloat;
  mean_sea_level_Pressure_Pa:= r.FieldByName('mean_sea_level_Pressure_Pa').AsFloat;
  Pressure_Pa:= r.FieldByName('Pressure_Pa').AsFloat;
end;

procedure TEnergySimpleDataRec.InsertIntoDB(db: TDBPqCon; nidPadre: integer );
var
  sql: string;

procedure apf( val: NReal );
begin
  sql:= sql +', '''+FloatToStr( val )+'''';
end;

begin
  self.nidPadre:= nidPadre;
  sql:= 'INSERT INTO mb_EnergySimpleData ( ';
  sql:= sql + 'nidPadre, dt, temperature_grC, winddirection_degr, ';
  sql:= sql + 'windspeed_km_p_h, windspeed_variation_km_p_h, ';
  sql:= sql + 'precipitation_rawModel_mm, solar_radiation_Wp_m2, ';
  sql:= sql + 'air_density_kg_p_m3, rel_hum_p100, ';
  sql:= sql + 'extraterestrial_Solar_Radiation_W_p_m2, ';
  sql:= sql + 'snowFraction, WindSpeed_rawModel_km_p_h, ';
  sql:= sql + 'winddirection_80m_degr, WindSpeed_80m_km_p_h, ';
  sql:= sql + 'temperature_rawModel_grC, mean_sea_level_Pressure_Pa, ';
  sql:= sql + 'Pressure_Pa ) VALUES ( ';
  sql:= sql + IntToStr(nidPadre )+', '''+ FormatDateTime( 'yyyy-mm-dd hh:nn:ss', dt)+''' ';
  apf( temperature_grC );
  apf( winddirection_degr );
  apf( windspeed_km_p_h );
  apf( windspeed_variation_km_p_h );
  apf( precipitation_rawModel_mm );
  apf( solar_radiation_Wp_m2 );
  apf( air_density_kg_p_m3 );
  apf( rel_hum_p100 );
  apf( extraterestrial_Solar_Radiation_W_p_m2 );
  apf( snowFraction );
  apf( WindSpeed_rawModel_km_p_h );
  apf( winddirection_80m_degr );
  apf( WindSpeed_80m_km_p_h );
  apf( temperature_rawModel_grC );
  apf( mean_sea_level_Pressure_Pa );
  apf( Pressure_Pa );
  sql:= sql + ' ); ';
  db.exec( sql );
end;

class procedure TEnergySimpleDataRec.CrearTabla( db: TDBPqCon );
var
   sql: string;
procedure apf( s: string );
begin
  sql:= sql +', '+s+' double precision';
end;

begin
  db.exec( 'DROP TABLE IF EXISTS mb_energysimpledata; ' );
  sql:= 'CREATE TABLE mb_EnergySimpleData ( ';
  sql:= sql +'nid serial PRIMARY KEY, nidPadre integer, dt timestamp';
  apf( ' temperature_grC ' );
  apf( ' winddirection_degr ' );
  apf( ' windspeed_km_p_h ' );
  apf( ' windspeed_variation_km_p_h ' );
  apf( ' precipitation_rawModel_mm ' );
  apf( ' solar_radiation_Wp_m2 ' );
  apf( ' air_density_kg_p_m3 ' );
  apf( ' rel_hum_p100 ' );
  apf( ' extraterestrial_Solar_Radiation_W_p_m2 ' );
  apf( ' snowFraction ' );
  apf( ' WindSpeed_rawModel_km_p_h ' );
  apf( ' winddirection_80m_degr ' );
  apf( ' WindSpeed_80m_km_p_h ' );
  apf( ' temperature_rawModel_grC ' );
  apf( ' mean_sea_level_Pressure_Pa ' );
  apf( ' Pressure_Pa ' );
  sql:= sql + '); ';
  db.exec( sql );
end;




function TEnergySimpleDataRec.v_mps: NReal;
begin
 result:= WindSpeed_80m_km_p_h * 0.2777777778; // 0.2777777778 = 1000/3600;
//  result:= WindSpeed_rawModel_km_p_h * 0.2777777778;
//  result:= windspeed_km_p_h * 0.2777777778;

end;

end.

