unit uModeladoParques;

interface

uses
  Classes, SysUtils, Math, matreal, xmatdefs,
  ucomplex,
  uro_aire,
  uauxiliares,
  udbconpg;

const
  VERSION_ARCHI_PARQUE = 2;

  vCorte = 12;
  pvCorte = vCorte * Vcorte * vcorte * 0.5 * 1.2 / 1.0E6;

  vCorte_0 = 2;
  pvCorte_0 = vCorte_0 * Vcorte_0 * vcorte_0 * 0.5 * 1.2 / 1.0E6;

  vCorte_Sup = 20;
  pvCorte_Sup = vCorte_Sup * Vcorte_Sup * vcorte_Sup * 0.5 * 1.2 / 1.0E6;

  show_padding = 30; // banda en metros al borde visible


type

  TCurvaAerogen = class
    nid: integer;
    modelo: string;
    Diametro: NReal;
    NPuntos: integer;
    curva_v, curva_P: TVectR;

    curva_pv3: TVectR;
    pv3min, pv3max: NReal;
    AreaRotor: NReal;
    constructor CreateLoadFromArchi(modelo, archi_curva: string);
    constructor CreateLoadFromFile(var f: textfile);
    procedure StoreInFile(var f: textfile);
    function P_MW(pv3: NReal): NReal;
    procedure Free;
    class procedure CrearTablaDB(db: TDBPQCon);
    constructor CreateLoadFromDB(r: TSQLQuery);
    function InsertIntoDB(db: TDBPQCon): integer;
    procedure UpdateDB(db: TDBPQCon);

  private
    // lo deven llamar los constructores para calcular le vector pv3
    procedure Prepararse;
  end;


  TPosicion = class
    // Coordenadas en sistema  Universal Transversa de Mercator (UTM)
    // Uuruguay: elipsoide GRS80, Huso 21 y 22, Zonas H y J.
    UTM_Uso: integer;
    UTM_Zona: char;
    UTM_xE, UTM_yN: NReal;

    // posición en la grilla sin rotar  ( x: Oeste-> Este, y :  Sur->Norte )
    z_base: Complex;

    // Calculados con cada nueva dirección de viento al generar la grilla rotada.
    z_rotado: Complex; // posición en la grilla rotada

    kFila: integer; // fila en la grilla de cálculo rotada
    jCol1, jCol2: integer; // primer y última columna en la grilla rotada

    constructor Create_Cartesianas(x, y: NReal);

    procedure Set_UTM(uso: integer; Zona: char; xE, yN: Nreal);
    constructor Create_LoadFromFile(var f: textfile);
    procedure StoreInFile(var f: textfile);
  end;


  TMolino = class
    nid: integer;
    nid_curvaAero: integer;
    nid_Parque: integer;

    curva: TCurvaAerogen;
    ro: NReal;

    posicion: TPosicion;

    // Estas variables son calculadas por el Parque.
    pv_prom: NReal; // Potencia de viento de acuerdo a la dirección actual
    PE_MW: NReal; // generando de acuerdo a la dirección actual

    constructor Create_Init(x, y: NReal; modeloAero, archi_curvaAero: string;
      nSectoresRo: integer);

    constructor Create_LoadFromFile(var f: textfile);

    procedure StoreInFile(var f: textfile);

    function P_MW(ang_gr, v3Libre: NReal): NReal;
    procedure Free;

    // Este procedimiento calcula la potencia electrica del molino usando
    // la curva del mismo. Para la potencia del viento incidente toma el promedio
    // de las potencia de viento de los puntos de la grilla rotada asociados al molino.
    // el procedimiento afecta luego del calculo los valores de PV en la grilla rotada
    // de forma que reflejen la potencia eléctrica extraída.x
    procedure CalcularAFectacionEnGrillaRotada(kfila_: integer;
      fila: TVectR; ang, fpv3pe, relRed: NReal);
    class procedure CrearTablaDB(db: TDBPQCon);
    constructor CreateLoadFromDB(r: TSQLQuery);
    function InsertIntoDB: integer;
    procedure UpdateDB;

  end;


  TDAOFMolino = array of TMolino;

  TParque = class
    Nombre: string;
    CantMolinos: integer;
    NSectoresDir: integer;
    Molinos: TDAOFMolino;

    // auxiliar para mapeo a pixels en CANVAS
    xp1, yp1, xp2, yp2: integer;
    dxa, dya: NReal; // ancho y alto en m
    dxp, dyp: integer; // ancho y alto en px
    show_xmin, show_xmax, show_ymin, show_ymax: NReal; // caja en m

    Meteo_UTM_uso: integer;
    Meteo_UTM_Zona: char;
    Meteo_UTM_xe, Meteo_UTM_yn: NReal;
    Meteo_ro: NReal;

    // Rectángulo en coordenadas cartesianas con el anemo en (0,0) que incluye
    // todos los aeros. Si se optimiza el modelo variando posicion de aeroes se
    // mueven dentro de este rectángulo.
    Aero_xmin, Aero_xmax, Aero_ymin, Aero_ymax: NReal;

    // Extensiòn del rectángulo para incluir el Anemómetro.
    Caja_xmin, Caja_xmax, Caja_ymin, Caja_ymax: NReal;
    Caja_dx, Caja_dy: NReal; // anchos de la caja

    z_BaseToRot: complex; // exp(j (alfa+90) )

    grillaRotada_dx, grillaRotada_dy, grillaRotada_MargenDx: NReal;
    grillaRotada: TMatR;

    alfa_terreno: NReal;  // Caculado para grilla de 1m
    alfa: NReal;
    // factor perturbante horizontal 0.5 *( dx/dy) * (1-exp( - alfa_terreno*dx )) de la grilla rotada
    beta: NReal; // 1 - 2 * alfa

    gamma_atmosfera: NReal; // factor de mezcla vertical
    gamma_sup: NReal;  // (dx / 300)* (1- exp( - gamma_atmosfera *dx ))
    gamma_inf: NReal;  // 1- gamma_sup

    fpv3pe: NReal; // factor de perdida delta_pv3 = fpv3pe * PE

    ang: NReal; // ángulo último cálculo
    pv_libre: NReal;  // pv_libre Ultimo cálculo

    k_Perdidas: NReal; // valor calculado a partir del Factor y el límite

    // Potencia Nominal.
    PotenciaNominal: NReal;

    // factor de pérdidas eléctricas. Especificado a la potencia nominal.
    FactorPerdidas: NReal;

    // Limitación de la potenica inyectada de acuerdo al contrato
    // de conección. Se supone que el parque controla siempre
    // la potencia inyectada para no superar este valor.
    LimiteDePotencia: NReal;



    // Matriz con los factores cada fila identifica un sector de 5° de ángulo
    // de ataque delviento.   De 0 a 355°. (72 filas)
    // Las columnas corresponden a la v elocidad del viento libre.
    // De 1 a 29mps cada 1mps.  (29 columnas)
    // Los coeficientes son V3_Anemo


    constructor Create_Init(aNombre: string; aCantMolinos, aNSectoresDir: integer;
      modeloAero, archi_curvaAero: string;
    //afectacion_v3: TAnemometro;
      alfa_terreno, gamma_atmosfera, fpv3pe: NReal;
      Aero_xmin_, Aero_xmax_, Aero_ymin_, Aero_ymax_: NReal;
      npuntos_x, npuntos_y: integer;
    // A considerar como grilla para distribuciòn inicial de los molinos
      grillaRotada_dx_, grillaRotada_dy_, grillaRotada_MargenDx_: NReal;
    // resoluciòn de la grilla rotada
      FactorPerdidas, PotenciaNominal, LimiteDePotencia: NReal);


    constructor Create_LoadFromFile(var f: textfile);
    procedure StoreInFile(var f: textfile);
    procedure SetCurvaAero(kAero: integer; aCurva: TCurvaAerogen);

    procedure Free;

    {function get_params_ptrs: TDAOfNReal;}

    procedure SetMeteo_UTM(uso: integer; Zona: char; xE, yN: Nreal);
    procedure SetAero_UTM(kMolino: integer; uso: integer; Zona: char; xE, yN: Nreal);
    procedure RecalcularCoordenadas;

    // Fija las coordenada despleglables y el frame en pixeles
    procedure SetPixFrame(xp1, yp1, xp2, yp2: integer);

    procedure ZToPix(var xp, yp: integer; z: Complex);
    procedure PixToZ(var z: Complex; xp, yp: integer);

    // busca el molino que esté más proximo al punto xp, yp y lo devuelve
    // si el mismo está a menos de un diametro del punto xp, yp
    function MolinoAt_xpyp(xp, yp: integer): TMolino;

    function GridValAtZ(z: complex): NReal;
    function GridValAtPix(xp, yp: integer): NReal;

    procedure Calc_PotMW_init_(ang: NReal);

    function Calc_PotMW(pv3: NReal): NReal;

    function Calc_PotMW_anemo(pv3_anemo: NReal): NReal;

    procedure Calcular_pv_fila(filaACalcular, filaAnterior: TVectR);

    function ParametrosPtr: TDAOfNRealPtr;



    procedure Calc_Auxiliares;


  private
    procedure AerosAfectarFila(kFila: integer; ang, relRed: NReal);


  end;




implementation



procedure TCurvaAerogen.Free;
begin
  curva_P.Free;
  curva_v.Free;
  curva_pv3.Free;
  inherited Free;
end;

procedure TCurvaAerogen.Prepararse;
var
  i: integer;
begin
  AreaRotor := pi * sqr(Diametro / 2);
  curva_pv3 := TVectR.Create_Init(NPuntos);
  //bucle
  for i := 1 to NPuntos do
    curva_pv3.pon_e(i, pv3Ofv(curva_v.e(i), ro_BASE));
  curva_pv3.MinMax(i, i, pv3min, pv3max);
end;

constructor TCurvaAerogen.CreateLoadFromArchi(modelo, archi_curva: string);
var
  f: textfile;
  k, i, j: integer;
  p, v: NReal;

begin
  inherited Create;
  nid := -1; // indicamos que todavía no tiene nid

  self.modelo := modelo;

  assignfile(f, archi_curva);
  reset(f);
  readln(f, Diametro);
  readln(f, NPuntos);

  curva_P := TVectR.Create_Init(NPuntos);
  curva_v := TVectR.Create_Init(NPuntos);
  //bucle
  for i := 1 to NPuntos do
  begin
    readln(f, v, p);
    curva_v.pon_e(i, v);
    curva_P.pon_e(i, p / 1000.0);
  end;
  closefile(f);
  Prepararse;
end;



constructor TCurvaAerogen.CreateLoadFromFile(var f: textfile);
var
  i: integer;
begin
  inherited Create;
  //  readln(f, nid );
  nid := -1;
  readln(f, modelo);
  readln(f, diametro);
  curva_v := TVectR.CreateLoadFromFile(f);
  curva_P := TVectR.CreateLoadFromFile(f);
  Prepararse;
end;


constructor TCurvaAerogen.CreateLoadFromDB(r: TSQLQuery);
begin
  nid := r.FieldByName('nid').AsInteger;
  modelo := r.FieldByName('modelo').AsString;
  diametro := r.FieldByName('diametro').AsFloat;
  curva_v := TVectR.Create_unserialize(r.FieldByName('curva_v').AsString);
  curva_P := TVectR.Create_unserialize(r.FieldByName('curva_P').AsString);
  prepararse;
end;

function TCurvaAerogen.InsertIntoDB( db: TDBPQCon ): integer;
var
  nid: integer;
  s: string;
begin
  nid := db.nextval('curvas_aerogen');
  s := 'INSERT INTO curvas_aerogen ( nid, modelo, diametro, curva_v, curva_P ) VALUES (';
  s := s + IntToStr(nid) + ', ''' + modelo + ''', ''' + FloatToStr(
    diametro) + ''', ''' + curva_v.serialize + ''', ''' + curva_P.serialize + ''' );';
  db.exec(s);
  Result := nid;
end;

procedure TCurvaAerogen.UpdateDB( db: TDBPQCon );
var
  nid: integer;
  s: string;
begin
  s := 'UPDATE curvas_aerogen SET modelo:= ''';
  s := s + modelo + ''', diametro= ''' + FloatToStr(diametro) + ''', curva_v=''' +
    curva_v.serialize + ''', curva_P=''' + curva_P.serialize + ''' WHERE nid =' +
    IntToStr(nid) + ';';
  db.exec(s);
end;

class procedure TCurvaAerogen.CrearTablaDB( db: TDBPQCon );
var
  s: string;
begin
  db.exec('CREATE TABLE curvas_aergen ( nid serial PRIMARY KEY, '
    + 'modelo VARCHAR(100), diametro double precision, curva_v text, curva_P text ); ');
  db.exec(s);
end;

procedure TCurvaAerogen.StoreInFile(var f: textfile);
begin
  writeln(f, nid);
  writeln(f, modelo);
  writeln(f, diametro);
  curva_v.StoreInFile(f);
  curva_P.StoreInFile(f);
end;


function TCurvaAerogen.P_MW(pv3: NReal): NReal;
var
  ind: NReal;
begin
  if (pv3 < pv3min) or (pv3 > pv3max) then
  begin
    Result := 0;
    exit;
  end;
  ind := curva_pv3.inv_interpol(pv3);
  Result := curva_P.interpol(ind);
end;




procedure TPosicion.Set_UTM(uso: integer; Zona: char; xE, yN: Nreal);
begin
  self.UTM_Uso := uso;
  self.UTM_Zona := Zona;
  self.UTM_xE := xE;
  self.UTM_yN := yN;
end;

constructor TPosicion.Create_Cartesianas(x, y: NReal);
begin
  inherited Create;
  self.z_base := cinit(x, y);
end;

constructor TPosicion.Create_LoadFromFile(var f: textfile);
var
  x, y: NReal;
  r: string;
begin
  readln(f, x, y);
  Create_Cartesianas(x, y);
  readln(f, r);
  UTM_Uso := nextint(r);
  UTM_Zona := nextpal(r)[1];
  UTM_xE := nextfloat(r);
  UTM_yN := nextfloat(r);
end;

procedure TPosicion.StoreInFile(var f: textfile);
begin
  writeln(f, z_base.re, #9, z_base.im);
  writeln(f, UTM_Uso, #9, UTM_Zona, #9, UTM_xE, #9, UTM_yN);
end;

(*
constructor TAnemometro_MatR.Create_Init(NSectoresDir_sombra, NVLibre3_sombra: integer;
  archi_Sombras: string = '');
var
  k, i, j: integer;
  archi, linea: TStringList;
  vLibre, vAnemo: NReal;
  factor: NReal;

begin
  inherited Create;
  VRotor3_porSombra := TMatR.Create_init(nSectoresDir_sombra, nVLibre3_sombra);
  // filas angulos, columnas VLibre3
  puntos_v3 := TVectR.Create_init(NVLibre3_sombra);


  for k := 1 to puntos_v3.n do
    puntos_v3.pon_e(k, power(k, 3));

  if archi_Sombras = '' then
  begin
    for k := 1 to VRotor3_porSombra.nf do
      for j := 1 to VRotor3_porSombra.nc do
        VRotor3_porSombra.pon_e(k, j, 1);
  end
  else
  begin

    archi := TStringList.Create;
    archi.LoadFromFile(archi_Sombras);
    linea := TStringList.Create;
    linea.Delimiter := #9;

    if NSectoresDir_sombra <> 72 then
      raise Exception.Create('OJO, ... archi_sombras tiene que tener 72 direccione');

    if NVLibre3_sombra <> 30 then
      raise Exception.Create('OJO, ... N puntos VLibre <> 30');

    //bucle
    for i := 1 to 72 do
    begin
      linea.CommaText := archi[i];
      for j := 0 to 29 do
      begin
        vLibre := (j + 1);
        factor := strtofloat(trim(linea[j]));
        vAnemo := vLibre / factor;
        VRotor3_porSombra.pon_e(i, j + 1, power(vAnemo, 3));
      end;
    end;
    archi.Free;
    linea.Free;
  end;

end;


procedure TAnemometro_MatR.Free;
begin
  VRotor3_porSombra.Free;
  puntos_v3.Free;
  inherited Free;
end;



constructor TAnemometro_MatR.Create_LoadFromFile(var f: textfile);
var
  r: string;
begin
  puntos_v3 := TVectR.CreateLoadFromFile(f);
  VRotor3_porSombra := TMatR.CreateLoadFromFile(f);
end;

procedure TAnemometro_MatR.StoreInFile(var f: textfile);

begin
  puntos_v3.StoreInFile(f);
  VRotor3_porSombra.StoreInFile(f);
end;

function TAnemometro_MatR.v3_LibreToLocal(ang_gr, v3Libre: NReal): NReal;
var
  fila, col: integer;
  Dir, discDir, dir1, V1, fracciondir, fraccionVel, FCd1V1, FCd1V2,
  FCd2V1, FCd2V2, FCV1, FCV2, FC: NReal;
  es0, es355, esValta, esVbaja: boolean;
  rDir: NReal;
  kDir, jDir: integer;
  peso_jDir: NReal;

  kV3, jV3: integer;
  peso_jV3: NReal;

begin
  rDir := (ang_gr / 5);
  peso_jDir := frac(rDir);
  kDir := (trunc(rDir) mod VRotor3_porSombra.nf) + 1;
  jDir := (kDir mod VRotor3_porSombra.nf) + 1;

  peso_jV3 := puntos_v3.inv_interpol(v3Libre);
  kV3 := trunc(peso_jV3);
  peso_jV3 := frac(peso_jV3);
  jV3 := kV3 + 1;
  if jV3 > VRotor3_porSombra.nc then
    jV3 := VRotor3_porSombra.nc;

  Result := (1 - peso_jDir) * ((1 - peso_jV3) * VRotor3_porSombra.e(
    kDir, kV3) + peso_jV3 * VRotor3_porSombra.e(kDir, jV3)) +
    peso_jDir * ((1 - peso_jV3) * VRotor3_porSombra.e(jDir, kV3) +
    peso_jV3 * VRotor3_porSombra.e(jDir, jV3));

end;


function TAnemometro_MatR.v3_LocalToLibre(ang_gr, v3Local: NReal): NReal;
var
  rDir: NReal;
  kDir, jDir: integer;
  peso_jDir: NReal;
  kV3, jV3: integer;
  peso_jV3: NReal;
  aux: TVectR;
  j: integer;
begin
  rDir := (ang_gr / 5);
  peso_jDir := frac(rDir);
  kDir := (trunc(rDir) mod VRotor3_porSombra.nf) + 1;
  jDir := (kDir mod VRotor3_porSombra.nf) + 1;

  aux := TVectR.Create_init(puntos_v3.n);
  for j := 1 to aux.n do
    aux.pon_e(j, VRotor3_porSombra.e(kDir, j) * (1 - peso_jDir) +
      VRotor3_porSombra.e(jDir, j) * peso_jDir);

  peso_jV3 := aux.inv_interpol(v3Local);
  kV3 := trunc(peso_jV3);
  peso_jV3 := frac(peso_jV3);
  jV3 := kV3 + 1;
  if jV3 > VRotor3_porSombra.nc then
    jV3 := VRotor3_porSombra.nc;

  Result := ((1 - peso_jV3) * aux.e(kV3) + peso_jV3 * aux.e(jV3));
  aux.Free;
end;




// Métodos de TAnemometro_Func = class( TAnemometro )


constructor TAnemometro_Func.Create_Init(NSectoresDir_sombra: integer);
begin
  inherited Create;
  // supongo 3 paràmetros por direcciòn
  parametros_dir := TMatR.Create_init(NSectoresDir_sombra, 3);

  //** inicializar para mapa sin afectaciòn.

end;

constructor TAnemometro_Func.Create_LoadFromFile(var f: textfile);
begin
  inherited Create;
  parametros_dir := TMatR.CreateFromXLT(f);
end;

procedure TAnemometro_Func.StoreInFile(var f: textfile);
begin
  parametros_dir.WriteXLT(f);
end;

procedure TAnemometro_Func.Free;
begin
  inherited Free;
  parametros_dir.Free;
end;

function TAnemometro_Func.v3_LibreToLocal(ang_gr, v3Libre: NReal): NReal;
begin
  Result := 0; /// poner en funciòn de los paràmetros
end;

function TAnemometro_Func.v3_LocalToLibre(ang_gr, v3Local: NReal): NReal;
begin
  Result := 0; /// poner en funciòn de los paràmetros
end;

*)

constructor TMolino.Create_Init(x, y: NReal; modeloAero, archi_curvaAero: string;
  nSectoresRo: integer);
  //  afectacion_v3: TAnemometro);
begin
  inherited Create;
  nid := -1;
  nid_curvaAero := -1;
  nid_Parque := -1;
  posicion := TPosicion.Create_cartesianas(x, y);
  //  self.afectacion_v3 := afectacion_v3;
  self.curva := TCurvaAerogen.CreateLoadFromArchi(modeloAero, archi_curvaAero);
  ro := 1.0;
end;


constructor TMolino.Create_LoadFromFile(var f: textfile);
var
  tipoA: string;

begin
  inherited Create;
  posicion := TPosicion.Create_LoadFromFile(f);
  self.curva := TcurvaAerogen.CreateLoadFromFile(f);
  readln(f, ro);
end;


procedure TMolino.StoreInFile(var f: textfile);
begin
  posicion.StoreInFile(f);
  self.curva.StoreInFile(f);
  writeln(f, ro);
end;


class procedure TMolino.CrearTablaDB( db: TDBPQCon );
var
  s: string;
begin
  s := 'CREATE TABLE molinos ( nid serial PRIMARY KEY, nid_curvaAero integer, ' +
    'nid_parque integer, UTM_zona double precision, ' +
    'UTM_uso double precision, UTM_xE double precision, ' +
    'UTM_yN double precision, ro double precision ) ';
  db.exec(s);
end;

constructor TMolino.CreateLoadFromDB(r: TSQLQuery);
begin
  nid := r.FieldByName('nid').AsInteger;
  nid_curvaAero := r.FieldByName('nid_curvaAero').AsInteger;
  nid_Parque := r.FieldByName('nid_parque').AsInteger;
  //  curva:= TCur
  raise Exception.Create('FALTA IMPLEMENTAR');
end;

function TMolino.InsertIntoDB: integer;
begin
  raise Exception.Create('FALTA IMPLEMENTAR');
end;

procedure TMolino.UpdateDB;
begin
  raise Exception.Create('FALTA IMPLEMENTAR');
end;


// Este procedimiento calcula la potencia electrica del molino usando
// la curva del mismo. Para la potencia del viento incidente toma el promedio
// de las potencia de viento de los puntos de la grilla rotada asociados al molino.
// el procedimiento afecta luego del calculo los valores de PV en la grilla rotada
// de forma que reflejen la potencia eléctrica extraída.x
// Pasamos la dirección del viento (ang) para que pueda calcular el valor de ro
procedure TMolino.CalcularAFectacionEnGrillaRotada(kfila_: integer;
  fila: TVectR; ang, fpv3pe, relRed: NReal);

var
  pv_E: NReal;
  j: integer;
  rel: NReal;
  jmed: NReal;
  max_delta_pv: NReal;
begin
  if kFila_ <> self.posicion.kFila then
    exit;

  pv_prom := 0;
  for j := posicion.jCol1 to posicion.jCol2 do
    pv_prom := pv_prom + fila.e(j);

  pv_prom := pv_prom / (posicion.jCol2 - posicion.jCol1 + 1);

  self.PE_MW := curva.P_MW(pv_prom * ro) * relRed;

  pv_E := PE_MW / curva.AreaRotor;

  if pv_prom < 1E-15 then
    rel := 0.01
  else
    rel := (pv_E * fpv3pe) / pv_prom;
  rel := 1 - rel;
  rel := max(0.00001, min(0.999999, rel));

  jMed := (posicion.jCol1 + posicion.jCol2) / 2;

  for j := posicion.jCol1 to posicion.jCol2 do
    fila.pon_e(j, fila.e(j) * rel);
end;



procedure TMolino.Free;
begin
  curva.Free;
  inherited Free;
end;


function TMolino.P_MW(ang_gr, v3Libre: NReal): NReal;
var
  V3Local: NReal;
begin
  //  V3Local := afectacion_v3.v3_LibreToLocal(ang_gr, v3Libre);
  Result := curva.P_MW(v3Local);
end;

constructor TParque.Create_Init(aNombre: string; aCantMolinos, aNSectoresDir: integer;
  modeloAero, archi_curvaAero: string;
  //afectacion_v3: TAnemometro;
  alfa_terreno, gamma_atmosfera, fpv3pe: NReal;
  Aero_xmin_, Aero_xmax_, Aero_ymin_, Aero_ymax_: NReal;
  // rectangulo en coordenadas cartesianas  que incluye TODOS los AEROS
  // con el ANEMO en el (0,0)
  npuntos_x, npuntos_y: integer;
  grillaRotada_dx_, grillaRotada_dy_, grillaRotada_MargenDx_: NReal;
  // resoluciòn de la grilla rotada
  FactorPerdidas, PotenciaNominal, LimiteDePotencia: NReal);

var
  Molino: TMolino;
  i: integer;
  kFil, jCol: integer;
  dx, dy: NReal;

begin
  Nombre := aNombre;
  CantMolinos := aCantMolinos;
  NSectoresDir := aNSectoresDir;
  //inicializo por defecto todos los molinos al mismo tipo
  Aero_xmin := Aero_xmin_;
  Aero_xmax := Aero_xmax_;
  Aero_ymin := Aero_ymin_;
  Aero_ymax := Aero_ymax_;
  grillaRotada_dx := grillaRotada_dx_;
  grillaRotada_dy := grillaRotada_dy_;
  grillaRotada_MargenDx := grillaRotada_MargenDx_;

  self.alfa_terreno := alfa_terreno;
  self.gamma_atmosfera := gamma_atmosfera;
  self.fpv3pe := fpv3pe;

  Self.FactorPerdidas := FactorPerdidas;
  Self.PotenciaNominal := PotenciaNominal;
  Self.LimiteDePotencia := LimiteDePotencia;

  dx := (Aero_xMax - Aero_xmin) / (NPuntos_x - 1);
  dy := (Aero_ymax - aero_ymin) / (NPuntos_y - 1);

  setlength(molinos, CantMolinos);

  kFil := 0;
  jCol := 0;
  for i := 0 to CantMolinos - 1 do
  begin
    Molino :=
      TMolino.Create_Init(Aero_xmin + jCol * dx, Aero_ymin + kFil *
      dy, modeloAero, archi_curvaAero, NSectoresDir);

    Molino.ro := 1;
    self.Molinos[i] := Molino;

    Inc(jCol);
    if jCol = npuntos_x then
    begin
      jCol := 0;
      Inc(kFil);
    end;

  end;
  Meteo_ro := 1;
  grillaRotada := nil;
end;



procedure TParque.Calc_Auxiliares;
begin
  LimiteDePotencia := max(LimiteDePotencia, 1.0e-8);
  k_Perdidas := FactorPerdidas / PotenciaNominal * (1 - FactorPerdidas);
end;



constructor TParque.Create_LoadFromFile(var f: textfile);
var
  i: integer;
  tipoA: string;
  version: integer;
  r: string;
begin
  inherited Create;
  readln(f, r);
  if pos('VERSION=', r) = 1 then
  begin
    Delete(r, 1, length('VERSION='));
    version := StrToInt(r);
    readln(f, Nombre);
  end
  else
  begin
    version := -1;
    nombre := r;
  end;

  readln(f, CantMolinos);
  readln(f, NSectoresDir);

  readln(f, Aero_xmin);
  readln(f, Aero_xmax);
  readln(f, Aero_ymin);
  readln(f, Aero_ymax);


  readln(f, alfa_terreno);
  readln(f, gamma_atmosfera);
  readln(f, fpv3pe);

  readln(f, FactorPerdidas);
  readln(f, LimiteDePotencia);
  if version <= 1 then
    PotenciaNominal := LimiteDePotencia
  else
    readln(f, PotenciaNominal);

  readln(f, Meteo_ro);

  readln(f, grillaRotada_dx);
  readln(f, grillaRotada_dy);
  readln(f, grillaRotada_MargenDx);

  setlength(molinos, CantMolinos);

  for i := 0 to CantMolinos - 1 do
    Molinos[i] := TMolino.Create_LoadFromFile(f);

  readln(f, Meteo_UTM_Zona);
  readln(f, Meteo_UTM_uso);
  readln(f, Meteo_UTM_xe);
  readln(f, Meteo_UTM_yn);

  Calc_Auxiliares;
end;

procedure TParque.StoreInFile(var f: textfile);
var
  i: integer;
begin
  writeln(f, 'VERSION=' + IntToStr(VERSION_ARCHI_PARQUE));
  writeln(f, Nombre);
  writeln(f, CantMolinos);
  writeln(f, NSectoresDir);
  writeln(f, Aero_xmin);
  writeln(f, Aero_xmax);
  writeln(f, Aero_ymin);
  writeln(f, Aero_ymax);

  writeln(f, alfa_terreno);
  writeln(f, gamma_atmosfera);
  writeln(f, fpv3pe);
  writeln(f, FactorPerdidas);
  writeln(f, LimiteDePotencia);
  writeln(f, PotenciaNominal);
  writeln(f, Meteo_ro);

  writeln(f, grillaRotada_dx);
  writeln(f, grillaRotada_dy);
  writeln(f, grillaRotada_MargenDx);

  for i := 0 to CantMolinos - 1 do
    Molinos[i].StoreInFile(f);

  writeln(f, Meteo_UTM_Zona);
  writeln(f, Meteo_UTM_uso);
  writeln(f, Meteo_UTM_xe);
  writeln(f, Meteo_UTM_yn);
end;

procedure TParque.Free;
var
  k: integer;
begin

  for k := 0 to CantMolinos - 1 do
  begin
    Molinos[k].Free;
  end;

  setlength(Molinos, 0);

  if grillaRotada <> nil then
    grillaROtada.Free;
  inherited Free;
end;


procedure TParque.SetCurvaAero(kAero: integer; aCurva: TCurvaAerogen);
begin
  Molinos[kAero].curva := aCurva;
end;




procedure TParque.Calcular_pv_fila(filaACalcular, filaAnterior: TVectR);
var
  j: integer;
begin
  for j := 2 to filaACalcular.n - 1 do
    filaACalcular.pon_e(j,
      (filaAnterior.e(j) * beta + (filaAnterior.e(j - 1) + filaAnterior.e(j + 1)) *
      alfa) * gamma_inf + pv_Libre * gamma_sup);
end;

function TParque.ParametrosPtr: TDAOfNRealPtr;
var
  res: TDAOfNRealPtr;
  kMolino: integer;
  nMolinos: integer;
begin
  nMolinos := length(self.Molinos);
  setlength(res, 6 + nMolinos);

  res[0] := @alfa_terreno;
  res[1] := @gamma_atmosfera;
  res[2] := @fpv3pe;
  res[3] := @FactorPerdidas;
  res[4] := @LimiteDePotencia;
  res[5] := @Meteo_ro;

  for kMolino := 0 to nMolinos - 1 do
    res[6 + kMolino] := @Molinos[kMolino].ro;
  Result := res;
end;


procedure TParque.AerosAfectarFila(kFila: integer; ang, relRed: NReal);
var
  kMolino: integer;
begin
  for kMolino := 0 to high(molinos) do
    molinos[kmolino].CalcularAFectacionEnGrillaRotada(
      kFila, GrillaRotada.Fila(kFila), ang, fpv3pe, relRed);

end;

procedure TParque.SetMeteo_UTM(uso: integer; Zona: char; xE, yN: Nreal);
begin
  Meteo_UTM_uso := uso;
  Meteo_UTM_Zona := Zona;
  Meteo_UTM_xe := xE;
  Meteo_UTM_yN := yN;
end;

procedure TParque.SetAero_UTM(kMolino: integer; uso: integer; Zona: char; xE, yN: Nreal);
begin
  molinos[kMolino - 1].posicion.Set_UTM(uso, zona, xE, yN);
end;

procedure TParque.RecalcularCoordenadas;
var
  kMolino: integer;
  aMolino: TMolino;
begin
  for kMolino := 0 to high(molinos) do
  begin
    aMolino := molinos[kMolino];
    aMolino.posicion.z_base :=
      cinit(aMolino.posicion.UTM_xE - Meteo_UTM_xe, aMolino.posicion.UTM_yN -
      Meteo_UTM_yN);
  end;
end;


procedure TParque.SetPixFrame(xp1, yp1, xp2, yp2: integer);
var
  aP: complex;
  kMolino: integer;
  dd: NReal;
begin
  self.xp1 := xp1;
  self.yp1 := yp1;
  self.xp2 := xp2;
  self.yp2 := yp2;

  dxp := xp2 - xp1;
  dyp := yp2 - yp1;

  Show_xmin := 0;
  Show_xmax := 0;
  Show_ymin := 0;
  Show_ymax := 0;

  for kMolino := 0 to high(molinos) do
  begin
    aP := molinos[kmolino].posicion.z_base;
    if aP.re < Show_xmin then
      show_xmin := aP.re
    else if aP.re > show_xmax then
      show_xmax := aP.Re;

    if ap.im < show_ymin then
      show_ymin := ap.im
    else if ap.im > show_ymax then
      show_ymax := ap.im;
  end;


  show_xmin := show_xmin - show_padding;
  show_xmax := show_xmax + show_padding;
  dxa := (show_xmax - show_xmin);

  show_ymin := show_ymin - show_padding;
  show_ymax := show_ymax + show_padding;
  dya := (show_ymax - show_ymin);


  if (dxp / dxa) > (dyp / dya) then
  begin
    // hay más píxeles/m en la horizontal  thengo que aumentar el ancho en m
    // para lograr la misma definición que la vertical
    dd := (-dxa + dya / dyp * dxp) / 2;
    show_xmin := show_xmin - dd;
    show_xmax := show_xmax + dd;
    dxa := (show_xmax - show_xmin);
  end
  else
  begin
    // hay más píxeles/m en la vertical
    dd := (-dya + dxa / dxp * dyp) / 2;
    show_ymin := show_ymin - dd;
    show_ymax := show_ymax + dd;
    dya := (show_ymax - show_ymin);
  end;

end;

procedure TParque.ZToPix(var xp, yp: integer; z: Complex);
begin
  xp := xp1 + round((dxp / dxa) * (z.re - show_xmin));
  yp := yp2 - round((dyp / dya) * (z.im - show_ymin));
end;

procedure TParque.PixToZ(var z: Complex; xp, yp: integer);
begin
  z.re := show_xmin + (dxa / dxp) * (xp - xp1);
  z.im := show_ymin + (dya / dyp) * (yp2 - yp);
end;


// busca el molino que esté más proximo al punto xp, yp y lo devuelve
// si el mismo está a menos de un diametro del punto xp, yp
function TParque.MolinoAt_xpyp(xp, yp: integer): TMolino;
var
  res, aMolino: TMolino;
  kMolino: integer;
  d, min_d: NReal;
  dz, z: Complex;

begin
  PixToZ(z, xp, yp);
  aMolino := Molinos[0];
  d := cmod(z - aMolino.posicion.z_base);
  min_d := d;
  res := aMolino;
  for kMolino := 1 to high(Molinos) do
  begin
    aMolino := Molinos[kMolino];
    d := cmod(z - aMolino.posicion.z_base);
    if d < min_d then
    begin
      min_d := d;
      res := aMolino;
    end;
  end;
  if min_d <= (3 * res.curva.Diametro) then
    Result := res
  else
    Result := nil;
end;


function TParque.GridValAtZ(z: complex): NReal;
var
  zRot: complex;
  kr, jr: NReal;
  k1, k2, j1, j2: integer;
  alfa_k, alfa_j, beta_k, beta_j: NReal;
begin
  zRot := z * self.z_BaseToRot;
  if (zRot.re < Caja_xmin) or (zRot.re > Caja_xmax) or (zRot.im < Caja_ymin) or
    (zRot.im > Caja_ymax) then
  begin
    Result := 0;
    exit;
  end;
  kr := (zRot.im - Caja_ymin) / Caja_dy;
  jr := (zRot.re - Caja_xmin) / Caja_dx;
  k1 := trunc(kr) + 1;
  k2 := k1 + 1;
  if k2 > grillaRotada.nf then
    k2 := grillaRotada.nf;
  j1 := trunc(jr) + 1;
  j2 := j1 + 1;
  if j2 > grillaRotada.nc then
    j2 := grillaRotada.nc;
  alfa_k := frac(kr);
  alfa_j := frac(jr);
  beta_k := 1 - alfa_k;
  beta_j := 1 - alfa_j;


  Result := beta_k * (beta_j * grillaRotada.e(k1, j1) + alfa_j *
    grillaRotada.e(k1, j2)) + alfa_k * (beta_j * grillaRotada.e(k2, j1) +
    alfa_j * grillaRotada.e(k2, j2));

end;

function TParque.GridValAtPix(xp, yp: integer): NReal;
var
  z: complex;
begin
  PixToZ(z, xp, yp);
  Result := GridValAtZ(z);
end;


procedure TParque.Calc_PotMW_init_(ang: NReal);
var
  kMolino: integer;
  aMolino: TMolino;
  ndivx, ndivy: integer;
  k, j: integer;
  res: NReal;
  xij, Ct, a, rd, vdj, alfa2: NReal;

begin
  Calc_Auxiliares;

  z_BaseToRot := cinit(cos(2 * pi * (ang + 180) / 360),
    sin(2 * pi * (ang + 180) / 360));
  Caja_xmin := 0;
  Caja_ymin := 0;
  Caja_xmax := 0;
  Caja_ymax := 0;

  for kMolino := 0 to high(Molinos) do
  begin
    aMolino := Molinos[kMolino];
    aMolino.posicion.z_rotado := aMolino.posicion.z_base * z_BaseToRot;
    if aMolino.posicion.z_rotado.re < Caja_xmin then
      Caja_xmin := aMolino.posicion.z_rotado.re
    else if aMolino.posicion.z_rotado.re > Caja_xmax then
      Caja_xmax := aMolino.posicion.z_rotado.re;
    if aMolino.posicion.z_rotado.im < Caja_ymin then
      Caja_ymin := aMolino.posicion.z_rotado.im
    else if aMolino.posicion.z_rotado.im > Caja_ymax then
      Caja_ymax := aMolino.posicion.z_rotado.im;
  end;

  Caja_xmin := Caja_xmin - grillaRotada_MargenDx;
  Caja_xmax := Caja_xmax + grillaRotada_MargenDx;

  ndivx := trunc((Caja_xmax - Caja_xmin) / grillaRotada_dx + 0.999);
  Caja_dx := (Caja_xmax - Caja_xmin) / ndivx;

  ndivy := trunc((Caja_ymax - Caja_ymin) / grillaRotada_dy + 0.999);
  Caja_dy := (Caja_ymax - Caja_ymin) / ndivy;


  if GrillaRotada <> nil then
    grillaRotada.Free;

  grillaRotada := TMatR.Create_Init(ndivy + 1, ndivx + 1);

  for kMolino := 0 to high(Molinos) do
  begin
    aMolino := Molinos[kMolino];
    aMolino.posicion.kFila :=
      round((aMolino.posicion.z_rotado.im - Caja_ymin) / Caja_dy) + 1;
    aMolino.posicion.jCol1 :=
      round((aMolino.posicion.z_rotado.re - aMolino.curva.Diametro / 2 - Caja_xmin) /
      Caja_dx) + 1;
    aMolino.posicion.jCol2 :=
      round((aMolino.posicion.z_rotado.re + aMolino.curva.Diametro / 2 - Caja_xmin) /
      Caja_dx) + 1;
  end;
end;


function TParque.Calc_PotMW(pv3: NReal): NReal;
var
  kMolino: integer;
  aMolino: TMolino;
  ndivx, ndivy: integer;
  k, j: integer;
  res: NReal;
  xij, Ct, a, rd, vdj, alfa2: NReal;
  relRed: NReal;
  cnt_Iter_relRed: integer;

label
  lbl_Recalcular;

begin
  pv_Libre := pv3;

  // cálculo de los pesos de afectación.


  // Alfa es el peso de los laterales y beta = 1- 2* alfa el peso del central.
  // Al aumentar la dy en relación a dx tiene que terminar pensando todos por igual alfa = 1/3
  alfa := (1 / 2) * (1 - exp(-alfa_terreno * Caja_dy / Caja_dx));
  beta := 1 - 2 * alfa;

  //gamma_sup es el peso de la Velocidad ARRIBA (a 500 m se supone = pvLIBRE) y gamma_inf es a
  gamma_sup := (1 - exp(-gamma_atmosfera * Caja_dy / 500.0));
  gamma_inf := 1 - gamma_sup;

  relRed := 1.0;
  cnt_Iter_relRed := 0;

  lbl_Recalcular:

      // Inicialización de las condiciones de borde.
    for j := 1 to grillaRotada.nc do
      GrillaRotada.pon_e(1, j, pv3);
  for k := 2 to grillaRotada.nf do
  begin
    GrillaRotada.pon_e(k, 1, pv3);
    GrillaRotada.pon_e(k, grillaRotada.nc, pv3);
  end;

  self.AerosAfectarFila(1, ang, relRed);

  for k := 2 to grillaRotada.nf do
  begin
    Calcular_pv_fila(grillaRotada.fila(k), grillaRotada.fila(k - 1));
    self.AerosAfectarFila(k, ang, relRed);
  end;

  //  GrillaRotada.WriteArchiXLT( 'c:\basura\interfer_pv3.xlt' );
  res := 0;

  for kMolino := 0 to high(Molinos) do
    res := res + molinos[kMolino].PE_MW;

  res := res - k_Perdidas * sqr(res);
  if res > LimiteDePotencia then
  begin
    relred := LimiteDePotencia / res * relRed;
    if cnt_Iter_relRed < 10 then
    begin
      Inc(cnt_Iter_RelRed);
      goto lbl_Recalcular;
    end;

    for kMolino := 0 to high(Molinos) do
      molinos[kMolino].PE_MW := molinos[kMolino].PE_MW * relRed;
    res := LimiteDePotencia;
  end;
  Result := res;
end;


function TParque.Calc_PotMW_anemo(pv3_anemo: NReal): NReal;
var
  convergio: boolean;
  cnt_iter: integer;
  pv3, PE_MW_Modelo, pv3Anemo_a, pv3Anemo_s, rel: NReal;
begin
  pv3anemo_a := pv3_anemo / Meteo_ro;

  if pv3anemo_a < 1E-10 then
  begin
    Result := 0;
    exit;
  end;

  pv3 := pv3anemo_a;
  convergio := False;
  cnt_iter := 0;
  repeat
    PE_MW_Modelo := Calc_PotMW(pv3);
    pv3Anemo_s := GridValAtZ(_0);
    rel := pv3Anemo_a / pv3Anemo_s;
    if (abs(rel - 1) < 0.01) then
      convergio := True
    else
    begin
      Inc(cnt_iter);
      pv3 := rel * pv3;
    end;
  until (cnt_iter > 10) or convergio;
  Result := PE_MW_Modelo;
end;


end.
