unit uparques_modeloteorico;

{$DEFINE FELIPE}


//{$DEFINE MODELO TEORICO MULTIPLE}

interface

uses
  Classes, SysUtils, Math, matreal, xmatdefs,
  ucomplex,
  uro_aire,
  {$IFNDEF SIN_DBCON}
  {$IFDEF pq_direct}
  udbconpg,
  {$ELSE}
  urosx,
  {$ENDIF}
  {$ENDIF}
  ucurvasaeros,
  uauxiliares;

const
  VERSION_ARCHI_PARQUE = 3;

  vCorte = 12;
  pvCorte = vCorte * Vcorte * vcorte * 0.5 * 1.2 / 1.0E6;

  vCorte_0 = 2;
  pvCorte_0 = vCorte_0 * Vcorte_0 * vcorte_0 * 0.5 * 1.2 / 1.0E6;

  vCorte_Sup = 20;
  pvCorte_Sup = vCorte_Sup * Vcorte_Sup * vcorte_Sup * 0.5 * 1.2 / 1.0E6;

  show_padding_ = 300; // banda en metros al borde visible


type
  {$IFNDEF SIN_DBCON}
  {$IFNDEF pq_direct}
  TDBPQCon = TDBrosxCon;
  {$ENDIF}
  {$ENDIF}


  TPosicion = class
    // Coordenadas en sistema  Universal Transversa de Mercator (UTM)
    // Uuruguay: elipsoide GRS80, Huso 21 y 22, Zonas H y J.
    UTM_Huso: integer;
    UTM_Zona: char;
    UTM_xE, UTM_yN: NReal;
    cota: NReal;

    // posición en la grilla sin rotar  ( x: Oeste-> Este, y :  Sur->Norte )
    z_base: Complex;

    // Calculados con cada nueva dirección de viento al generar la grilla rotada.
    z_rotado: Complex; // posición en la grilla rotada

    kFila: integer; // fila en la grilla de cálculo rotada
    jCol1, jCol2: integer; // primer y última columna en la grilla rotada

    constructor Create_Cartesianas(x, y: NReal);
    constructor Create_UTM(xHuso: integer; xZona: char; xe, yn: NReal);
    procedure Set_UTM(huso: integer; Zona: char; xE, yN: Nreal);
    constructor Create_LoadFromFile(var f: textfile);
    procedure StoreInFile(var f: textfile);

    function UTM_ZonaHusoStr: string;
  end;


  TMolino = class
    nid: integer;
    nro: integer; // identificador dentro de la central.
    nid_curvaAero: integer;
    nid_central: integer;

    curva: TCurvaAerogen;
    etha: NReal;

    posicion: TPosicion;

    // Estas variables son calculadas por el Parque.
    pv_prom: NReal; // Potencia de viento de acuerdo a la dirección actual
    PE_MW: NReal; // generando de acuerdo a la dirección actual

    constructor Create_Init(x, y: NReal; modeloAero, archi_curvaAero: string;
      nSectoresRo: integer);

    constructor Create_LoadFromFile(var f: textfile);

    procedure StoreInFile(var f: textfile);

    procedure Free;

    // Este procedimiento calcula la potencia electrica del molino usando
    // la curva del mismo. Para la potencia del viento incidente toma el promedio
    // de las potencia de viento de los puntos de la grilla rotada asociados al molino.
    // el procedimiento afecta luego del calculo los valores de PV en la grilla rotada
    // de forma que reflejen la potencia eléctrica extraída.x
    procedure CalcularAFectacionEnGrillaRotada(kfila_: integer;
      fila: TVectR; ang, fpv3pe, relRed, roDato: NReal);

    {$IFNDEF SIN_DBCON}    class procedure CrearTablaDB(db: TDBPQCon);
    constructor CreateLoadFromDB(db: TDBPQCon; r: TSQLQuery);
    function InsertIntoDB(db: TDBPQCon): integer;
    procedure UpdateDB(db: TDBPQCon);
    procedure UpdateDB_Modelo(db: TDBPQCon);
    {$ENDIF}
  end;


  TDAOFMolino = array of TMolino;

  { TParque_ModeloTeorico }

  TParque_ModeloTeorico = class
    nid_central: integer;
    nid_modelo: integer;

    Nombre: string;
    CantMolinos: integer;
    N_Sectores_Dir: integer;
    Molinos: TDAOFMolino;

    // Curva_s representativa de la SUMA de las curvas de los aerogeneradores
    // para su funcionamiento entre 0 y PMáx. Se utiliza solo para tener una figura
    // de cómo sería la respuesta del parque sin considerar efectos de estela ni de
    // afectación de velocidades por posición.
    Curva_s: TCurvaAerogen;

    // auxiliar para mapeo a pixels en CANVAS
    xp1, yp1, xp2, yp2: integer;
    dxa, dya: NReal; // ancho y alto en m
    dxp, dyp: integer; // ancho y alto en px
    show_xmin, show_xmax, show_ymin, show_ymax: NReal; // caja en m

    Meteo_UTM_huso: integer;
    Meteo_UTM_Zona: char;
    Meteo_UTM_xe, Meteo_UTM_yn: NReal;
    Meteo_cota: NReal;

    // Extensiòn del rectángulo para incluir el Anemómetro.
    Caja_xmin, Caja_xmax, Caja_ymin, Caja_ymax: NReal;
    Caja_dx, Caja_dy: NReal; // anchos de la caja

    z_BaseToRot: complex; // exp(j (alfa+90) )

    grilla_Rotada_dx, grilla_Rotada_dy, grilla_Rotada_Margen_Dx: NReal;
    grillaRotada: TMatR;

  //=============================================
  // PARAMETROS DE CALIBRACION GLOBAL DEL PARQUE
  //---------------------------------------------
    Meteo_etha: TVectR; // a0 , a1 ... an , b1 ... bn desarrollo de Fourier etha( ang )
    alfa_terreno: NReal;  // Mezclador horizontal calculado para grilla de 1m
    gamma_atmosfera: NReal; // Mezclador vertical calculado para grilla de 1m
    fpv3pe: NReal; // factor de perdida delta_pv3 = fpv3pe * PE
    Factor_Perdidas: NReal; // factor de pérdidas eléctricas. Especificado a la potencia nominal.
  //---------------------------------------------

    ang: NReal; // ángulo último cálculo
    pv_libre: NReal;  // pv_libre Ultimo cálculo

    k_Perdidas: NReal; // valor calculado a partir del Factor y el límite

    // Potencia Nominal.
    Potencia_Nominal: NReal;


    // Limitación de la potenica inyectada de acuerdo al contrato
    // de conección. Se supone que el parque controla siempre
    // la potencia inyectada para no superar este valor.
    Potencia_Autorizada_MW: NReal;


    // valor efectivo para la dirección actual. Se calcula en base
    // a Meteo_etha para la dirección actual en Calc_PotMW_init_
    Meteo_etha_dir: NReal;

    constructor Create_Init(aNombre: string; aCantMolinos,
      aN_Sectores_Dir: integer; modeloAero, archi_curvaAero: string;
  alfa_terreno, gamma_atmosfera, fpv3pe: NReal;
  // rectangulo en coordenadas cartesianas  que incluye TODOS los AEROS
  // con el ANEMO en el (0,0). Esto es para una defición "artificial" junto con los
  // parámetros npuntos_* siguientes
  Aero_xmin_, Aero_xmax_, Aero_ymin_, Aero_ymax_: NReal;
  // Posiciones imaginaria de los aeros para una inicialización
  // artificial del Modelo plantando los aeros en una grilla de npuntos_x por n_puntos_y
  // estos parámetros no tienen utilidad real dado que se conoce la posición de los aeros
  npuntos_x, npuntos_y: integer;
 grilla_Rotada_dx_, grilla_Rotada_dy_, grilla_Rotada_Margen_Dx_: NReal;
  Factor_Perdidas, Potencia_Nominal, Potencia_autorizada_MW: NReal);


    constructor Create_LoadFromFile(var f: textfile);
    constructor Create_LoadFromArchi( const archi: string );

    {$IFNDEF SIN_DBCON}
    constructor Create_LoadFromDB( db: TDBPQCon; r_central: TSQLQuery );
    constructor CreateMultiParque_LoadFromDB(db: TDBPQCon; nid_centrales: TDAofNInt;nid_central_anemo:integer);
    procedure Update( db: TDBPQCon );
    {$ENDIF}

    procedure StoreInFile(var f: textfile);
    procedure StoreInArchi(const archi: string );
    procedure SetCurvaAero(kAero: integer; aCurva: TCurvaAerogen);

    procedure Free;

    {function get_params_ptrs: TDAOfNReal;}

    procedure SetMeteo_UTM( huso: integer; Zona: char; xE, yN: Nreal);
    procedure SetAero_UTM(kMolino: integer; uso: integer; Zona: char; xE, yN: Nreal);
    procedure RecalcularCoordenadas;

    // Fija las coordenada despleglables y el frame en pixeles
    procedure SetPixFrame(xp1, yp1, xp2, yp2: integer);

    procedure ZToPix(var xp, yp: integer; z: Complex);
    procedure PixToZ(var z: Complex; xp, yp: integer);

    // busca el molino que esté más proximo al punto xp, yp y lo devuelve
    // si el mismo está a menos de un diametro del punto xp, yp
    function MolinoAt_xpyp(xp, yp: integer): TMolino;

    function GridValAtZ(z: complex): NReal;
    function GridValAtPix(xp, yp: integer): NReal;

    procedure Calc_PotMW_init_(ang: NReal);

    function Calc_PotMW_pvLibre(roDato, pv3, cgm: NReal): NReal;

    function Calc_PotMW_anemo(roDato,pv3_anemo: NReal; cgm: NReal;  var flg_Convergio: boolean;
      var pv3_Anemo_s, pv3_VientoLibre: NReal): NReal;

    procedure Calcular_pv_fila(filaACalcular, filaAnterior: TVectR);

    function ParametrosPtr: TDAOfNRealPtr;

    procedure Calc_Auxiliares;

  private
    // factor perturbante horizontal 0.5 *( dx/dy) * (1-exp( - alfa_terreno*dx )) de la grilla rotada
      alfa_lateral: NReal;
      alfa_central: NReal; // 1 - 2 * alfa

      gamma_sup: NReal;  // (dx / 300)* (1- exp( - gamma_atmosfera *dx ))
      gamma_central: NReal;  // 1- gamma_sup


    procedure AerosAfectarFila(kFila: integer; ang, relRed,roDato: NReal);


  end;

{$IFDEF MODELO TEORICO MULTIPLE}
TDAOFTParque_ModeloTeorico = array of TParque_ModeloTeorico;

  { TParque_ModeloTeorico Múltiple}

{ TParque_ModeloTeorico_Multiple }

TParque_ModeloTeorico_Multiple = class

 Parques_MT: TDAOFTParque_ModeloTeorico;

 GrillaRotada:TMatR;

 function Calc_PotMW_anemo(roDato, pv3_anemo: NReal; cgm_vect: TVectR;
   var flg_Convergio: boolean; var pv3_Anemo_s, pv3_VientoLibre: NReal): TVectR;

 procedure Calc_PotMW_init_(ang: NReal);

 function Calc_PotMW_pvLibre(roDato, pv3, cgm: NReal): NReal;


end;



implementation

{ TParque_ModeloTeorico_Multiple }

function TParque_ModeloTeorico_Multiple.Calc_PotMW_anemo(roDato,
  pv3_anemo: NReal; cgm_vect: TVectR; var flg_Convergio: boolean; var pv3_Anemo_s,
  pv3_VientoLibre: NReal): TVectR;
  var

  kMolino,cant_centrales: integer;
  k, j: integer;
  res: NReal;
  Pmax: NReal;
  cnt_Iter_relRed: integer;
  LimitePot,relRed: TVectR;
label
  lbl_Recalcular;

begin

  cant_centrales:=cgm_vect.n;
  LimitePot:=TVectR.Create_Init(cant_centrales);

  for k:=0 to cant_centrales-1 do
  begin
    Pmax:=Parques_MT[k].Potencia_Autorizada_MW;
    Pmax:=min(Pmax,cgm_vect.e(k+1));
    LimitePot.pon_e(k,Pmax);
  end;

  pv_Libre := pv3;

  // cálculo de los pesos de afectación.

  relRed :=TVectR.Create_Init(cant_centrales);
  relRed.Ceros;
  relRed.MasReal(1);

  cnt_Iter_relRed := 0;

  lbl_Recalcular:

  // Inicialización de las condiciones de borde.
  for j := 1 to grillaRotada.nc do
    GrillaRotada.pon_e(1, j, pv3);

  for k := 2 to grillaRotada.nf do
  begin
    GrillaRotada.pon_e(k, 1, pv3);
    GrillaRotada.pon_e(k, grillaRotada.nc, pv3);
  end;

  for k:=1 to cant_centrales do
    Parques_MT[k-1].AerosAfectarFila(1, ang, relRed.e(k), roDato);//JF:cambiar esta funcion para que considere grilla conjunta.

  for k := 2 to grillaRotada.nf do
  begin
    Calcular_pv_fila(grillaRotada.fila(k), grillaRotada.fila(k - 1));
    self.AerosAfectarFila(k, ang, relRed, roDato);
  end;

  //  GrillaRotada.WriteArchiXLT( 'c:\basura\interfer_pv3.xlt' );
  res := 0;

  for kMolino := 0 to high(Molinos) do
    res := res + molinos[kMolino].PE_MW;

  res := max(0, res - k_Perdidas * sqr(res));
  if res > LimitePot then
  begin
    relred := LimitePot / res * relRed;
    if cnt_Iter_relRed < 10 then
    begin
      Inc(cnt_Iter_RelRed);
      goto lbl_Recalcular;
    end;

    for kMolino := 0 to high(Molinos) do
      molinos[kMolino].PE_MW := molinos[kMolino].PE_MW * relRed;
    res := LimitePot;
  end;
  Result := res;

end;

procedure TParque_ModeloTeorico_Multiple.Calc_PotMW_init_(ang: NReal);
begin

end;

function TParque_ModeloTeorico_Multiple.Calc_PotMW_pvLibre(roDato, pv3,
  cgm: NReal): NReal;
begin

end;

{$ENDIF}

implementation


procedure TPosicion.Set_UTM(huso: integer; Zona: char; xE, yN: Nreal);
begin
  self.UTM_Huso := huso;
  self.UTM_Zona := Zona;
  self.UTM_xE := xE;
  self.UTM_yN := yN;
end;

constructor TPosicion.Create_Cartesianas(x, y: NReal);
begin
  inherited Create;
  self.z_base := cinit(x, y);
end;

constructor TPosicion.Create_UTM(xHuso: integer; xZona: char; xe, yn: NReal);
begin
  inherited Create;
  self.z_base := _0;
  UTM_Huso := xHuso;
  UTM_Zona := xZona;
  UTM_xE := xe;
  UTM_yN := yn;
end;

constructor TPosicion.Create_LoadFromFile(var f: textfile);
var
  x, y: NReal;
  r: string;
begin
  readln(f, x, y);
  Create_Cartesianas(x, y);
  readln(f, r);
  UTM_Huso := nextint(r);
  UTM_Zona := nextpal(r)[1];
  UTM_xE := nextfloat(r);
  UTM_yN := nextfloat(r);
end;

procedure TPosicion.StoreInFile(var f: textfile);
begin
  writeln(f, z_base.re, #9, z_base.im);
  writeln(f, UTM_Huso, #9, UTM_Zona, #9, UTM_xE, #9, UTM_yN);
end;

function TPosicion.UTM_ZonaHusoStr: string;
begin
  result:= UTM_Zona+ IntToStr( UTM_huso );
end;

constructor TMolino.Create_Init(x, y: NReal; modeloAero, archi_curvaAero: string;
  nSectoresRo: integer);
  //  afectacion_v3: TAnemometro);
begin
  inherited Create;
  nid := -1;
  nid_curvaAero := -1;
  nid_central := -1;
  posicion := TPosicion.Create_cartesianas(x, y);
  //  self.afectacion_v3 := afectacion_v3;
  self.curva := TCurvaAerogen.CreateLoadFromArchi(modeloAero, archi_curvaAero);
  etha := 1.0;
end;


constructor TMolino.Create_LoadFromFile(var f: textfile);
var
  tipoA: string;

begin
  inherited Create;
  readln( f, nid );
  readln( f, nid_curvaAero );
  readln( f, nid_central );
  posicion := TPosicion.Create_LoadFromFile(f);
  self.curva := TcurvaAerogen.CreateLoadFromFile(f);
  readln(f, etha);
end;


procedure TMolino.StoreInFile(var f: textfile);
begin
  writeln( f, nid );
  writeln( f, nid_curvaAero );
  writeln( f, nid_central );
  posicion.StoreInFile(f);
  self.curva.StoreInFile(f);
  writeln(f, etha);
end;


{$IFNDEF SIN_DBCON}
class procedure TMolino.CrearTablaDB(db: TDBPQCon);
var
  s: string;
begin
  s := 'CREATE TABLE molinos ( nid serial PRIMARY KEY, nid_curvaAero integer, ' +
    'nidcentral integer, zona_utm double precision, ' +
    'huso_utm double precision, xe double precision, ' +
    'yn double precision, cota double precision, etha double precision ) ';
  DB.exec(s);
end;

constructor TMolino.CreateLoadFromDB(db: TDBPQCon; r: TSQLQuery);
begin
  inherited Create;
  nid := r.FieldByName('nid').AsInteger;
  nid_curvaAero := r.FieldByName('nid_curvaaero').AsInteger;
  nid_central := r.FieldByName('nidcentral').AsInteger;
  nro:= r.FieldByName('nro').AsInteger;
  curva := TCurvaAerogen.CreateLoadFromDB_nid(db, nid_curvaAero);
  etha := r.FieldByName('etha').AsFloat;
  posicion := TPosicion.Create_UTM(r.FieldByName('huso_utm').AsInteger,
    r.FieldByName('zona_utm').AsString[1], r.FieldByName('xe').AsFloat,
    r.FieldByName('yn').AsFloat);
end;

function TMolino.InsertIntoDB(db: TDBPQCon): integer;
var
  s: string;
begin
  nid := DB.nextval('molinos');
  s := 'INSERT INTO molinos (nid, nid_curvaaero, nidcentral, etha, huso_utm, zona_utm, xe, yn, cota ) '
    + ' VALUES ( ' + IntToStr(nid) + ', ' + IntToStr(nid_curvaaero) + ', ' + IntToStr(
    nid_central) + ', ' + FloatToStr(etha) + ', ' + IntToStr(
    posicion.utm_huso) + ', ''' + posicion.UTM_Zona + ''', ' + FloatToStr(
    posicion.UTM_xE) + ', ' + FloatToStr(posicion.UTM_yN) + ', ' + FloatToStr(
    posicion.cota);

  DB.Exec(s);
  Result := nid;
end;

procedure TMolino.UpdateDB(db: TDBPQCon);
var
  s: string;
begin
  s := 'UPDATE molinos SET ' + 'nid_curvaaero=' + IntToStr(nid_curvaaero) + ', ' +
    'nidcentral =' + IntToStr(nid_central) + ', ' + 'etha = ' + FloatToStr(
    etha) + ', ' + 'huso_utm = ' + IntToStr(posicion.utm_huso) + ', ' +
    'zona_utm = ''' + posicion.UTM_Zona + ''', ' + 'xe = ' + FloatToStr(
    posicion.UTM_xE) + ', ' + 'yn = ' + FloatToStr(posicion.UTM_yN) + ', ' +
    'cota = ' + FloatToStr(posicion.cota) + ' WHERE nid = ' + IntToStr(nid);
  DB.Exec(s);

end;

procedure TMolino.UpdateDB_Modelo(db: TDBPQCon);
var
  s: string;
begin
  s := 'UPDATE molino SET etha= ''' + FloatToStr(etha) + ''' WHERE nid =' +
    IntToStr(nid) + ';';
  DB.exec(s);
end;
{$ENDIF}


// Este procedimiento calcula la potencia electrica del molino usando
// la curva del mismo. Para la potencia del viento incidente toma el promedio
// de las potencia de viento de los puntos de la grilla rotada asociados al molino.
// el procedimiento afecta luego del calculo los valores de PV en la grilla rotada
// de forma que reflejen la potencia eléctrica extraída.x
// Pasamos la dirección del viento (ang) para que pueda calcular el valor de ro
procedure TMolino.CalcularAFectacionEnGrillaRotada(kfila_: integer;
  fila: TVectR; ang, fpv3pe, relRed, roDato: NReal);

var
  pv_E: NReal;
  j: integer;
  rel: NReal;
  ajuste_ro: NReal;
begin
  if kFila_ <> self.posicion.kFila then
    exit;

  ajuste_ro := Ajuste_Cp_ro(roDato);
  //JFP: Ajusta la curva del aero según la densidad dato (roDato),
  //ya que la misma fue cargada para la densidad base (ro_BASE).

  pv_prom := 0;
  for j := posicion.jCol1 to posicion.jCol2 do
    pv_prom := pv_prom + fila.e(j);

  assert( posicion.jCol2 >= posicion.jCol1, 'Error, posicion.jCol2 < posicion.jCol1 ');
  pv_prom := pv_prom / (posicion.jCol2 - posicion.jCol1 + 1);

  self.PE_MW := curva.P_MW(pv_prom * etha) * ajuste_ro * relRed;

  pv_E := PE_MW / curva.AreaRotor;

  if pv_prom < 1E-15 then
    rel := 0.01
  else
    rel := (pv_E * fpv3pe) / pv_prom;
  rel := 1 - rel;
  rel := max(0.00001, min(0.999999, rel));

  for j := posicion.jCol1 to posicion.jCol2 do
    fila.pon_e(j, fila.e(j) * rel);
end;



procedure TMolino.Free;
begin
  curva.Free;
  inherited Free;
end;



constructor TParque_ModeloTeorico.Create_Init(aNombre: string;
  aCantMolinos, aN_Sectores_Dir: integer; modeloAero, archi_curvaAero: string;
  //afectacion_v3: TAnemometro;
  alfa_terreno, gamma_atmosfera, fpv3pe: NReal;

  // rectangulo en coordenadas cartesianas  que incluye TODOS los AEROS
  // con el ANEMO en el (0,0). Esto es para una defición "artificial" junto con los
  // parámetros npuntos_* siguientes
  Aero_xmin_, Aero_xmax_, Aero_ymin_, Aero_ymax_: NReal;
  // Posiciones imaginaria de los aeros para una inicialización
  // artificial del Modelo plantando los aeros en una grilla de npuntos_x por n_puntos_y
  // estos parámetros no tienen utilidad real dado que se conoce la posición de los aeros
  npuntos_x, npuntos_y: integer;

  grilla_Rotada_dx_, grilla_Rotada_dy_, grilla_Rotada_Margen_Dx_: NReal;
  // resoluciòn de la grilla rotada
  Factor_Perdidas, Potencia_Nominal, Potencia_autorizada_MW: NReal);

var
  Molino: TMolino;
  i: integer;
  kFil, jCol: integer;
  dx, dy: NReal;

begin
  Nombre := aNombre;
  CantMolinos := aCantMolinos;
  N_Sectores_Dir := aN_Sectores_Dir;
  //inicializo por defecto todos los molinos al mismo tipo
  grilla_Rotada_dx := grilla_Rotada_dx_;
  grilla_Rotada_dy := grilla_Rotada_dy_;
  grilla_Rotada_Margen_Dx := grilla_Rotada_Margen_Dx_;

  self.alfa_terreno := alfa_terreno;
  self.gamma_atmosfera := gamma_atmosfera;
  self.fpv3pe := fpv3pe;

  Self.Factor_Perdidas := Factor_Perdidas;
  Self.Potencia_Nominal := Potencia_Nominal;
  Self.Potencia_autorizada_MW := Potencia_autorizada_MW;

  dx := (Aero_xMax_ - Aero_xmin_) / (NPuntos_x - 1);
  dy := (Aero_ymax_ - aero_ymin_) / (NPuntos_y - 1);

  setlength(molinos, CantMolinos);

  kFil := 0;
  jCol := 0;
  for i := 0 to CantMolinos - 1 do
  begin
    Molino :=
      TMolino.Create_Init(Aero_xmin_ + jCol * dx, Aero_ymin_ + kFil *
      dy, modeloAero, archi_curvaAero, N_Sectores_Dir);

    Molino.etha := 1;
    self.Molinos[i] := Molino;

    Inc(jCol);
    if jCol = npuntos_x then
    begin
      jCol := 0;
      Inc(kFil);
    end;

  end;
  Meteo_etha := TVectR.Create_init(11);
  Meteo_etha.Ceros;
  Meteo_etha.pon_e(1, 1.0);

  grillaRotada := nil;
end;



procedure TParque_ModeloTeorico.Calc_Auxiliares;
var
  aMolino: TMolino;
  k: integer;
  aofc: array of TCurvaAerogen;
begin
  Potencia_autorizada_MW := max(Potencia_autorizada_MW, 1.0e-8);
  k_Perdidas := Factor_Perdidas / Potencia_Nominal; // * (1 - Factor_Perdidas);

  (* cálcula la curva_s suma de todos los molinos *);
  if length( Molinos ) > 0 then
  begin
    aMolino:= molinos[0];
    curva_s:= TCurvaAerogen.Create_Clone( aMolino.curva );
    if length( Molinos ) > 1 then
    begin
      setlength( aofc, length( Molinos ) - 1 );
      for k:= 1 to high( molinos ) do
        aofc[k-1]:= molinos[k].curva;
      curva_s.SumarCurvas( aofc );
      setlength( aofc, 0 );
    end;
  end;
end;


constructor TParque_ModeloTeorico.Create_LoadFromFile(var f: textfile);
var
  i: integer;
  tipoA: string;
  version: integer;
  r: string;
  Meteo_etha_nr: NReal;
begin
  inherited Create;
  curva_S:= nil;

  setSeparadoresGlobales;

  readln(f, r);

  Delete(r, 1, length('VERSION='));
  version := StrToInt(r);

  readln(f, nid_Central );
  readln(f, nid_modelo );

  readln(f, Nombre);
  readln(f, CantMolinos);
  readln(f, N_Sectores_Dir);

  readln(f, alfa_terreno);
  readln(f, gamma_atmosfera);
  readln(f, fpv3pe);

  readln(f, Factor_Perdidas);
  readln(f, Potencia_Autorizada_MW);
  if version <= 1 then
    Potencia_Nominal := Potencia_Autorizada_MW
  else
    readln(f, Potencia_Nominal);

  if version < 3 then
  begin
    readln(f, Meteo_etha_nr);
    Meteo_etha := TVectR.Create_init(11);
    Meteo_etha.Ceros;
    Meteo_etha.pon_e(1, Meteo_etha_nr);
  end
  else
    Meteo_etha := TVectR.CreateLoadFromFile(f);

  readln(f, grilla_Rotada_dx);
  readln(f, grilla_Rotada_dy);
  readln(f, grilla_Rotada_Margen_Dx);

  setlength(molinos, CantMolinos);

  for i := 0 to CantMolinos - 1 do
    Molinos[i] := TMolino.Create_LoadFromFile(f);

  readln(f, Meteo_UTM_Zona);
  readln(f, Meteo_UTM_huso);
  readln(f, Meteo_UTM_xe);
  readln(f, Meteo_UTM_yn);

  setSeparadoresLocales;

  Calc_Auxiliares;

end;

constructor TParque_ModeloTeorico.Create_LoadFromArchi(const archi: string);
var
  f: textfile;
begin
  inherited Create;
  curva_s:= nil;
  assignfile(f, archi);
  reset(f);
  Create_LoadFromFile(f);
  closefile(f);
end;


procedure TParque_ModeloTeorico.StoreInFile(var f: textfile);
var
  i: integer;
begin
  setSeparadoresGlobales;

  writeln(f, 'VERSION=' + IntToStr(VERSION_ARCHI_PARQUE));
  writeln(f, nid_central );
  writeln( f, nid_modelo );
  writeln(f, Nombre);
  writeln(f, CantMolinos);
  writeln(f, N_Sectores_Dir);

  writeln(f, alfa_terreno);
  writeln(f, gamma_atmosfera);
  writeln(f, fpv3pe);
  writeln(f, Factor_Perdidas);
  writeln(f, Potencia_Autorizada_MW);
  writeln(f, Potencia_Nominal);
  Meteo_etha.StoreInFile(f);

  writeln(f, grilla_Rotada_dx);
  writeln(f, grilla_Rotada_dy);
  writeln(f, grilla_Rotada_Margen_Dx);

  for i := 0 to CantMolinos - 1 do
    Molinos[i].StoreInFile(f);

  writeln(f, Meteo_UTM_Zona);
  writeln(f, Meteo_UTM_huso);
  writeln(f, Meteo_UTM_xe);
  writeln(f, Meteo_UTM_yn);
  setSeparadoresLocales;
end;

procedure TParque_ModeloTeorico.StoreInArchi(const archi: string);
var
  f: textfile;
begin
  assignfile(f, archi);
  rewrite(f);
  storeInFile(f);
  closefile(f);
end;

{$IFNDEF SIN_DBCON}
constructor TParque_ModeloTeorico.Create_LoadFromDB(db: TDBPQCon;
  r_central: TSQLQuery);
var
  i: integer;
  tipoA: string;
  dsModelo, dsMolinos: TSQLQuery;

  flg_creando_modelo: boolean;
  s: string;
begin
  inherited Create;
  setSeparadoresGlobales;
  curva_s:= nil;

  nid_Central := r_central.FieldByName('nid').AsInteger;
  Nombre := r_central.FieldByName('centralgeneradora').AsString;

  dsModelo := DB.query('SELECT * FROM modelo_teorico WHERE central = ' + IntToStr(
    nid_Central ) + ' ORDER BY dt_creacion DESC LIMIT 1');


  if dsModelo.EOF then
  begin
    nid_modelo := DB.nextval('modelo_teorico');
    N_Sectores_Dir := 72;
    alfa_terreno := 0.5; // alfa_terreno  Mezclador horizontal
    gamma_atmosfera := 0.4; // gamma_atmosfera  Mezclador Vertical
    fpv3pe := 1.8;
    Factor_Perdidas := 0.035;
    potencia_autorizada_mw := r_central.FieldByName('potencia_autorizada_mw').AsFloat;
    Potencia_Nominal := potencia_autorizada_mw;
    Meteo_etha := TVectR.Create_init(11);
    Meteo_etha.Ceros;
    Meteo_etha.pon_e(1, 1.0);
    grilla_Rotada_dx := 9;
    grilla_Rotada_dy := 1;
    grilla_Rotada_Margen_dx := 1000;
    Meteo_UTM_Zona := r_central.FieldByName('meteo_utm_zona').AsString[1];
    Meteo_UTM_huso := r_central.FieldByName('meteo_utm_huso').AsInteger;
    Meteo_UTM_xe := r_central.FieldByName('Meteo_UTM_xe').AsFloat;
    Meteo_UTM_yn := r_central.FieldByName('meteo_utm_yn').AsFloat;
    Meteo_cota := r_central.FieldByName('meteo_cota').AsFloat;
    flg_creando_modelo := True;
    DB.exec('INSERT INTO modelo_teorico (nid, central, dt_creacion, dt_actualizacion ) '
      + ' VALUES ( ' + IntToStr(nid_modelo) + ', ' + IntToStr(nid_central) + ', now(), now() )');
  end
  else
  begin
    nid_modelo := dsModelo.FieldByName('nid').AsInteger;
    flg_creando_modelo := False;
    N_Sectores_Dir := dsModelo.FieldByName('n_sectores_dir').AsInteger;
    alfa_terreno := dsModelo.FieldByName('alfa_terreno').AsFloat;
    gamma_atmosfera := dsModelo.FieldByName('gamma_atmosfera').AsFloat;
    fpv3pe := dsModelo.FieldByName('fpv3pe').AsFloat;
    Factor_Perdidas := dsModelo.FieldByName('factor_perdidas').AsFloat;
    potencia_autorizada_mw := dsModelo.FieldByName('potencia_autorizada_mw').AsFloat;
    Potencia_Nominal := dsModelo.FieldByName('potencia_nominal').AsFloat;
    s := dsModelo.FieldByName('meteo_etha').AsString;
    if pos('|', s) = 0 then
    begin
      Meteo_etha := TVectR.Create_init(11);
      Meteo_etha.Ceros;
      Meteo_etha.pon_e(1, StrToFloat(s));
    end
    else
      Meteo_etha := TVectR.Create_unserialize(s);
    grilla_Rotada_dx := dsModelo.FieldByName('grilla_rotada_dx').AsFloat;
    grilla_Rotada_dy := dsModelo.FieldByName('grilla_rotada_dy').AsFloat;
    grilla_Rotada_Margen_Dx := dsModelo.FieldByName('grilla_rotada_margen_dx').AsFloat;
    Meteo_UTM_Zona := dsModelo.FieldByName('meteo_utm_zona').AsString[1];
    Meteo_UTM_huso := dsModelo.FieldByName('meteo_utm_huso').AsInteger;
    Meteo_UTM_xe := dsModelo.FieldByName('meteo_utm_xe').AsFloat;
    Meteo_UTM_yn := dsModelo.FieldByName('meteo_utm_yn').AsFloat;
    Meteo_Cota := dsModelo.FieldByName('meteo_cota').AsFloat;
  end;
  dsModelo.Free;

  CantMolinos := DB.FuncI('SELECT count(*) FROM molinos WHERE nidcentral=' +
    IntToStr(nid_central));
  setlength(molinos, CantMolinos);
  dsMolinos := DB.query('SELECT * FROM molinos WHERE nidcentral=' + IntToStr(nid_central));

  for i := 0 to CantMolinos - 1 do
  begin
    Molinos[i] := TMolino.CreateLoadFromDB(db, dsMolinos);
    dsMolinos.Next;
  end;

  if flg_creando_modelo then
    for i := 0 to CantMolinos - 1 do
      db.exec('INSERT INTO modelo_teorico_etha_molinos ( nid_modelo, nid_molino, etha ) VALUES ( '
        + IntToStr(nid_modelo) + ', ' + IntToStr(Molinos[i].nid) + ', 1 )');

  for i := 0 to CantMolinos - 1 do
    Molinos[i].etha := DB.funcF(
      'SELECT etha FROM modelo_teorico_etha_molinos WHERE nid_modelo = ' +
      IntToStr(nid_modelo) + ' AND nid_molino = ' + IntToStr(Molinos[i].nid));

  dsMolinos.Free;

  RecalcularCoordenadas;
  if flg_creando_modelo then
    Update(db);

  Calc_Auxiliares;
  setSeparadoresLocales;
end;
constructor TParque_ModeloTeorico.CreateMultiParque_LoadFromDB(db: TDBPQCon;
  nid_centrales: TDAofNInt;nid_central_anemo:integer);

var
  i,h,nidCentral,CantMolinosTotal,j: integer;
  tipoA: string;

  r_central,dsModelo, dsMolinos: TSQLQuery;

  flg_creando_modelo,es: boolean;
  s: string;

begin
  inherited Create;
  setSeparadoresGlobales;
  curva_s:= nil;


 Nombre:='Superparque';
 potencia_autorizada_mw:=0;
 Potencia_Nominal:=0;
 nid_Central := 0;

 for h:=0 to length(nid_centrales)-1 do begin

 nid_Central:= nid_Central*100 + nid_centrales[h];  end;

  dsModelo := DB.query('SELECT * FROM modelo_teorico WHERE central = ' + IntToStr(
    nid_Central ) + ' ORDER BY dt_creacion DESC LIMIT 1');


  if dsModelo.EOF then
  begin

    flg_creando_modelo := True;

    for h:=0 to length(nid_centrales)-1 do begin

    r_central := db.query('SELECT * FROM modelo_teorico WHERE central = ' + IntToStr(nid_centrales[h]));
    es:=(nid_centrales[h]=nid_central_anemo);

    if h=0 then begin

    nid_modelo := DB.nextval('modelo_teorico');
    N_Sectores_Dir := 72;
    alfa_terreno := 0.5; // alfa_terreno  Mezclador horizontal
    gamma_atmosfera := 0.4; // gamma_atmosfera  Mezclador Vertical
    fpv3pe := 1.8;
    Factor_Perdidas := 0.035;
    end;

    if es then
    begin
    Meteo_etha := TVectR.Create_init(11);
    Meteo_etha.Ceros;
    Meteo_etha.pon_e(1, 1.0);
    Meteo_UTM_Zona := r_central.FieldByName('meteo_utm_zona').AsString[1];
    Meteo_UTM_huso := r_central.FieldByName('meteo_utm_huso').AsInteger;
    Meteo_UTM_xe := r_central.FieldByName('Meteo_UTM_xe').AsFloat;
    Meteo_UTM_yn := r_central.FieldByName('meteo_utm_yn').AsFloat;
    Meteo_cota := r_central.FieldByName('meteo_cota').AsFloat;
    grilla_Rotada_dx := 9;
    grilla_Rotada_dy := 1;
    grilla_Rotada_Margen_dx := 1000;

    end;

    Nombre := Nombre +'_'+ r_central.FieldByName('central').AsString ;
    potencia_autorizada_mw := potencia_autorizada_mw + r_central.FieldByName('potencia_autorizada_mw').AsFloat;
    Potencia_Nominal := Potencia_Nominal + r_central.FieldByName('potencia_nominal').AsFloat;
    end;


    DB.exec('INSERT INTO modelo_teorico (nid, central, dt_creacion, dt_actualizacion ) '
      + ' VALUES ( ' + IntToStr(nid_modelo) + ', ' + IntToStr(nid_central) + ', now(), now() )');
  end
  else
  begin

    flg_creando_modelo := False;
    nid_modelo := dsModelo.FieldByName('nid').AsInteger;

    N_Sectores_Dir := dsModelo.FieldByName('n_sectores_dir').AsInteger;
    alfa_terreno := dsModelo.FieldByName('alfa_terreno').AsFloat;
    gamma_atmosfera := dsModelo.FieldByName('gamma_atmosfera').AsFloat;
    fpv3pe := dsModelo.FieldByName('fpv3pe').AsFloat;
    Factor_Perdidas := dsModelo.FieldByName('factor_perdidas').AsFloat;


    s := dsModelo.FieldByName('meteo_etha').AsString;
    if pos('|', s) = 0 then
    begin
      Meteo_etha := TVectR.Create_init(11);
      Meteo_etha.Ceros;
      Meteo_etha.pon_e(1, StrToFloat(s));
    end
    else
        Meteo_etha := TVectR.Create_unserialize(s);

    Meteo_UTM_Zona := dsModelo.FieldByName('meteo_utm_zona').AsString[1];
    Meteo_UTM_huso := dsModelo.FieldByName('meteo_utm_huso').AsInteger;
    Meteo_UTM_xe := dsModelo.FieldByName('meteo_utm_xe').AsFloat;
    Meteo_UTM_yn := dsModelo.FieldByName('meteo_utm_yn').AsFloat;
    Meteo_Cota := dsModelo.FieldByName('meteo_cota').AsFloat;
    grilla_Rotada_dx := dsModelo.FieldByName('grilla_rotada_dx').AsFloat;
    grilla_Rotada_dy := dsModelo.FieldByName('grilla_rotada_dy').AsFloat;
    grilla_Rotada_Margen_Dx := dsModelo.FieldByName('grilla_rotada_margen_dx').AsFloat;

    potencia_autorizada_mw := dsModelo.FieldByName('potencia_autorizada_mw').AsFloat;
    Potencia_Nominal := dsModelo.FieldByName('potencia_nominal').AsFloat;

  end;
  dsModelo.Free;
 //////////////////////////////////////////////////////////////////////////////

  CantMolinosTotal:=0;

  for h:=0 to length(nid_centrales)-1 do
  begin
  CantMolinos := DB.FuncI('SELECT count(*) FROM molinos WHERE nidcentral=' +
    IntToStr(nid_centrales[h]));

  CantMolinosTotal:=CantMolinosTotal+CantMolinos;
  end;

  setlength(molinos, CantMolinosTotal);
  j:=0;

  for h:=0 to length(nid_centrales)-1 do
  begin
  dsMolinos := DB.query('SELECT * FROM molinos WHERE nidcentral=' + IntToStr(nid_centrales[h]));
  CantMolinos := DB.FuncI('SELECT count(*) FROM molinos WHERE nidcentral=' +
    IntToStr(nid_centrales[h]));

  for i:=0 to CantMolinos-1 do
  begin
    Molinos[j+i] := TMolino.CreateLoadFromDB(db, dsMolinos);
    dsMolinos.Next;
  end;
  j:=j+CantMolinos;
  end;

  if flg_creando_modelo then
         for i := 0 to CantMolinosTotal - 1 do
          db.exec('INSERT INTO modelo_teorico_etha_molinos ( nid_modelo, nid_molino, etha ) VALUES ( '
          + IntToStr(nid_modelo) + ', ' + IntToStr(Molinos[i].nid) + ', 1 )');



  for i := 0 to CantMolinosTotal - 1 do
    Molinos[i].etha := DB.funcF(
      'SELECT etha FROM modelo_teorico_etha_molinos WHERE nid_modelo = ' +
      IntToStr(nid_modelo) + ' AND nid_molino = ' + IntToStr(Molinos[i].nid));

  CantMolinos:=CantMolinosTotal;

  dsMolinos.Free;

  RecalcularCoordenadas;

  if flg_creando_modelo then
    Update(db);

  Calc_Auxiliares;
  setSeparadoresLocales;


 end;




procedure TParque_ModeloTeorico.Update(db: TDBPQCon);
var
  i: integer;
  tipoA: string;
  dsModelo, dsMolinos: TSQLQuery;
  aMolino: TMolino;
  s: string;

begin
  for i := 0 to CantMolinos - 1 do
  begin
    aMolino := molinos[i];
    s := 'UPDATE modelo_teorico_etha_molinos SET etha = ' + FloatToStr(
      aMolino.etha) + ' WHERE nid_modelo = ' + IntToStr(
      nid_modelo) + ' AND nid_molino = ' + IntToStr(aMolino.nid);
    DB.exec(s);
  end;

  s := 'UPDATE modelo_teorico SET ' + 'n_sectores_dir= ' + IntToStr(
    n_sectores_dir) + ', alfa_terreno = ' + FloatToStr(
    alfa_terreno) + ', gamma_atmosfera = ' + FloatToStr(
    gamma_atmosfera) + ', fpv3pe = ' + FloatToStr(fpv3pe) + ', factor_perdidas = ' +
    FloatToStr(factor_perdidas) + ', potencia_autorizada_mw =' +
    FloatToStr(potencia_autorizada_mw) + ', potencia_nominal = ' +
    FloatToStr(potencia_nominal) + ', meteo_etha = ''' +
    meteo_etha.serialize + ''', meteo_cota = ' + FloatToStr(
    meteo_cota) + ', grilla_rotada_dx = ' + FloatToStr(
    grilla_rotada_dx) + ', grilla_rotada_dy = ' + FloatToStr(
    grilla_rotada_dy) + ', grilla_rotada_margen_dx = ' + FloatToStr(
    grilla_rotada_margen_dx) + ', meteo_utm_zona = ''' + meteo_utm_zona + ''''
    + ', meteo_utm_huso = ' + FloatToStr(meteo_utm_huso) + ', meteo_utm_xe = ' +
    FLoatToStr(meteo_utm_xe) + ', meteo_utm_yn = ' + FloatToStr(
    meteo_utm_yn) + ' WHERE nid = ' + IntToStr(nid_modelo);
  DB.exec(s);
end;

{$ENDIF}


procedure TParque_ModeloTeorico.Free;
var
  k: integer;
begin

  for k := 0 to CantMolinos - 1 do
  begin
    Molinos[k].Free;
  end;

  setlength(Molinos, 0);

  if grillaRotada <> nil then
    grillaROtada.Free;

  if Curva_s <> nil then
    Curva_s.Free;
  inherited Free;
end;


procedure TParque_ModeloTeorico.SetCurvaAero(kAero: integer; aCurva: TCurvaAerogen);
begin
  Molinos[kAero].curva := aCurva;
end;




procedure TParque_ModeloTeorico.Calcular_pv_fila(filaACalcular, filaAnterior: TVectR);
var
  j: integer;
  m: NReal;
begin
  for j := 2 to filaACalcular.n - 1 do
  begin
    m:=  (filaAnterior.e(j) * alfa_central + (filaAnterior.e(j - 1) + filaAnterior.e(j + 1)) *
      alfa_lateral) * gamma_central + pv_Libre * gamma_sup;
    filaACalcular.pon_e(j, m);
  end;
end;

function TParque_ModeloTeorico.ParametrosPtr: TDAOfNRealPtr;
var
  res: TDAOfNRealPtr;
  k, kMolino: integer;
  nMolinos: integer;
  cnt_par: integer;

begin
  nMolinos := length(self.Molinos);

  setlength(res, Meteo_etha.n + 1 ); // + nMolinos

  cnt_par := 0;

  // Meteo_etha
  for k := 1 to Meteo_etha.n do
  begin
    res[cnt_par] := @Meteo_etha.pv[k];
    Inc(cnt_par);
  end;
  (*
  res[cnt_par] := @alfa_terreno;
  Inc(cnt_par);
  res[cnt_par] := @gamma_atmosfera;
  Inc(cnt_par);
  res[cnt_par] := @fpv3pe;
  Inc(cnt_par);
    *)
  res[cnt_par] := @Factor_Perdidas;
  Inc(cnt_par);
(*
  for kMolino := 0 to nMolinos - 1 do
  begin
    res[cnt_par] := @Molinos[kMolino].etha;
    Inc(cnt_par);
  end;
  *)
  Result := res;
end;


procedure TParque_ModeloTeorico.AerosAfectarFila(kFila: integer;
  ang, relRed, roDato: NReal);
var
  kMolino: integer;
begin
  for kMolino := 0 to high(molinos) do
    molinos[kmolino].CalcularAFectacionEnGrillaRotada(
      kFila, GrillaRotada.Fila(kFila), ang, fpv3pe, relRed, roDato);

end;

procedure TParque_ModeloTeorico.SetMeteo_UTM(huso: integer; Zona: char; xE, yN: Nreal);
begin
  Meteo_UTM_huso := huso;
  Meteo_UTM_Zona := Zona;
  Meteo_UTM_xe := xE;
  Meteo_UTM_yN := yN;
end;

procedure TParque_ModeloTeorico.SetAero_UTM(kMolino: integer; uso: integer;
  Zona: char; xE, yN: Nreal);
begin
  molinos[kMolino - 1].posicion.Set_UTM(uso, zona, xE, yN);
end;

procedure TParque_ModeloTeorico.RecalcularCoordenadas;
var
  kMolino: integer;
  aMolino: TMolino;
begin
  for kMolino := 0 to high(molinos) do
  begin
    aMolino := molinos[kMolino];
    aMolino.posicion.z_base :=
      cinit(aMolino.posicion.UTM_xE - Meteo_UTM_xe, aMolino.posicion.UTM_yN -
      Meteo_UTM_yN);
  end;
end;


procedure TParque_ModeloTeorico.SetPixFrame(xp1, yp1, xp2, yp2: integer);
var
  aP: complex;
  kMolino: integer;
  dd: NReal;
begin
  self.xp1 := xp1;
  self.yp1 := yp1;
  self.xp2 := xp2;
  self.yp2 := yp2;

  dxp := xp2 - xp1;
  dyp := yp2 - yp1;

  Show_xmin := 0;
  Show_xmax := 0;
  Show_ymin := 0;
  Show_ymax := 0;

  for kMolino := 0 to high(molinos) do
  begin
    aP := molinos[kmolino].posicion.z_base;
    if aP.re < Show_xmin then
      show_xmin := aP.re
    else if aP.re > show_xmax then
      show_xmax := aP.Re;

    if ap.im < show_ymin then
      show_ymin := ap.im
    else if ap.im > show_ymax then
      show_ymax := ap.im;
  end;


  show_xmin := show_xmin - show_padding_;
  show_xmax := show_xmax + show_padding_;
  dxa := (show_xmax - show_xmin);

  show_ymin := show_ymin - show_padding_;
  show_ymax := show_ymax + show_padding_;
  dya := (show_ymax - show_ymin);


  if (dxp / dxa) > (dyp / dya) then
  begin
    // hay más píxeles/m en la horizontal  thengo que aumentar el ancho en m
    // para lograr la misma definición que la vertical
    dd := (-dxa + dya / dyp * dxp) / 2;
    show_xmin := show_xmin - dd;
    show_xmax := show_xmax + dd;
    dxa := (show_xmax - show_xmin);
  end
  else
  begin
    // hay más píxeles/m en la vertical
    dd := (-dya + dxa / dxp * dyp) / 2;
    show_ymin := show_ymin - dd;
    show_ymax := show_ymax + dd;
    dya := (show_ymax - show_ymin);
  end;

end;

procedure TParque_ModeloTeorico.ZToPix(var xp, yp: integer; z: Complex);
begin
  xp := xp1 + round((dxp / dxa) * (z.re - show_xmin));
  yp := yp2 - round((dyp / dya) * (z.im - show_ymin));
end;

procedure TParque_ModeloTeorico.PixToZ(var z: Complex; xp, yp: integer);
begin
  z.re := show_xmin + (dxa / dxp) * (xp - xp1);
  z.im := show_ymin + (dya / dyp) * (yp2 - yp);
end;


// busca el molino que esté más proximo al punto xp, yp y lo devuelve
// si el mismo está a menos de un diametro del punto xp, yp
function TParque_ModeloTeorico.MolinoAt_xpyp(xp, yp: integer): TMolino;
var
  res, aMolino: TMolino;
  kMolino: integer;
  d, min_d: NReal;
  z: Complex;

begin
  PixToZ(z, xp, yp);
  aMolino := Molinos[0];
  d := cmod(z - aMolino.posicion.z_base);
  min_d := d;
  res := aMolino;
  for kMolino := 1 to high(Molinos) do
  begin
    aMolino := Molinos[kMolino];
    d := cmod(z - aMolino.posicion.z_base);
    if d < min_d then
    begin
      min_d := d;
      res := aMolino;
    end;
  end;
  if min_d <= (3 * res.curva.Diametro) then
    Result := res
  else
    Result := nil;
end;


function TParque_ModeloTeorico.GridValAtZ(z: complex): NReal;
var
  zRot: complex;
  kr, jr: NReal;
  k1, k2, j1, j2: integer;
  alfa_k, alfa_j, beta_k, beta_j: NReal;
begin
  zRot := z * self.z_BaseToRot;
  if (zRot.re < Caja_xmin) or (zRot.re > Caja_xmax) or (zRot.im < Caja_ymin) or
    (zRot.im > Caja_ymax) then
  begin
    Result := 0;
    exit;
  end;
  kr := (zRot.im - Caja_ymin) / Caja_dy;
  jr := (zRot.re - Caja_xmin) / Caja_dx;
  k1 := trunc(kr) + 1;
  k2 := k1 + 1;
  if k2 > grillaRotada.nf then
    k2 := grillaRotada.nf;
  j1 := trunc(jr) + 1;
  j2 := j1 + 1;
  if j2 > grillaRotada.nc then
    j2 := grillaRotada.nc;
  alfa_k := frac(kr);
  alfa_j := frac(jr);
  beta_k := 1 - alfa_k;
  beta_j := 1 - alfa_j;


  Result := beta_k * (beta_j * grillaRotada.e(k1, j1) + alfa_j *
    grillaRotada.e(k1, j2)) + alfa_k * (beta_j * grillaRotada.e(k2, j1) +
    alfa_j * grillaRotada.e(k2, j2));

end;

function TParque_ModeloTeorico.GridValAtPix(xp, yp: integer): NReal;
var
  z: complex;
begin
  PixToZ(z, xp, yp);
  Result := GridValAtZ(z);
end;


procedure TParque_ModeloTeorico.Calc_PotMW_init_(ang: NReal);
var
  kMolino: integer;
  aMolino: TMolino;
  ndivx, ndivy: integer;
  k: integer;
  NArmonicos: integer;

begin

  Calc_Auxiliares;

  Meteo_etha_dir := max( 0.1, Meteo_etha.FourierEval(ang / 180 * pi) );

  z_BaseToRot := cinit(cos(2 * pi * (ang + 180) / 360),
    sin(2 * pi * (ang + 180) / 360));
  Caja_xmin := 0;
  Caja_ymin := 0;
  Caja_xmax := 0;
  Caja_ymax := 0;

  for kMolino := 0 to high(Molinos) do
  begin
    aMolino := Molinos[kMolino];
    aMolino.posicion.z_rotado := aMolino.posicion.z_base * z_BaseToRot;
    if aMolino.posicion.z_rotado.re < Caja_xmin then
      Caja_xmin := aMolino.posicion.z_rotado.re
    else if aMolino.posicion.z_rotado.re > Caja_xmax then
      Caja_xmax := aMolino.posicion.z_rotado.re;
    if aMolino.posicion.z_rotado.im < Caja_ymin then
      Caja_ymin := aMolino.posicion.z_rotado.im
    else if aMolino.posicion.z_rotado.im > Caja_ymax then
      Caja_ymax := aMolino.posicion.z_rotado.im;
  end;

  Caja_xmin := Caja_xmin - grilla_Rotada_Margen_Dx;
  Caja_xmax := Caja_xmax + grilla_Rotada_Margen_Dx;

  Caja_ymin := Caja_ymin - grilla_Rotada_Margen_Dx / 2;
  Caja_ymax := Caja_ymax + grilla_Rotada_Margen_Dx / 2;

  ndivx := trunc((Caja_xmax - Caja_xmin) / grilla_Rotada_dx + 0.999);
  Caja_dx := (Caja_xmax - Caja_xmin) / ndivx;

  ndivy := trunc((Caja_ymax - Caja_ymin) / grilla_Rotada_dy + 0.999);
  Caja_dy := (Caja_ymax - Caja_ymin) / ndivy;

  // Alfa es el peso de los laterales y beta = 1- 2* alfa el peso del central.
  // Al aumentar la dy en relación a dx tiene que terminar pensando todos por igual alfa = 1/3
  alfa_lateral := (1 / 2) * (1 - exp(-alfa_terreno * Caja_dy / Caja_dx));
  alfa_central := 1 - 2 * alfa_lateral;

  //gamma_sup es el peso de la Velocidad ARRIBA (a 500 m se supone = pvLIBRE) y gamma_inf es a
  gamma_sup := (1 - exp(-gamma_atmosfera * Caja_dy / 500.0));
  gamma_central := 1 - gamma_sup;

  assert( alfa_lateral >= 0, 'Error, alfa_lateral < 0' );
  assert( alfa_lateral <= 0.5, 'Error, alfa_lateral > 0.5' );
  assert( alfa_central >= 0, 'Error, alfa_central < 0' );
  assert( alfa_central <= 1, 'Error, alfa_lateral > 1' );
  assert( gamma_sup > 0, 'Error, gamma_sup < 0' );
  assert( gamma_sup < 1, 'Error, gamma_sup > 1' );
  assert( gamma_central > 0, 'Error, central_sup < 0' );
  assert( gamma_central < 1, 'Error, central_sup > 1' );

  if GrillaRotada <> nil then
    grillaRotada.Free;

  grillaRotada := TMatR.Create_Init(ndivy + 1, ndivx + 1);

  for kMolino := 0 to high(Molinos) do
  begin
    aMolino := Molinos[kMolino];
    aMolino.posicion.kFila :=
      round((aMolino.posicion.z_rotado.im - Caja_ymin) / Caja_dy) + 1;
    aMolino.posicion.jCol1 :=
      round((aMolino.posicion.z_rotado.re - aMolino.curva.Diametro / 2 - Caja_xmin) /
      Caja_dx) + 1;
    aMolino.posicion.jCol2 :=
      round((aMolino.posicion.z_rotado.re + aMolino.curva.Diametro / 2 - Caja_xmin) /
      Caja_dx) + 1;
  end;


end;


function TParque_ModeloTeorico.Calc_PotMW_pvLibre(roDato, pv3, cgm: NReal ): NReal;
var
  kMolino: integer;
  k, j: integer;
  res: NReal;
  relRed: NReal;
  cnt_Iter_relRed: integer;
  LimitePot: NReal;
label
  lbl_Recalcular;

begin
  LimitePot:= min( cgm, Potencia_Autorizada_MW );

  pv_Libre := pv3;

  // cálculo de los pesos de afectación.

  relRed := 1.0;
  cnt_Iter_relRed := 0;

  lbl_Recalcular:

      // Inicialización de las condiciones de borde.
    for j := 1 to grillaRotada.nc do
      GrillaRotada.pon_e(1, j, pv3);

  for k := 2 to grillaRotada.nf do
  begin
    GrillaRotada.pon_e(k, 1, pv3);
    GrillaRotada.pon_e(k, grillaRotada.nc, pv3);
  end;

  self.AerosAfectarFila(1, ang, relRed, roDato);

  for k := 2 to grillaRotada.nf do
  begin
    Calcular_pv_fila(grillaRotada.fila(k), grillaRotada.fila(k - 1));
    self.AerosAfectarFila(k, ang, relRed, roDato);
  end;

  //  GrillaRotada.WriteArchiXLT( 'c:\basura\interfer_pv3.xlt' );
  res := 0;

  for kMolino := 0 to high(Molinos) do
    res := res + molinos[kMolino].PE_MW;

  res := max(0, res - k_Perdidas * sqr(res));
  if res > LimitePot then
  begin
    relred := LimitePot / res * relRed;
    if cnt_Iter_relRed < 10 then
    begin
      Inc(cnt_Iter_RelRed);
      goto lbl_Recalcular;
    end;

    for kMolino := 0 to high(Molinos) do
      molinos[kMolino].PE_MW := molinos[kMolino].PE_MW * relRed;
    res := LimitePot;
  end;
  Result := res;
end;

function TParque_ModeloTeorico.Calc_PotMW_anemo(roDato, pv3_anemo: NReal;
  cgm: NReal;
  var flg_Convergio: boolean; var pv3_Anemo_s,
  pv3_VientoLibre: NReal): NReal;
var
  cnt_iter: integer;
  pv3, PE_MW_Modelo, rel: NReal;
begin
  if pv3_anemo < 1E-10 then
  begin
    Result := 0;
    exit;
  end;

  pv3 := pv3_anemo / Meteo_etha_dir;
  flg_convergio := False;
  cnt_iter := 0;
  repeat
    PE_MW_Modelo := Calc_PotMW_pvLibre(roDato, pv3, cgm );
    pv3_Anemo_s := GridValAtZ(_0) * Meteo_etha_dir;
    rel := pv3_Anemo / pv3_Anemo_s;
    if (abs(rel - 1) < 0.01) then
      flg_convergio := True
    else
    begin
      Inc(cnt_iter);
      pv3 := rel * pv3;
    end;
  until (cnt_iter > 10) or flg_convergio;
  pv3_VientoLibre:= pv3;
  Result := PE_MW_Modelo;
end;


end.
