unit uActorNodal;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  uCosa, uGlobs, uActores, uFechas, unodos, uCosaConNombre, uFichasLPD,
  xMatDefs, Classes, contnrs, uauxiliares;

type

  { TActorNodal }

  TActorNodal = class(TActor)

    function ReferenciaAlNodo(nodo: TNodo): boolean; virtual; abstract;

  end;

  { TActorUniNodal }

  TActorUniNodal = class(TActorNodal)

  public
    P: TDAOfNReal; // Inyectada como generador al nodo

    (**************************************************************************)
    (*               A T R I B U T O S   P E R S I S T E N T E S              *)
    (**************************************************************************)

    Nodo: TNodo;
    barras_flucar: TDAOfNint;
    codigos_flucar: TDAOfString;
    (**************************************************************************)

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; Nodo: TNodo);
    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    {$ENDIF}
    function ReferenciaAlNodo(nodo: TNodo): boolean; override;
    procedure PubliVars; override;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure Free; override;

    {$IFDEF MIGRACION_PERSISTENCIA}
    class function CreateDataColumnList(xClaseDeCosa:TClaseDeCosa; xVersion: Integer = -2): TDataColumnListOfCosa; override;
    {$ENDIF}

    procedure AfterInstantiation(); override;

    published

      {$IFDEF MIGRACION_PERSISTENCIA}
      property _Nodo: TNodo read nodo write nodo;
      property _barras_flucar: TDAOfNint read barras_flucar write barras_flucar;
      property _codigos_flucar: TDAOfString read codigos_flucar write codigos_flucar;
      {$ENDIF}


  end;

  { TActorBiNodal }

  TActorBiNodal = class(TActorNodal)
  public
    (**************************************************************************)
    (*               A T R I B U T O S   P E R S I S T E N T E S              *)
    (**************************************************************************)

    NodoA, NodoB: TNodo;
    (**************************************************************************)

    // resultados del despacho
    P_NodoA, P_NodoB: TDAOfNReal; //[MW] Potencia inyectada como generador en cada Nodo.

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; NodoA, NodoB: TNodo);

    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    {$ENDIF}
    function ReferenciaAlNodo(nodo: TNodo): boolean; override;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure Free; override;

    {$IFDEF MIGRACION_PERSISTENCIA}
    class function CreateDataColumnList(xClaseDeCosa:TClaseDeCosa; xVersion: Integer=-2): TDataColumnListOfCosa; override;
    {$ENDIF}

    published

      {$IFDEF MIGRACION_PERSISTENCIA}
      property _NodoA: TNodo read NodoA write NodoA;
      property _NodoB: TNodo read NodoB write NodoB;
      {$ENDIF}
  end;

procedure AlInicio;
procedure ALFinal;

implementation

//--------------------------
// Mtodos de TActorUniNodal
//==========================

constructor TActorUniNodal.Create(
 capa: integer; nombre: string; nacimiento, muerte: TFecha;
  lpdUnidades: TFichasLPD; Nodo: TNodo );
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades);
  self.Nodo := Nodo;
  barras_flucar := nil;
  codigos_flucar := nil;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TActorUniNodal.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);

  f.IniciarLecturaRetrasada;

  f.rdReferencia('nodo', TCosa(self.Nodo), Self);

  if f.Version < 49 then
  begin
    barras_flucar := nil;
    codigos_flucar := nil;
  end
  else
  begin
    f.rd('barras_flucar', barras_flucar);
    f.rd('codigos_flucar', codigos_flucar);
  end;
  f.EjecutarLectura;

  if f.Version >= 49 then
    clear_vacias(codigos_flucar);

  self.Nodo := nil;


end;
{$ENDIF}

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure TActorUniNodal.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wrReferencia('nodo', Nodo);
  f.wr('barras_flucar', barras_flucar);
  f.wr('codigos_flucar', codigos_flucar);
end;
{$ENDIF}

function TActorUniNodal.ReferenciaAlNodo(nodo: TNodo): boolean;
begin
  Result := self.nodo = nodo;
end;

procedure TActorUniNodal.PubliVars;
begin
  inherited PubliVars;
  PublicarVariableVR('P', '[MW]', 6, 2, P, True, True);
end;

procedure TActorUniNodal.PrepararMemoria(globs: TGlobs);
begin
  inherited PrepararMemoria(globs);
  setlength(P, globs.NPostes);
end;


procedure TActorUniNodal.Free;
begin
  setlength(P, 0);
  inherited Free;
end;


{$IFDEF MIGRACION_PERSISTENCIA}
class function TActorUniNodal.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: Integer): TDataColumnListOfCosa;
begin
  Result:=inherited CreateDataColumnList(xClaseDeCosa,xVersion);
  Result.AddCosaColumn('nodo');
  Result.AddDAOfIntegerColumn('barras_flucar', 49);
  Result.AddDAOfStringColumn ('codigos_flucar',49);
end;
{$ENDIF}

procedure TActorUniNodal.AfterInstantiation;
begin

  inherited AfterInstantiation;
  clear_vacias(codigos_flucar);
  self.Nodo := nil;

end;


//-------------------------
// Mtodos de TActorBiNodal
//=========================

constructor TActorBiNodal.Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
  lpdUnidades: TFichasLPD; NodoA, NodoB: TNodo);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades);
  self.NodoA := NodoA;
  self.NodoB := NodoB;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TActorBiNodal.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rdReferencia('nodoA', TCosa(NodoA), Self);
  f.rdReferencia('nodoB', TCosa(NodoB), Self);
  f.EjecutarLectura;
end;
{$ENDIF}

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure TActorBiNodal.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wrReferencia('nodoA', NodoA);
  f.wrReferencia('nodoB', NodoB);
end;
{$ENDIF}

function TActorBiNodal.ReferenciaAlNodo(nodo: TNodo): boolean;
begin
  Result := (nodoA = nodo) or (nodoB = nodo);
end;

procedure TActorBiNodal.PrepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
  setlength(P_NodoA, globs.NPostes);
  setlength(P_NodoB, globs.NPostes);
end;

procedure TActorBiNodal.Free;
begin
  setlength(P_NodoA, 0);
  setlength(P_NodoB, 0);
  inherited Free;
end;

{$IFDEF MIGRACION_PERSISTENCIA}
class function TActorBiNodal.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: Integer): TDataColumnListOfCosa;
begin
  Result:=inherited CreateDataColumnList(xClaseDeCosa,xVersion);
  Result.AddCosaColumn('nodoA');
  Result.AddCosaColumn('nodoB');
end;
{$ENDIF}

procedure AlInicio;
begin
  //  registrarClaseDeCosa(TActorUniNodal.ClassName, TActorUniNodal);
  //  registrarClaseDeCosa(TActorBiNodal.ClassName, TActorBiNodal);

end;

procedure AlFinal;
begin
end;

end.
