(*Se agrega la capacidad de tener requerimientos de reserva
al 8/5/15 por Micho mvarela@adme.com.uy*)
unit unodos;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  xmatdefs, usimplex, uglobs,
  SysUtils,
  Classes,
  ucosa,
  uCosaConNombre,
  uActores,
  //uINodos,
  ufichasLPD,
  ufechas;

resourcestring
  rsNodo = 'Nodo';

type

  { TListNodo }
  TListNodo = class(TListaDeCosasConNombre)
  {$IFDEF MIGRACION_PERSISTENCIA}    class function CreateDataColumnList(
      xClaseDeCosa: TClaseDeCosa; xVersion: integer = -2): TDataColumnListOfCosa;
      override;
  {$ENDIF}
  end;


  { TNodo }

  TNodo = class(TActor)
  public

    (**************************************************************************)
    (*               A T R I B U T O S   P E R S I S T E N T E S              *)
    (**************************************************************************)

    ZonaFlucar: integer;

    (**************************************************************************)


    // resultados
    cmarg: TDAOfNReal;

    cReserva: TDAOfNReal;
    //la idea es guardar los multiplicadores si son el costo de reserva.

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; ZonaFlucar: integer);

    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    {$ENDIF}

    procedure PrepararMemoria(globs: TGlobs); override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;
    procedure opt_cargue(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;
    {$IFDEF SPXCONLOG}
    procedure spx_NombrarVariables(s: TSimplex); override;
    {$ENDIF}
    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure Free; override;

    procedure PubliVars; override;
    {$IFDEF MIGRACION_PERSISTENCIA}
    class function CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
      xVersion: integer = -2): TDataColumnListOfCosa; override;
     {$ENDIF}
  published
  {$IFDEF MIGRACION_PERSISTENCIA}
    property _ZonaFlucar: integer read ZonaFlucar write ZonaFlucar;
  {$ENDIF}

  end;



procedure AlInicio;
procedure AlFinal;



implementation

{ TListNodo }

{$IFDEF MIGRACION_PERSISTENCIA}
class function TListNodo.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: integer): TDataColumnListOfCosa;
begin
  Result := inherited CreateDataColumnList(xClaseDeCosa, xVersion);
  Result.ItemListClaseDeCosa := TNodo;
end;

{$ENDIF}

//-------------------
// Métodos de TNodo
//===================
constructor TNodo.Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
  lpdUnidades: TFichasLPD; ZonaFlucar: integer);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades);
  Self.ZonaFlucar := ZonaFlucar;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TNodo.Create_ReadFromText(f: TArchiTexto);
begin
  ZonaFlucar := 0;
  inherited Create_ReadFromText(f);
  if f.Version >= 66 then
  begin
    f.rd('ZonaFlucar', ZonaFlucar);
  end;
end;

{$ENDIF}

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure TNodo.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('ZonaFlucar', ZonaFlucar);
end;

{$ENDIF}

procedure TNodo.PrepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
  setlength(cmarg, globs.NPostes);
  if globs.flg_ReservaRotante then
    setlength(cReserva, globs.NPostes);
end;

function TNodo.InfoAd_: string;
begin
  Result := '';
end;

class function TNodo.DescClase: string;
begin
  Result := rsNodo;
end;

procedure TNodo.opt_nvers(var ivar, ivae, ires: integer);
begin
  // el nodo no agreaga variables de control
  Self.ires := ires;
  // El NODO agrega una restricción por cada POSTE

  if globs.flg_ReservaRotante then
    ires := ires + 2 * globs.NPostes//;
  //si está definido reserva guardo 2 filas para cada poste: Una ec de Potencia y otra de Reserva
  else
    ires := ires + globs.NPostes;//;
end;

procedure TNodo.opt_cargue(s: TSimplex);
var
  ip: integer;
begin
  // nada, cada demanda carga el termino constante y cada generador los variables
{$IFDEF SPXCONLOG}
  spx_NombrarVariables(s);
{$ENDIF}
  if globs.flg_ReservaRotante then
    for ip := 0 to 2 * globs.NPostes - 1 do
      s.FijarRestriccionIgualdad(ires + ip)//;
  else
    for ip := 0 to globs.NPostes - 1 do
      s.FijarRestriccionIgualdad(ires + ip);//;
end;

procedure TNodo.opt_fijarRestriccionesDeCaja(s: TSimplex);
begin
  // no tiene variables no tiene restricciones de caja
end;

procedure TNodo.opt_leerSolucion(s: TSimplex);
var
  iposte: integer;
begin
  // no tiene variables no tiene que leer las variables.
  // lo que podemos leer es el multiplicador de Lagrange para tener
  // el costo marginal del nodo.
  for iposte := 0 to globs.NPostes - 1 do
    cmarg[iposte] := -s.ymult(ires + iposte) * globs.invDurPos[iposte];
    {if usalasdejuego.reserva then
    begin
      for iposte := globs.NPostes to 2*globs.NPostes - 1 do
      cReserva[iposte] :=-s.ymult(ires + iposte) * globs.invDurPos[iposte]
    end;}
end;

{$IFDEF SPXCONLOG}
procedure TNodo.spx_NombrarVariables(s: TSimplex);
var
  iposte: integer;
begin
  for iposte := 0 to globs.NPostes - 1 do
    s.set_NombreRest(ires + iposte, nombre + 'res-nodo' + IntToStr(iposte + 1));
end;

{$ENDIF}

function TNodo.getNombreVar(ivar: integer; var nombre: string): boolean;
begin
  Result := False;
end;

function TNodo.getNombreRes(ires: integer; var nombre: string): boolean;
begin
  if globs.flg_ReservaRotante then
  begin
    if (ires >= self.ires) and (ires < self.ires + 2 * globs.NPostes) then
    begin
      nombre := self.nombre + 'res-nodo' + IntToStr(ires - self.ires + 1);
      Result := True;
    end
    else
      Result := False;
  end
  else
  begin
    if (ires >= self.ires) and (ires < self.ires + globs.NPostes) then
    begin
      nombre := self.nombre + 'res-nodo' + IntToStr(ires - self.ires + 1);
      Result := True;
    end
    else
      Result := False;
  end;
end;

procedure TNodo.Free;
begin
  setlength(cmarg, 0);
  inherited Free;
end;

procedure TNodo.PubliVars;
begin
  inherited PubliVars;
  PublicarVariableVR('cmg', '[USD/MWh]', 6, 1, cmarg, True, True);
  {PublicarVariableVR('cReserva', '[USD/MWh]', 6, 1, cReserva, True, True);}
end;

{$IFDEF MIGRACION_PERSISTENCIA}
class function TNodo.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: integer): TDataColumnListOfCosa;
begin
  Result := inherited CreateDataColumnList(xClaseDeCosa, xVersion);
  Result.AddIntegerColumn('ZonaFlucar', 0, 66);
end;

{$ENDIF}

procedure AlInicio;
begin
  registrarClaseDeCosa(TNodo.ClassName, TNodo);
  registrarClaseDeCosa(TListNodo.ClassName, TListNodo);
end;

procedure AlFinal;
begin
end;

end.
