(*
  Una cosa es un objeto que puede guardarse.
  Una TListaDeCosas, es una lista de cosas
*)

// Este define, permite que al leer se escriban en consola la clase y el nombre
// de las cosas leídas si son del tipo TCosaConNombre. Sirve a los efectos de
// debuggin cuando se tranca durante la lectura de la sala de juegos.
{xDEFINE WRLN_COSAS_CON_NOMBRE_LEIDAS}

// Este define hace que se escriban todas las líneas leídas.
{xDEFINE LECTURA_VERBOSA}

unit ucosa;

interface

uses
  SysUtils, Classes, Dialogs, xmatdefs, ufechas, typinfo,
  ucampolink,
  ubuscaarchivos,
  {$IFDEF CASOS_MACROVARS} uparametros, {$ENDIF}
  contnrs, RegExpr, uCosaHashList,
  uauxiliares,
  uevaluador,
  uversion_architexto,
  {$IFDEF MIGRACION_PERSISTENCIA}
  unitLecturaArchivo,
  {$ENDIF}
  uconstantesSimSEE;

resourcestring
  rs_NoSeEncuentraArchivo = 'No se encuentra el archivo: ';
  rs_ElArchivo = 'El archivo';
  rs_VersionPosterior =
    'esta escrito con una versión posterior a la instalada. Debe actualizar el software.';
  rs_NoInstanciasDe = 'No hay instancias de cosas de la clase';


const
  formatoReales = ffFixed;

type
  TProcMsg = procedure(msg: string);

  TArchiTexto = class; // definido más adelante.

  // Define la referenica a un archivo externo.
  // se utiliza para centralizar las dependencias.
  TArchiRef = class
  private
    archiRefStr: string;
    function get_archi: string;
    procedure set_archi(const archi_str: string);

  public
    constructor Create(archiRef_: string);
    procedure Free; virtual;

    // permite pasar carpeta en cuyo caso si no encuentra el archivo
    // con la archiRefStr original y sino busca en la lista_caminos
    function testear: boolean;
    property archi: string read get_archi write set_archi;
  end;


  // Clase de todas las cosas que usamos en una sala.

  { TCosa }

  TCosa = class;
  TClaseDeCosa = class of TCosa;

  TDAofCosa = array of TCosa;


  TCosa = class{$IFDEF MIGRACION_PERSISTENCIA}(TPersistent){$ENDIF}
  public
    capa: integer;
    {$IFDEF CASOS_MACROVARS}
    par_lst: TListaDeParametros;
    {$ENDIF}

    //dfusco@20150422
    id: integer;

    constructor Create; overload; virtual;
    constructor Create(capa: integer); overload; virtual;

    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); virtual;
    procedure WriteToText(f: TArchiTexto); virtual;
    {$ENDIF}

    // crea el archivo y lo sobreescribe si existe y llama a WriteToText
    procedure WriteToArchi(archi: string);

    // retorna el texto como se guarda en el archivo de texto de  la sala.
    function asMemo: TStrings;
    class function Create_FromMemo(const memo: TStrings): TCosa;
    function Create_Clone: TCosa; virtual;

    // obtiene el string infoAd pero lo recorta a 20 caracteres.
    function InfoAd_20: string;

    function InfoAd_: string; virtual;
    class function DescClase: string; virtual;

    class function recDef: TRecLink; virtual;

    procedure Free; virtual;

    function Apodo: string; virtual;
    // en las cosas con nombre retorna el nombre, en las sin nombre retonra '-'

    procedure ActivarCapas(const capasActivas: TDAOfNInt); virtual;

    function Validate(): boolean; virtual;

    {$IFDEF MIGRACION_PERSISTENCIA}
    //Al comienzo del programa se crea la BD con las tablas en la última versión (-2).
    //Luego cuando se lee el archivo se crea una tabla en la versión de éste (DataTableAux) y se convierte a la última versión.
    class function CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
      xVersion: integer = -2): TDataColumnListOfCosa; virtual; overload;

    class function CreateDataConversionList(): TListDataConversion; virtual;
    {$ENDIF}

    // Procedimiento que se debera ejecuar luego de creada la Instancia
    // desde un TDataStore.
    procedure AfterInstantiation(); virtual;

  published
    {$IFDEF MIGRACION_PERSISTENCIA}
    property _Capa: integer read capa write capa;
    {$ENDIF}
  end;

  TCualquierClase = class of TObject;

  TDAOfTCosa = array of TCosa;

  { TListaDeCosas }
  TListaDeCosas = class(TCosa)

    function getItem(i: integer): TCosa;
    procedure setItem(i: integer; cosa: TCosa);
    function getCapacity: integer;
    procedure setCapacity(newCapacity: integer);
  public

    (**************************************************************************)
    (*              A T R I B U T O S   P E R S I S T E N T E S               *)
    (**************************************************************************)

    {dfusco@20141029
     Se agrega atributo que representa el largo de la lista.}
    n: integer;
    lst: TList;
    (**************************************************************************)

    idCarpeta: string; // identificador para salvar en carpeta
    constructor Create(capa: integer; idCarpeta: string);
    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    {$ENDIF}
    function Add(cosa: TCosa): integer; virtual;

    //Si freeElementos = true llama a ClearFreeElementos, sino a Clear,
    //luego hace add de todas las cosas en listaDeCosas
    procedure ponerIgualA(listaDeCosas: TListaDeCosas; freeElementos: boolean);
    procedure ActivarCapas(const CapasActivas: TDAOfNInt); override;
    procedure Free; override;
    //Libera la lista sin destruir los objetos que contiene
    procedure FreeSinElemenentos;
    function Count: integer;
    procedure insert(indice: integer; cosa: TCosa);
    function Remove(cosa: TCosa): integer;
    procedure Delete(indice: integer);
    procedure Pack;
    procedure Exchange(indice1, indice2: integer);
    function IndexOf(cosa: TCosa): integer;
    function replace(cosaARemplazar, cosaNueva: TCosa): integer;
    procedure Clear;
    procedure ClearFreeElementos;
    procedure Sort(Compare: TListSortCompare);
    function getList: TList;
    function toArray: TDAOfTCosa;
    property Capacity: integer read getCapacity write setCapacity;
    property items[i: integer]: TCosa read getItem write setItem; default;

    {$IFDEF MIGRACION_PERSISTENCIA}
    class function CreateDataColumnList(xClaseDeCosa: TClaseDeCosa = nil;
      xVersion: integer = -2): TDataColumnListOfCosa; override;
    {$ENDIF}

  published
    {$IFDEF MIGRACION_PERSISTENCIA}
    property _n: integer read n write n;
    property _lst: TList read lst write lst;
    {$ENDIF}
  end;

(* La función ProcesoLineaRes puede retornar laguno de los siguientes
valores. No se prevé un código de error pues la función lanza
excepciones con cualquier error que detecte. *)
  TResultadoProcesoLinea = (
    CTPLR_CONTINUAR,   // continuar procesando
    CTPLR_FIN_OBJETO,  // se llegó al fin del objeto (cerrar la lista)
    CTPLR_ABRIR_NUEVO_OBJETO); // Se encontró un objeto abrir otra lista

  TTipoCampo = (
    CTC_NReal,
    CTC_String,
    CTC_ArchRef,
    CTC_NInt,
    CTC_NCardinal,
    CTC_Boolean,
    CTC_DAOfNReal,
    CTC_DAOfNInt,
    CTC_DAOfBoolean,
    CTC_DAOfString,
    CTC_StringList,
    CTC_Fecha,
    CTC_TDataRowOfCosa,
    CTC_Cosa,
    CTC_Referencia);


  TProcFijarValorPorDefecto = procedure(const nombre: string;
    pVal: Pointer; tipo: TTipoCampo; const klinea: integer);




  { TListCamposParaLeer }
  TListCamposParaLeer = class(TList)
  public
    {$IFDEF MIGRACION_PERSISTENCIA}
    DataRow: TDataRowOfCosa;
    {$ENDIF}
    cosa: TCosa;

    lecturaRetrasada, leiFinDeObjeto: boolean;
    constructor Create(IdentificadorDeLaClaseDelObjeto: string);
    procedure Esperar_NReal(const Nombre: string; var Variable: NReal;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_String(const Nombre: string; var Variable: string;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_ArchRef(const Nombre: string; var Variable: TArchiRef;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_NInt(const Nombre: string; var Variable: integer;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_NCardinal(const Nombre: string; var Variable: cardinal;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_Boolean(const Nombre: string; var Variable: boolean;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_DAOfNReal(const Nombre: string; var Variable: TDAOfNReal;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_DAOfNInt(const Nombre: string; var Variable: TDAOfNInt;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_DAOfBoolean(const Nombre: string; var Variable: TDAOfBoolean;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_DAOfString(const Nombre: string; var Variable: TDAOfString;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_StringList(const Nombre: string; var Variable: TStringList;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_Fecha(const Nombre: string; var Variable: TFecha;
      proc_vpd: TProcFijarValorPorDefecto); overload;
    procedure Esperar_Fecha(const Nombre: string; var Variable: TFecha;
      proc_vpd: TProcFijarValorPorDefecto; flgError: integer); overload;

    {$IFDEF MIGRACION_PERSISTENCIA}
    procedure Esperar_TDataRowOfCosa(const nombre: string;
      var variable: TDataRowOfCosa; proc_vpd: TProcFijarValorPorDefecto);
    {$ENDIF}
    procedure Esperar_Cosa(const Nombre: string; var Variable: TCosa;
      proc_vpd: TProcFijarValorPorDefecto);
    procedure Esperar_Referencia(const Nombre: string; var Variable: TCosa;
      proc_vpd: TProcFijarValorPorDefecto; referente: TCosa);

    function ProcesarLinea(klinea: integer; linea: string; var extres: string;
      var ColumnName: string ): TResultadoProcesoLinea;
    procedure Free;
  private
    IdentificadorDeLaClaseDelObjeto: string; // identificador de la clase de objeto
    lineaAbierta: boolean;
    xNombre, xval: string; // valores bajo procesamiento
    procedure ResolverCampoValor(const xNombre, xVal: string);
    procedure Ejecutar_vpds(klinea: integer);
    procedure Ejecutar_vpd(klinea: integer; nombreCampo: string);
{$IFDEF DBG}
    //Para debug, muestra los campos que la lista tiene por leer
    function NombresCamposPorLeer: TDAofString;
{$ENDIF}
  end;

  TCampoParaLeer = class
  public
    nombre: string;
    tipo: TTipoCampo;
    pVal: pointer;
    proc_vpd: TProcFijarValorPorDefecto;
    referente: TCosa;
    flgError: integer; // 0 es showmessage, 1 no muestra nada
    constructor Create(const Nombre: string; pVal: Pointer; Tipo: TTipoCampo;
      proc_vpd: TProcFijarValorPorDefecto); overload;
    constructor Create(const Nombre: string; pVal: Pointer; Tipo: TTipoCampo;
      proc_vpd: TProcFijarValorPorDefecto; flgError: integer); overload;
    constructor Create(const Nombre: string; pVal: Pointer; Tipo: TTipoCampo;
      proc_vpd: TProcFijarValorPorDefecto; referente: TCosa); overload;
    procedure Ejecutar_vpd(const klinea: integer);
  end;

  ExcepcionNoSePidioElCampo = class(Exception)
  end;

  ExcepcionResolverCampoValor = class(Exception)
  end;

  ExcepcionClaseNoRegistrada = class(Exception)
  end;

  ExcepcionPersistentAttributeDoesNotHasProperty = class(Exception)
  end;

  ExcepcionPersistentPropertyHasToPointAtAttribute = class(Exception)
  end;

  TListaDeObjetosPorLeer = class(TList)
  public
    constructor Create;
    procedure Free;
    function ListaActiva: TListCamposParaLeer;
    procedure FreeListaPorLeer; // libera la última lista(cerró objeto)
    procedure InicioObjeto(const NombreClase: string);
    procedure FinalObjeto;
  end;

  TArchiTexto = class
  public
    klinea: integer;
    Version: integer;
    nombreArchivo: string;
    aux_idCarpeta: string;
    // auxiliar para pasaje de parámetros de TListaDeCosas CreateReadFromText

    unArchivoPorCosa: boolean;
    padre: TArchiTexto;
    ramas: TList;

    constructor CreateForRead(const nombreArchivo: string; abortarEnError: boolean);
    constructor CreateForWrite(const nombreArchivo: string;
      crearBackupSiExiste: boolean; maxNBackups: integer);

    procedure Free;
    function NextLine: string;
    procedure writeline(const s: string);

    procedure wr(const nombre: string; const valor: NReal; precision: integer;
      decimales: integer); overload;
    procedure wr(const nombre: string; const valor: NReal); overload;
    procedure wr(const nombre: string; const valor: string); overload;
    procedure wrArchRef(const nombre: string; const valor: TArchiRef);
    procedure wr(const nombre: string; const valor: NInt); overload;
    procedure wr(const nombre: string; const valor: cardinal); overload;
    procedure wr(const nombre: string; const valor: boolean); overload;
    procedure wr(const nombre: string; const valor: TDAOfNReal;
      precision: integer; decimales: integer); overload;
    procedure wr(const nombre: string; const valor: TDAOfNReal); overload;
    procedure wr(const nombre: string; const valor: TDAOfNInt); overload;
    procedure wr(const nombre: string; const valor: TDAOfBoolean); overload;
    procedure wr(const nombre: string; const valor: TDAOfString); overload;
    procedure wr(const nombre: string; const valor: TStringList); overload;
    procedure wr(const nombre: string; const valor: TFecha); overload;
    procedure wr(const nombre: string; const valor: TCosa); overload;
    procedure wrReferencia(const nombre: string; const valor: TCosa);

    procedure rd(const nombre: string; var valor: NReal); overload;
    procedure rd(const nombre: string; var valor: string); overload;
    procedure rdArchRef(const nombre: string; var valor: TArchiRef);
    procedure rd(const nombre: string; var valor: NInt); overload;
    procedure rd(const nombre: string; var valor: cardinal); overload;
    procedure rd(const nombre: string; var valor: boolean); overload;
    procedure rd(const nombre: string; var valor: TDAOfNReal); overload;
    procedure rd(const nombre: string; var valor: TDAOfNInt); overload;
    procedure rd(const nombre: string; var valor: TDAOfString); overload;
    procedure rd(const nombre: string; var valor: TDAOfBoolean); overload;
    procedure rd(const nombre: string; var valor: TStringList); overload;
    procedure rd(const nombre: string; var valor: TFecha); overload;
    // flgError= 0 lo mismo que sin flg. flgError= 1 no avisa del error
    procedure rd(const nombre: string; var valor: TFecha; flgError: integer); overload;
    {$IFDEF MIGRACION_PERSISTENCIA}
    procedure rd(const nombre: string; var valor: TDataRowOfCosa); overload;
    {$ENDIF}
    procedure rd(const nombre: string; var valor: TCosa); overload;
    procedure rdReferencia(const nombre: string; var valor: TCosa; referente: TCosa);


    procedure IniciarLecturaRetrasada;
    procedure EjecutarLectura;

    function CreateRamaForWrite(
      const Carpeta, NombreClase, NombreCosa: string): TArchiTexto;
    function CreateRamaForRead(
      const Carpeta, NombreClase, NombreCosa: string): TArchiTexto;
  private
    f: textfile;
    f_abierto: boolean;

    listaObjetosPorLeer: TListaDeObjetosPorLeer; //TListCamposParaLeer;
    abortarEnError: boolean;
    oldFileMode: byte;

    indentstr: string;
    function EOF: boolean;
    function quedanCosasDelObjetoActualPorLeer(claseObjeto: string): boolean;
    function ListaPorLeer: TListCamposParaLeer;

    {$IFDEF MIGRACION_PERSISTENCIA}
    //dfusco@20141001
    {Este procedimiento crea una instancia de TDataRowOfCosa a partir de la
    clase del campo que se leyó del archivo de texto (Ej: <+TGlobs>) y se hace
    la lectura de cada columna definida (DataColumnList) y el valor se guarda en
    la celda correspondiete de la fila creada.
    Sustituye al procedimiento ReadCosaFromText
    }
    procedure ReadDataRowFromText(var DataRow: TDataRowOfCosa;
      ColumnName: string; cosaStrId: string);
    {$ENDIF}
    procedure ReadCosaFromText(var cosa: TCosa; cosaStrId: string);

  end;



procedure ValorPorDefectoNulo(const nombre: string; pVal: Pointer;
  tipo: TTipoCampo; const klinea: integer);
procedure ValorPorDefecto_ERROR(const nombre: string; pVal: Pointer;
  tipo: TTipoCampo; const klinea: integer);

var
  registro_de_ClasesDeCosas_: TListaDeCosas;
  procMsgValorPorDefecto: TProcMsg;
  procMsgErrorLectura: TProcMsg;
  procMsgAdvertenciaLectura: TProcMsg;

  MrFlag: boolean = False;

{$IFDEF CNT_COSAS}
type
  TCntCosasClase = class
  private
    clase: TClass;
    cnt: integer;
  public
    constructor Create(xclase: TClass);
  end;

  TCntCosas = class
  private
    CntCosasClases: TList {of TCntCosasClase};
  public
    cnt_Cosas: integer;

    constructor Create;
    function incCntCosasClase(clase: TClass): integer;
    function decCntCosasClase(clase: TClass): integer;
    function CntCosasClase(clase: TClass): integer;
    procedure Free;
  end;

var
  CantCosas: TCntCosas;

{$ENDIF}

function registrarClaseDeCosa(const nombre: string; Clase: TClaseDeCosa;
  CreateDataTable: boolean = True): integer;
function getClaseOf(const nombre: string): TCualquierClase;
// si la clase no está registrada retorna '<CLASE_NO_REGISTRADA>'
function getIdStrOf(clase: TClaseDeCosa): string;

//Retorna TRUE si la clase child está en el árbol de herencias de la clase
// Parent y false en caso contrario.
function IsChild(Parent, Child: string): boolean; overload;
function IsChild(Parent, Child: TClass): boolean; overload;

procedure LimpiarRegistroDeClasesDeCosas;

// Retorna la cadena de padres separados por '->'
function ParentsStrList(clase: TClass): string;

procedure AlInicio;
procedure AlFinal;

procedure FreeAndNil(var obj: TCosa);
function referenciaACosa(cosa: TCosa): string;

type
  ExceptionFileNotFound = class(Exception)
  public
    arch: string;
    constructor Create(arch, comentarios: string);
  end;

  { TFichaClase }

  TFichaClase = class(TCosa)
  public
    strClaseId: string;
    clase: TClaseDeCosa;
    //dfusco@20141023 se agrega la lista de definición de campos.
    DataColumnList: TList;
    constructor Create(strClaseID: string; clase: TClaseDeCosa);
    procedure Free; override;
  end;

var
  // Lista de referencias a archivos externos.
  ReferenciasAArchivosExternos: TList;

  // Lista de caminos de búsquedas de los archivos externos.
  // La aplicación debe agregar aquí las rutas de búqueda que
  // se deben usar en la función testear de las referencias a
  // archivos.
  lista_caminos: TBuscaArchivos;


  {$IFDEF MIGRACION_PERSISTENCIA}
  //dfusco-fbarreto@20141023 El DataStore pasa a ser global para crear las
  //estructuras de las tablas AlInicio.
  DATA_STORE: TDataStore;
  {$ENDIF}

  //Lista de Cosas con nombre en el DATA:STORE
  cosasConNombre: TListaDeCosas;

// retorna la constante de versión de formato de sala almacenada en la unidad
// uversion_architexto
function version_ArchiTexto: integer;

{$IFDEF MIGRACION_PERSISTENCIA}
procedure PrintDataSet(DataSet: TDataStore);
procedure PrintDataSetToFile(DataSet: TDataStore);
{$ENDIF}

implementation

uses
  //  uVirtualTablesDefs,
  uCosaConNombre, usalasdejuego, uFuenteConstante, uActores, unodos,
  ufichasLPD, udemandadetallada;

var
  global_FlgError: integer;


function version_ArchiTexto: integer;
begin
  Result := uversion_architexto.version_ArchiTexto;
end;

{$IFDEF MIGRACION_PERSISTENCIA}
procedure PrintDataSet(DataSet: TDataStore);
var
  z: integer;
  j: integer;
  i: integer;
  a: integer;
begin
  WriteLn();
  WriteLn('---');
  WriteLn('SetName = ' + DataSet.SetName);
  for z := 0 to DataSet.DataTableListCount - 1 do
  begin
    WriteLn('   ' + 'TableName = ' + DataSet.DataTableList[z].TableName);
    if DataSet.DataTableList[z].TableName = 'dt_TFichasLPD' then
      a := 1;
    for j := 0 to DataSet.DataTableList[z].DataRowListCount - 1 do
    begin
      WriteLn('      ' + 'RowId = ' +
        IntToStr(DataSet.DataTableList[z].DataRowList[j]._RowId) + ' | SIUD = ' +
        GetEnumName(TypeInfo(TSIUD), Ord(DataSet.DataTableList[z].DataRowList[j].SIUD)));
      for i := 0 to DataSet.DataTableList[z].DataColumnListCount - 1 do
      begin
        WriteLn('         ' + DataSet.DataTableList[z].DataColumnList[i].ColumnName +
          ' = ' + DataSet.DataTableList[z].DataRowList[j].DataCellList[i].sval);
      end;
    end;
  end;
  WriteLn('---');
  WriteLn();
end;

procedure PrintDataSetToFile(DataSet: TDataStore);
var
  z: integer;
  j: integer;
  i: integer;
  a: integer;
  f: TextFile;
  xFileName: string;
  oldFileMode: byte;
begin

  xFileName := 'pepe.txt';
  assignFile(f, xFileName);
  oldFileMode := FileMode;
  filemode := fmOpenWrite;
  Rewrite(f);
  if ioresult <> 0 then
    raise Exception.Create('TArchiTexto.CreateForWrite(' + xFileName + ')');

  WriteLn(f);
  WriteLn(f, '---');
  WriteLn(f, 'SetName = ' + DataSet.SetName);
  for z := 0 to DataSet.DataTableListCount - 1 do
  begin
    WriteLn(f, '   ' + 'TableName = ' + DataSet.DataTableList[z].TableName);
    if DataSet.DataTableList[z].TableName = 'dt_TFichasLPD' then
      a := 1;
    for j := 0 to DataSet.DataTableList[z].DataRowListCount - 1 do
    begin
      WriteLn(f, '      ' + 'RowId = ' +
        IntToStr(DataSet.DataTableList[z].DataRowList[j]._RowId) + ' | SIUD = ' +
        GetEnumName(TypeInfo(TSIUD), Ord(DataSet.DataTableList[z].DataRowList[j].SIUD)));
      for i := 0 to DataSet.DataTableList[z].DataColumnListCount - 1 do
      begin
        WriteLn(f, '         ' +
          DataSet.DataTableList[z].DataColumnList[i].ColumnName + ' = ' +
          DataSet.DataTableList[z].DataRowList[j].DataCellList[i].sval);
      end;
    end;
  end;
  WriteLn(f, '---');
  WriteLn(f);

  CloseFile(f);
  filemode := oldFileMode;
end;
{$ENDIF}


//---------------------------------
// Métodos de TArchiRef
//=================================


constructor TArchiRef.Create(archiRef_: string);
begin
  archiRefStr := archiRef_;
  ReferenciasAArchivosExternos.add(self);
  //writeln( 'TArchiRef.Crete: '+archiRef_ );
end;

procedure TArchiRef.Free;
var
  i: integer;
begin
  i := ReferenciasAArchivosExternos.IndexOf(Self);
  if i >= 0 then
    ReferenciasAArchivosExternos.Delete(i);
  inherited Free;
end;

function TArchiRef.get_archi: string;
begin
  Result := archiRefStr;
end;

procedure TArchiRef.set_archi(const archi_str: string);
begin
  archiRefStr := archi_str;
end;

function TArchiRef.testear: boolean;
var
  nombre: string;
  camino: string;
  carpeta: string;
  buscando: boolean;
  k: integer;

begin
  if FileExists(archiRefStr) then
    Result := True
  else
  begin
    buscando := lista_caminos.Count > 0;
    k := 0;
    while buscando and (k < lista_caminos.Count) do
    begin
      carpeta := lista_caminos[k];
      if carpeta <> '' then
      begin
        nombre := ExtractFileName(archiRefStr);
        camino := ExtractFilePath(archiRefStr);
        if length(camino) > length(nombre) then
        begin
          Delete(nombre, 1, length(camino));
        end;
        nombre := carpeta + barraDir + nombre;
        if FileExists(nombre) then
        begin
          archiRefStr := nombre;
          buscando := False;
        end
        else
          Inc(k);
      end;
    end;
    Result := not buscando;
  end;
end;



//---------------------------------
// Métodos de ExceptionFileNotFound
//=================================

constructor ExceptionFileNotFound.Create(arch, comentarios: string);
begin
  inherited Create(rs_NoSeEncuentraArchivo + ': ' + arch + '. ' + comentarios);
  self.arch := arch;
end;




//-----------------
// Métodos de TCosa
//=================


constructor TCosa.Create;
begin
  inherited Create;
  self.id := -1;
end;

constructor TCosa.Create(capa: integer);
begin
  inherited Create;
  self.capa := capa; // CapaPorDefecto
end;



function TCosa.Apodo: string;
begin
  Result := '-';
end;


procedure TCosa.ActivarCapas(const capasActivas: TDAOfNInt);
begin
  // este no hace nada.
end;

procedure TCosa.WriteToArchi(archi: string);
var
  f: TArchiTexto;
begin
  f := TArchiTexto.CreateForWrite(archi, False, 0);
  try
    f.wr(apodo, Self);
  finally
    f.Free;
  end;
end;

function TCosa.Create_Clone: TCosa;
var
  f: TArchiTexto;
  a: TCosa;
  dirTemporales: string;
begin
  {dfusco@20150423
   Si se está en la nueva versión del programa (PERSISTENCIA TEXTO) no voy a
   clonar el objeto, me voy a quedar con el mismo porque necesito el cosa.id.
   Si clono como está acá el id lo pierdo porque no persiste.
   Como el objeto es temporal no hay problema en hacer esto...
   Al abrir y cerrar la ventana se va a destruir porque se guarda siempre en
   la base de datos. }
  {$IFDEF MIGRACION_PERSISTENCIA}
  Result := self;
  {$ELSE}
  a := nil;
  dirTemporales := getDir_Tmp;

  f := TArchiTexto.CreateForWrite(dirTemporales + 'aux_clonar.tmp', False, 0);
  try
    f.wr('_', self);
  finally
    f.Free;
  end;

  f := TArchiTexto.CreateForRead(dirTemporales + 'aux_clonar.tmp', True);
  try
    f.rd('_', a);
    Result := a;
  finally
    f.Free;
  end;
  {$ENDIF}
end;

function TCosa.asMemo: TStrings;
var
  f: TArchiTexto;
  ff: textfile;
  a: TCosa;
  dirTemporales: string;
  r: string;
  res: TStringList;
begin
  dirTemporales := getDir_Tmp;
  f := TArchiTexto.CreateForWrite(dirTemporales + 'aux_clonar.tmp', False, 0);
  try
    f.wr('_', self);
  finally
    f.Free;
  end;

  res := TStringList.Create;
  assignfile(ff, dirTemporales + 'aux_clonar.tmp');
  reset(ff);
  while not EOF(ff) do
  begin
    readln(ff, r);
    res.add(r);
  end;
  closefile(ff);
  Result := res;
end;

class function TCosa.Create_FromMemo(const memo: TStrings): TCosa;
var
  f: TArchiTexto;
  ff: textfile;
  a: TCosa;
  dirTemporales: string;
  k: integer;
begin
  a := nil;
  dirTemporales := getDir_Tmp;
  assignfile(ff, dirTemporales + 'aux_clonar.tmp');
  rewrite(ff);
  for k := 0 to memo.Count - 1 do
    writeln(ff, memo[k]);
  closefile(ff);

  f := TArchiTexto.CreateForRead(dirTemporales + 'aux_clonar.tmp', True);
  try
    f.rd('_', a);
    Result := a;
  finally
    f.Free;
  end;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TCosa.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create;
  if f.version < 91 then
    capa := 0
  else
  begin
    f.IniciarLecturaRetrasada;
    f.rd('capa', capa);
    f.EjecutarLectura;
  end;
end;

{$ENDIF}

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure TCosa.WriteToText(f: TArchiTexto);
begin
  f.wr('capa', capa);
end;

{$ENDIF}

procedure TCosa.Free;
var
  k: integer;
begin
  ucosaConNombre.eliminar_referencias_del(self);
  inherited Free;
end;


function TCosa.InfoAd_: string;
begin
  Result := IntToStr(capa) + ', ';
end;

function TCosa.InfoAd_20: string;
var
  res: string;
begin
  res := infoAd_;
  if length(res) > 20 then
  begin
    Delete(res, 21, length(res) - 20);
    res := res + '...';
  end;
  Result := res;
end;


class function TCosa.DescClase: string;
begin
  Result := 'Cosa';
end;

class function TCosa.recDef: TRecLink;
var
  res: TRecLink;
begin
  res:= TRecLink.Create;
  res.add( TCampoLink.Create('capa', TCosa(nil).capa ) );
  result:= res;
end;


{$IFDEF MIGRACION_PERSISTENCIA}
class function TCosa.CreateDataConversionList: TListDataConversion;
begin
  Result := TListDataConversion.Create;
end;

class function TCosa.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: integer): TDataColumnListOfCosa;
begin
  Result := TDataColumnListOfCosa.Create(xClaseDeCosa, xVersion);
  Result.AddIntegerColumn('capa', 0);
end;
{$ENDIF}

procedure TCosa.AfterInstantiation;
begin

end;

function TCosa.Validate: boolean;
begin
  Result := True;
end;

{$IFDEF CNT_COSAS}
constructor TCntCosasClase.Create(xclase: TClass);
begin
  clase := xclase;
  cnt := 1;
end;

constructor TCntCosas.Create;
begin
  cnt_Cosas := 0;
  CntCosasClases := TList.Create;
end;

function TCntCosas.incCntCosasClase(clase: TClass): integer;
var
  i: integer;
  resultado: integer;
  aux: TCntCosasClase;
begin
  resultado := 0;
  for i := 0 to CntCosasClases.Count - 1 do
    if TCntCosasClase(CntCosasClases[i]).clase = clase then
    begin
      TCntCosasClase(CntCosasClases[i]).cnt := TCntCosasClase(CntCosasClases[i]).cnt + 1;
      resultado := TCntCosasClase(CntCosasClases[i]).cnt;
      break;
    end;
  if resultado = 0 then
  begin
    aux := TCntCosasClase.Create(clase);
    resultado := aux.cnt;
    CntCosasClases.Add(aux);
  end;
  cnt_Cosas := cnt_Cosas + 1;
  Result := resultado;
end;

function TCntCosas.decCntCosasClase(clase: TClass): integer;
var
  i, ipos: integer;
begin
  ipos := -1;
  for i := 0 to CntCosasClases.Count - 1 do
    if TCntCosasClase(CntCosasClases[i]).clase = clase then
    begin
      TCntCosasClase(CntCosasClases[i]).cnt := TCntCosasClase(CntCosasClases[i]).cnt - 1;
      ipos := i;
      cnt_Cosas := cnt_Cosas - 1;
      break;
    end;
  if ipos <> -1 then
    Result := TCntCosasClase(CntCosasClases[ipos]).cnt
  else
    raise Exception.Create(rs_NoInstanciasDe + ': ' + clase.ClassName);
end;

function TCntCosas.CntCosasClase(clase: TClass): integer;
var
  i, ipos: integer;
begin
  ipos := -1;
  for i := 0 to CntCosasClases.Count - 1 do
    if TCntCosasClase(CntCosasClases[i]).clase = clase then
    begin
      ipos := i;
      break;
    end;
  if ipos <> -1 then
    Result := TCntCosasClase(CntCosasClases[i]).cnt
  else
    Result := 0;
end;

procedure TCntCosas.Free;
var
  i: integer;
begin
  for i := 0 to CntCosasClases.Count - 1 do
    TCntCosasClase(CntCosasClases[i]).Free;
  CntCosasClases.Free;
  inherited Free;
end;

{$ENDIF}

// métodos de TFichaClase
constructor TFichaClase.Create(strClaseID: string; clase: TClaseDeCosa);
begin
  inherited Create;
  Self.strClaseId := strClaseId;
  self.clase := clase;
end;

procedure TFichaClase.Free;
begin
  inherited Free;
end;

{$IFDEF MIGRACION_PERSISTENCIA}
procedure TArchiTexto.ReadDataRowFromText(var DataRow: TDataRowOfCosa;
  ColumnName: string; cosaStrId: string);
var
  clase: TClaseDeCosa;
  j, jj, a: integer;
  Name, aaa: string;
  Count: integer;
  DataColumn: TDataColumnOfCosa;

  DataTypeDataRow: TDataTypeDataRowOfCosa;
  DataTableName: string;
  DataTable: TDataTableOfCosa;
  h: integer;

  DataRowChild: TDataRowOfCosa;

  n: integer;
  esPeriodica: boolean;
  nombre: string;

  cl: boolean; //bandera que me indica si TListaDeCosasConNombre tendrá elementos (n>0)
  z: integer;
  DataColumn2: TDataColumnOfCosa;
  DataRow2: TDataRowOfCosa;
  DataTable2: TDataTableOfCosa;

  index: integer;

  sreferencia: string;
  claseReferido: TClaseDeCosa;
  nombreClaseReferido, nombreAtributoReferido: string;

  dataTableAux: TDataTableOfCosa;
  dataRowAux: TDataRowOfCosa;

  DataTableOfList: TDataTableOfListOfCosa;
  DataRowOfList: TDataRowOfCosa;
  LastIdParent: integer;
  strValor: string;
  adt: TDateTime;
  ad: NReal;
  b: string;

begin

  cl := False;

  clase := TClaseDeCosa(getClaseOf(CosaStrId));
  if clase = nil then
    raise ExcepcionClaseNoRegistrada.Create(
      'TArchiTexto.ReadDataRowFromText: Error, ' + CosaStrId +
      ' No es una clase registrada. Procesando Línea: ' + IntToStr(klinea));

  if clase.ClassName = 'TListGenerador' then
    a := 1;

  //Tabla auxiliar que corresponde a la version del archiText

  DataTableName := 'dt_' + clase.ClassName;
  DataTableAux := TDataTableOfCosa.Create(DataTableName, clase, self.Version);
  DataRowAux := TDataRowOfCosa.Create(DataTableAux, 'dr_' + ColumnName);
  DataRowAux.SIUD := TSIUD.TSELECT;
  DataTableAux.AddDataRow(DataRowAux);
  Count := dataTableAux.DataColumnListCount;
  self.IniciarLecturaRetrasada;
  j := 0;
  while j < Count do
  begin
    DataRowAux.AddDataCell(TDataCellOfCosa.Create);
    DataColumn := DataTableAux.DataColumnList[j];

    if LowerCase(DataColumn.ColumnName) = 'flg_controlcotaobjetivoinferior' then
      a := 1;

    if (Datacolumn.ColumnName[1] <> ':') then
    begin
      if (DataColumn.DataType is TDataTypeDataRowOfCosa) then
      begin
        DataRowChild := nil;
        self.rd(DataColumn.ColumnName, DataRowChild);
        self.EjecutarLectura; //FUNDAMENTAL PARA QUE TODO FUNCIONE BIEN!
        if (DataRowChild <> nil) then
          DataRowAux.LastDataCell.sval :=
            TDataTableOfCosa(DataRowChild.DataTableOwner).TableName + '.' +
            IntToStr(DataRowChild._RowId);
      end
      else if (DataColumn.DataType is TDataTypeListOfCosa) then
      begin
        if (cl = True) then
        begin
          DataTableOfList := TDataTableOfListOfCosa(
            DATA_STORE.GetDataTableByName('dt_TList'));
          if Assigned(DataTableOfList) then
          begin
            LastIdParent := DataTableOfList.GetNextIdParent;
            DataRowAux.LastDataCell.sval := DataTableOfList.TableName +
              '.IdParent.' + IntToStr(LastIdParent);
          end;
        end;
      end
      {else if (DataColumn.DataType is TDataTypeReferenciaOfCosa) then
      begin
        TDataTypeReferenciaOfCosa(DataColumn.DataType).referente:=DataRow;
        self.rd(DataColumn.ColumnName, DataRowAux.LastDataCell.sval);
      end}
      else
      begin
        //TRATAMIENTO ESPECIAL CUANDO HAY INSTANCIAS DE LISTAS (el campo 'n' me indica esto)
        if (LowerCase(DataColumn.ColumnName) = 'n') then
        begin
          n := 0;
          self.rd(DataColumn.ColumnName, n);
          self.EjecutarLectura;
          dataRowAux.LastDataCell.sval := IntToStr(n);
          if (n > 0) then
            cl := True; //VIENE UNA LISTA!
        end
        else
        begin
          self.rd(DataColumn.ColumnName, dataRowAux.LastDataCell._sval);

          if DataColumn.DataType is TDataTypeDoubleOfCosa then
          begin
            self.EjecutarLectura();

            //SE CONVIERTE EL REAL AL FORMATO LOCAL
            b := dataRowAux.LastDataCell._sval;
            if b = '0.87' then
              a := 1;

            ad := StrToFloat(dataRowAux.LastDataCell._sval);
            uauxiliares.setSeparadoresLocales();
            dataRowAux.LastDataCell._sval := FloatToStr(ad);
            b := dataRowAux.LastDataCell._sval;
            uauxiliares.setSeparadoresGlobales();
          end;
          if DataColumn.DataType is TDataTypeDateOfCosa then
          begin
            self.EjecutarLectura();

            //SE CONVIERTE LA FECHA AL FORMATO LOCAL
            adt := StrToDate(dataRowAux.LastDataCell._sval);
            uauxiliares.setSeparadoresLocales();
            dataRowAux.LastDataCell._sval := DateToStr(adt);
            uauxiliares.setSeparadoresGlobales();
          end;
        end;
      end;
    end;
    if (j = Count - 1) then
      //Si se leyó último campo de los definidos en TListaDeCosasConNombre
      if (cl = True) then // Lo que viene ahora son las instancias de la lista
      begin
        //self.EjecutarLectura; //Fuerzo lectura del resto de los campos hasta las instancias de la lista.
        for z := 0 to n - 1 do
        begin
          DataRow2 := nil;
          self.rd(':', DataRow2);

          //ReadDataRowFromText(var DataRow: TDataRowOfCosa; ColumnName: string; cosaStrId: string)

          self.EjecutarLectura;
          DataTable2 := TDataTableOfCosa(DataRow2.DataTableOwner);
          DataRow2.RowName := 'dr_:' + IntToStr(DataTable2.DataRowListCount);

          //CREO EL REGISTRO EN LA TABLA DE LISTAS:
          DataRowOfList := TDataRowOfCosa.Create(DataTableOfList,
            'dr_' + IntToStr(DataTableOfList.DataRowListCount));
          DataRowOfList.SIUD := TSIUD.TSELECT;
          DataRowOfList.AddDataCell(TDataCellOfCosa.Create(IntToStr(LastIdParent)));
          DataRowOfList.AddDataCell(TDataCellOfCosa.Create(
            DataTable2.TableName + '.' + IntToStr(DataRow2._RowId)));
          DataTableOfList.AddDataRow(DataRowOfList);

        end;
        cl := False;
      end;
    j := j + 1;
  end;
  self.EjecutarLectura;


  { SE OBTIENE UNA FILA CONVERTIDA A LA ÚLTIMA VERSIÓN
    --------------------------------------------------
    DataTableAux: es una tabla con la versión del archivo.
    DataTable: es una tabla con la última versión posible (VERSION=-2). }


  (********************************
   VER EL TEMA DE LA VERSION ENTRE DataTable y DataTableAux, así controlamos y
   no entra todo el tiempo. ***************************************************)
  if clase.ClassName = 'TListGenerador' then
    a := 1;

  DataTable := DATA_STORE.GetDataTableByName(DataTableName);
  DataRow := TDataRowOfCosa.Create(DataTable, DataRowAux.RowName);
  DataRow.SIUD := DataRowAux.SIUD;
  DataTable.AddDataRow(DataRow);
  DataTable.ConvertDataRowToNewestVersion(DataRow, DataTableAux);
end;
{$ENDIF}

procedure TArchiTexto.ReadCosaFromText(var cosa: TCosa; cosaStrId: string);
var
  clase: TClaseDeCosa;
begin
  clase := TClaseDeCosa(getClaseOf(CosaStrId));
  if clase = nil then
    raise ExcepcionClaseNoRegistrada.Create(
      'TArchiTexto.ReadCosaFromText: Error, ' + CosaStrId +
      ' No es una clase registrada. Procesando Línea: ' + IntToStr(klinea));

  {$IFNDEF MIGRACION_PERSISTENCIA}
  cosa := clase.Create_ReadFromText(Self);
  {$ENDIF}

{$IFDEF WRLN_COSAS_CON_NOMBRE_LEIDAS}
  if cosa is TCosaConNombre then
    writeln(IntToStr(klinea) + ', Se leyó: ' + cosaStrId + '...' + cosa.apodo);
{$ENDIF}
end;


//-------------------------
// Métodos de TListaDeCosas
//=========================

constructor TListaDeCosas.Create(capa: integer; idCarpeta: string);
begin
  inherited Create(capa);
  self.idCarpeta := idCarpeta;
  lst := TList.Create;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TListaDeCosas.Create_ReadFromText(f: TArchiTexto);
var
  k: integer;
  cosas: array of TCosa;
begin
  inherited Create_ReadFromText(f);
  self.idCarpeta := f.aux_idCarpeta;
  f.IniciarLecturaRetrasada;
  f.rd('n', n);
  f.EjecutarLectura;
  lst := TList.Create;
  SetLength(cosas, n);
  f.IniciarLecturaRetrasada;
  for k := 0 to n - 1 do
    f.rd(':', TCosa(cosas[k]));
  f.EjecutarLectura;

  //Esto es porque puede estar mal grabado el archivo y no se carguen las n cosas
  for k := 0 to n - 1 do
    if cosas[k] <> nil then
      lst.Add(cosas[k]);
  SetLength(cosas, 0);
end;

{$ENDIF}

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure TListaDeCosas.WriteToText(f: TArchiTexto);
var
  n, k: integer;
  cosa: TCosa;
begin
  inherited WriteToText(f);
  n := lst.Count;
  f.wr('n', n);
  for k := 0 to n - 2 do
  begin
    cosa := lst.items[k];
    f.wr(':', cosa);
    f.writeline('');
  end;
  if n > 0 then
  begin
    cosa := lst.items[n - 1];
    f.wr(':', cosa);
  end;
end;

{$ENDIF}

function TListaDeCosas.Add(cosa: TCosa): integer;
begin
  Result := self.lst.Add(cosa);
end;

procedure TListaDeCosas.ponerIgualA(listaDeCosas: TListaDeCosas; freeElementos: boolean);
var
  i: integer;
begin
  if freeElementos then
    ClearFreeElementos
  else
    Clear;

  lst.Capacity := listaDeCosas.lst.Count;
  for i := 0 to listaDeCosas.lst.Count - 1 do
    lst.Add(listaDeCosas.lst[i]);
end;


procedure TListaDeCosas.ActivarCapas(const CapasActivas: TDAOfNInt);
var
  k: integer;
  ac: TCosa;
begin
  k := 0;
  while (k < Count) do
  begin
    ac := items[k];
    if kInArray(CapasActivas, ac.capa) >= 0 then
    begin
      ac.ActivarCapas(CapasActivas);
      Inc(k);
    end
    else
    begin
      ac.Free;
      Delete(k);
    end;
  end;
end;

procedure TListaDeCosas.Free;
var
  k: integer;
begin
  for k := 0 to lst.Count - 1 do
    if lst.Items[k] <> nil then
      TCosa(lst.items[k]).Free;
  lst.Free;
  inherited Free;
end;

procedure TListaDeCosas.FreeSinElemenentos;
begin
  lst.Free;
  inherited Free;
end;

function TListaDeCosas.Count: integer;
begin
  Result := lst.Count;
end;

procedure TListaDeCosas.insert(indice: integer; cosa: TCosa);
begin
  lst.Insert(indice, cosa);
end;

function TListaDeCosas.Remove(cosa: TCosa): integer;
begin
  Result := lst.Remove(cosa);
end;

procedure TListaDeCosas.Delete(indice: integer);
begin
  lst.Delete(indice);
end;

procedure TListaDeCosas.Pack;
begin
  lst.Pack;
end;

procedure TListaDeCosas.Exchange(indice1, indice2: integer);
begin
  lst.Exchange(indice1, indice2);
end;

function TListaDeCosas.IndexOf(cosa: TCosa): integer;
begin
  Result := lst.IndexOf(cosa);
end;

function TListaDeCosas.replace(cosaARemplazar, cosaNueva: TCosa): integer;
var
  i: integer;
begin
  i := 0;
  while (i < Count) and (lst.Items[i] <> cosaARemplazar) do
    i := i + 1;
  if i < Count then
  begin
    lst.Items[i] := cosaNueva;
    Result := i;
  end
  else
    Result := -1;
end;

procedure TListaDeCosas.Clear;
begin
  lst.Clear;
end;

procedure TListaDeCosas.ClearFreeElementos;
var
  i: integer;
begin
  for i := 0 to lst.Count - 1 do
    TCosa(items[i]).Free;
  lst.Clear;
end;

procedure TListaDeCosas.Sort(Compare: TListSortCompare);
begin
  lst.Sort(Compare);
end;

function TListaDeCosas.getList: TList;
begin
  Result := lst;
end;

function TListaDeCosas.toArray: TDAOfTCosa;
var
  i: integer;
  res: TDAOfTCosa;
begin
  SetLength(res, lst.Count);
  for i := 0 to lst.Count - 1 do
    res[i] := lst[i];
  Result := res;
end;

function TListaDeCosas.getItem(i: integer): TCosa;
begin
  Result := lst.items[i];
end;

procedure TListaDeCosas.setItem(i: integer; cosa: TCosa);
begin
  lst.items[i] := cosa;
end;

function TListaDeCosas.getCapacity: integer;
begin
  Result := lst.Capacity;
end;

procedure TListaDeCosas.setCapacity(newCapacity: integer);
begin
  lst.Capacity := newCapacity;
end;

{$IFDEF MIGRACION_PERSISTENCIA}
class function TListaDeCosas.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: integer): TDataColumnListOfCosa;
begin
  Result := inherited CreateDataColumnList(xClaseDeCosa, xVersion);
  Result.AddIntegerColumn('n', 0);
  Result.AddListColumn('lst');
end;
{$ENDIF}


/////////////////////////
//Procedimientos Globales
/////////////////////////

//dfusco@20141023
//Se modifica la función para pasar al registro la lista de definicion de campos
//de la clase
function registrarClaseDeCosa(const nombre: string; Clase: TClaseDeCosa;
  CreateDataTable: boolean): integer;
var
  ficha: TFichaClase;
  {$IFDEF MIGRACION_PERSISTENCIA}
  DataTable: TDataTableOfCosa;
  {$ENDIF}
begin
  {$IFDEF MIGRACION_PERSISTENCIA}
  if Assigned(DATA_STORE.GetDataTableByName('dt_' + Clase.ClassName)) then
    raise Exception.Create('La clase ' + Clase.ClassName +
      ' ya fue registrada, verificar procedimientos AlInicio');
  if CreateDataTable then
  begin
    DataTable := TDataTableOfCosa.Create('dt_' + Clase.ClassName, TClaseDeCosa(Clase));
    DATA_STORE.AddDataTable(DataTable);
  end;
  {$ENDIF}

  ficha := TFichaClase.Create(nombre, clase);
  Result := registro_de_ClasesDeCosas_.lst.Add(ficha);
end;

function getClaseOf(const nombre: string): TCualquierClase;
var
  ipos: integer;
  res: TCualquierClase;
  slow: string;
begin
  res := nil;
  slow := LowerCase(nombre);
  for ipos := 0 to registro_de_ClasesDeCosas_.lst.Count - 1 do
  begin
    //dfusco@20141017
    //Para comparar los string se convierte todo a LowerCase
    if LowerCase(TFichaClase(registro_de_ClasesDeCosas_.lst[ipos]).strClaseId) =
      slow then
    begin
      res := TFichaClase(registro_de_ClasesDeCosas_.lst.items[ipos]).clase;
      break;
    end;
  end;

  if res = nil then
    raise Exception.Create('ucosa.getClaseOf: Clase de Actor desconocida ' + nombre)
  else
    Result := res;
end;

// si la clase no está registrada retorna '<CLASE_NO_REGISTRADA>'
function getIdStrOf(clase: TClaseDeCosa): string;
var
  ipos: integer;
  buscando: boolean;
  b: string;
begin
  buscando := True;
  b := clase.ClassName;
  for ipos := 0 to registro_de_ClasesDeCosas_.lst.Count - 1 do
    if TFichaClase(registro_de_ClasesDeCosas_.lst.items[ipos]).Clase = Clase then
    begin
      buscando := False;
      break;
    end;
  if buscando then
    Result := '<CLASE_NO_REGISTRADA>'
  else
    Result := TFichaClase(registro_de_ClasesDeCosas_.lst.items[ipos]).strClaseId;
end;


// Retorna la cadena de padres separados por #9
function ParentsStrList(clase: TClass): string;
var
  res: string;
  padre: TClass;
begin
  res := clase.ClassName;
  padre := clase;
  while assigned(padre.ClassParent) do
  begin
    padre := padre.ClassParent;
    res := padre.ClassName + '->' + res;
  end;
  Result := res;

end;

function IsChild(Parent, Child: string): boolean;
begin
  Result := IsChild(getClaseOf(Parent), getClaseOf(Child));
end;

function IsChild(Parent, Child: TClass): boolean;
begin
  if not Assigned(Child) then
  begin
    Result := False;
    exit;
  end;
  //  writeln(Parent.ClassName,' = ',Child.ClassName);
  if (Parent.ClassName = Child.ClassParent.ClassName) then
    Result := True
  else if Child.ClassParent.ClassName = TObject.ClassName then
    Result := False
  else
    Result := IsChild(Parent, Child.ClassParent);
end;

procedure LimpiarRegistroDeClasesDeCosas;
var
  k: integer;
begin
  for k := 0 to registro_de_ClasesDeCosas_.lst.Count - 1 do
    TFichaClase(registro_de_ClasesDeCosas_.lst.items[k]).Free;
  registro_de_ClasesDeCosas_.lst.Clear;
end;

procedure AlInicio;
begin
  ReferenciasAArchivosExternos := TList.Create;
  lista_caminos := TBuscaArchivos.Create;

  {$IFDEF MIGRACION_PERSISTENCIA}
  DATA_STORE := TDataStore.Create(nil);
  {$ENDIF}
{$IFDEF CNT_COSAS}
  CantCosas := TCntCosas.Create;
{$ENDIF}
  registro_de_ClasesDeCosas_ := TListaDeCosas.Create(0, 'RegistroClasesDeActor');
  //  registrarClaseDeCosa( TCosa.ClassName, TCosa );
  registrarClaseDeCosa(TListaDeCosas.ClassName, TListaDeCosas);
 {$IFDEF MIGRACION_PERSISTENCIA}
  DATA_STORE.AddDataTable(TDataTableOfListOfCosa.Create);
{$ENDIF}
end;

procedure AlFinal;
begin
  //  LimpiarRegistroDeClasesDeCosas;
  registro_de_ClasesDeCosas_.Free;
{$IFDEF CNT_COSAS}
  CantCosas.Free;
{$ENDIF}
  ReferenciasAArchivosExternos.Free;
  lista_caminos.Free;
end;

procedure FreeAndNil(var obj: TCosa);
var
  aux: TCosa;
begin
  aux := obj;
  obj := nil;
  aux.Free;
end;

function referenciaACosa(cosa: TCosa): string;
begin
  if assigned(cosa) and (cosa is TCosaConNombre) then
    Result := '<' + cosa.ClassName + '.' + TCosaConNombre(cosa).nombre + '>'
  else
    Result := '<?.?>';
end;

constructor TListaDeObjetosPorLeer.Create;
begin
  inherited Create;
end;

procedure TListaDeObjetosPorLeer.Free;
var
  k: integer;
begin
  for k := 0 to Count - 1 do
    TListCamposParaLeer(items[k]).Free;
  self.Clear;
  inherited Free;
end;

function TListaDeObjetosPorLeer.ListaActiva: TListCamposParaLeer;
begin
  if Count > 0 then
    Result := items[Count - 1]
  else
  begin
    raise Exception.Create('Error, no hay un Objeto activo para lectura ');
    Result := nil;
  end;
end;

procedure TListaDeObjetosPorLeer.FreeListaPorLeer;
begin
  if Count > 0 then
  begin
    TListCamposParaLeer(items[Count - 1]).Free;
    Delete(Count - 1);
  end;
end;

procedure TListaDeObjetosPorLeer.InicioObjeto(const NombreClase: string);
var
  a: TListCamposParaLeer;
begin
  a := TListCamposParaLeer.Create(NombreClase);
  add(a);
end;

procedure TListaDeObjetosPorLeer.FinalObjeto;
var
  {$IFDEF MIGRACION_PERSISTENCIA}
  xDataRow: TDataRowOfCosa;
  {$ENDIF}
  xCosa: TCosa;
  nombreDelCampoObjeto: string;
  lpl: TListCamposParaLeer;
  i: integer;
  res: boolean;
begin
  // procesar hasta el final del objeto
  lpl := ListaActiva;
  {$IFDEF MIGRACION_PERSISTENCIA}
  xDataRow := lpl.DataRow;
  {$ENDIF}
  xCosa := lpl.cosa;
  FreeListaPorLeer;
  lpl := ListaActiva;
  nombreDelCampoObjeto := lowercase(lpl.xNombre);
  res := False;
  for i := 0 to lpl.Count - 1 do
    if lowercase(TCampoParaLeer(lpl[i]).nombre) = nombreDelCampoObjeto then
    begin

      {$IFDEF MIGRACION_PERSISTENCIA}
      if xDataRow <> nil then
        TDataRowOfCosa(TCampoParaLeer(lpl.items[i]).pVal^) := xDataRow;
      {$ENDIF}
      if xCosa <> nil then
        TCosa(TCampoParaLeer(lpl.items[i]).pVal^) := xCosa;
      {$IFDEF MIGRACION_PERSISTENCIA}
      if (xDataRow = nil) and (xCosa = nil) then
        raise Exception.Create(
          'TListaDeObjetosPorLeer.FinalObjeto: xDataRow=xCosa=nil');
      {$ELSE}
      if xCosa = nil then
        raise Exception.Create(
          'TListaDeObjetosPorLeer.FinalObjeto: xCosa=nil');
      {$ENDIF}
      TCampoParaLeer(lpl.items[i]).Free;
      lpl.Delete(i);
      res := True;
      break;
    end;
  if not res then
    raise Exception.Create('Aparentemente el objeto: ' +
      lpl.IdentificadorDeLaClaseDelObjeto + ' no tienen definido el campo: ' +
      nombreDelCampoObjeto);
end;


constructor TArchiTexto.CreateForRead(const nombreArchivo: string;
  abortarEnError: boolean);
var
  s: string;
  iposver: integer;
begin
  setSeparadoresGlobales;
  if FileExists(nombreArchivo) then
  begin
    inherited Create;

    f_abierto := False;
    unArchivoPorCosa := False;
    padre := nil;
    ramas := nil;


    Self.NombreArchivo := nombreArchivo;
    oldFileMode := FileMode;
    filemode := fmOpenRead;

    assignfile(f, NombreARchivo);
    {$I-}
    reset(f);
    {$I+}
    if ioresult = 0 then
      f_abierto := True
    else
      raise Exception.Create('TArchiTexto.CreateForRead: No pude abrir archivo: ' +
        NombreArchivo);

    self.abortarEnError := abortarEnError;

    readln(f, s);
    s := trim(s);

    klinea := 0;

    iposver := pos('VERSION=', s);
    if iposver >= 1 then
    begin
      Delete(s, 1, iposver - 1 + length('VERSION='));
      s := trim(s);
      version := StrToInt(s);
      Inc(klinea);
    end
    else
      version := 0;

    // parche chancho para poder leer archivos de la gente de la regasificadora
    if (version < 10000) and (version > VERSION_ArchiTexto) then
      raise Exception.Create(rs_ElArchivo + ' ' + nombreArchivo +
        ' ' + rs_VersionPosterior);


    if version <= 45 then
      setSeparadoresGlobales_FechaUruguaya_V45_;

    listaObjetosPorLeer := TListaDeObjetosPorLeer.Create;
    indentstr := '';
    listaObjetosPorLeer.InicioObjeto('TODO');
  end
  else
    raise Exception.Create('TArchiTexto.CreateForRead: ' +
      rs_NoSeEncuentraArchivo + nombreArchivo);
end;



constructor TArchiTexto.CreateForWrite(const nombreArchivo: string;
  crearBackupSiExiste: boolean; maxNBackups: integer);
begin
  inherited Create;
  setSeparadoresGlobales;

  unArchivoPorCosa := False;
  padre := nil;
  ramas := nil;

  f_abierto := False;
  Self.NombreArchivo := NombreArchivo;

  if crearBackupSiExiste and FileExists(nombreArchivo) then
    uconstantesSimSEE.backupearArchivoAntesDeSalvar(nombreArchivo, maxNBackups);
  assignFile(f, NombreArchivo);
  oldFileMode := FileMode;
  filemode := fmOpenWrite;
   {$I-}
  Rewrite(f);
   {$I+}
  if ioresult = 0 then
    f_abierto := True
  else
    raise Exception.Create('TArchiTexto.CreateForWrite(' + nombreArchivo + ')');

  writeln(f, 'VERSION=' + IntToStr(VERSION_ArchiTexto));
  klinea := 1;
  listaObjetosPorLeer := TListaDeObjetosPorLeer.Create;
end;

function TArchiTexto.CreateRamaForWrite(
  const Carpeta, NombreClase, NombreCosa: string): TArchiTexto;
var
  s: string;
  i: integer;
  a: TArchiTexto;
begin
  if ramas = nil then
    ramas := TList.Create;
  i := pos('.sse', nombreArchivo);
  s := copy(nombreArchivo, 1, i - 1);
  if carpeta <> '' then
    s := s + '/' + carpeta + '/';

  s := s + '$' + NombreClase + '$' + NombreCosa + '.sse';

  a := TArchiTexto.CreateForWrite(s, False, 0);
  a.unArchivoPorCosa := True;
  ramas.Add(a);
  Result := a;
end;


function TArchiTexto.CreateRamaForRead(
  const carpeta, NombreClase, NombreCosa: string): TArchiTexto;
var
  s: string;
  i: integer;
  a: TArchiTexto;
begin
  if ramas = nil then
    ramas := TList.Create;
  i := pos('.sse', nombreArchivo);
  s := copy(nombreArchivo, 1, i - 1);
  if carpeta <> '' then
    s := s + '/' + carpeta + '/';

  s := s + '$' + NombreClase + '$' + NombreCosa + '.sse';
  a := TArchiTexto.CreateForRead(s, self.abortarEnError);
  a.unArchivoPorCosa := True;
  ramas.Add(a);
  Result := a;
end;


procedure TArchiTexto.Free;
var
  i: integer;
  //  a: TArchiTexto;
begin
  setSeparadoresLocales;

  if listaObjetosPorLeer <> nil then
    listaObjetosPorLeer.Free;

  if ramas <> nil then
  begin
    if ramas.Count > 0 then
      raise Exception.Create(
        'Toy tratando de FREE un ArchiTexto pero tiene RAMAS, revisar esto')
    else
      ramas.Free;
(* creo que no es necesario esto, si lo de arriba salta, talvez tenga que
   usar lo de abajo
    for i:= 0 to ramas.Count - 1 do
    begin
      a:= ramas.Items[i];
      a.padre:= self.padre;
      padre.ramas.Add( a );
    end;
*)
  end;

  if padre <> nil then
  begin
    // quitamos este archivo de la lista del padre
    i := padre.ramas.IndexOf(self);
    padre.ramas.Delete(i);
  end;

  if (f_abierto) then
  begin
    f_abierto := False;
    closefile(f);
  end;
  FileMode := oldFileMode;


  inherited Free;
end;

function TArchiTexto.NextLine: string;
var
  r: string;
begin
  if not system.EOF(f) then
  begin
    readln(f, r);
    Inc(klinea);
    Result := trim(r);
  end
  else
    raise Exception.Create('Llegué al fin del archivo en forma inesperada.');
end;

procedure TArchiTexto.writeline(const s: string);
begin
  writeln(f, indentstr + s);
end;

function TArchiTexto.EOF: boolean;
begin
  Result := system.EOF(f);
end;

function TArchiTexto.quedanCosasDelObjetoActualPorLeer(claseObjeto: string): boolean;
var
  s: string;
begin
  s := NextLine;
  //Solo puede haber lineas vacias, sino devuelvo true
  while (s <> '<-' + claseObjeto + '>;') and (trim(s) = '') do
    s := NextLine;
  if s = '<-' + claseObjeto + '>;' then
    Result := False
  else
    Result := True;
end;

function TArchiTexto.ListaPorLeer: TListCamposParaLeer;
begin
  Result := listaObjetosPorLeer.ListaActiva;
end;

procedure TArchiTexto.wr(const nombre: string; const valor: NReal;
  precision: integer; decimales: integer);
begin
  writeline(nombre + '= ' + FloatToStrF(valor, ffGeneral, precision, decimales) + ';');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: NReal);
begin
  wr(nombre, valor, CF_PRECISION, CF_DECIMALES);
end;

procedure TArchiTexto.wr(const nombre: string; const valor: string);
begin
  writeline(nombre + '= ' + valor + ';');
end;

procedure TArchiTexto.wrArchRef(const nombre: string; const valor: TArchiRef);
begin
  if valor <> nil then
    writeline(nombre + '= ' + valor.archiRefStr + ';')
  else
    writeline(nombre + '= ;');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: NInt);
begin
  writeline(nombre + '= ' + IntToStr(valor) + ';');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: cardinal);
begin
  writeline(nombre + '= ' + IntToStr(valor) + ';');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: TFecha);
begin
  if valor <> nil then
    writeline(nombre + '= ' + valor.AsStr + ';')
  else
    writeline(nombre + '= 0;');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: TCosa);
var
  CosaStrId: string;
  clase: TClaseDeCosa;
  b: string;
begin
  if valor <> nil then
  begin
    b := valor.ClassName;
    clase := TClaseDeCosa(valor.ClassType);
    CosaStrId := getIdStrOf(clase);
    writeline(nombre + '= <+' + CosaStrId + '>');
    indentstr := indentstr + '  ';
    {$IFNDEF MIGRACION_PERSISTENCIA}
    valor.WriteToText(Self);
    {$ENDIF}
    Delete(indentstr, length(indentstr) - 1, 2);
    writeline('<-' + CosaStrId + '>;');
  end
  else
    writeline(nombre + '= NIL;');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: boolean);
begin
  if valor then
    writeline(nombre + '= 1;')
  else
    writeline(nombre + '= 0;');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: TDAOfNReal;
  precision: integer; decimales: integer);
var
  k: integer;
begin
  Write(f, indentstr + nombre, '= [', length(valor), '| ');
  if length(valor) > 0 then
  begin
    Write(f, FloatToStrF(valor[0], ffFixed, precision, decimales));
    for k := 1 to high(valor) do
    begin
      if ((k mod 12) = 0) then
        writeline('  ');
      Write(f, ', ', FloatToStrF(valor[k], ffFixed, precision, decimales));
    end;
  end;
  writeln(f, '];');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: TDAOfNReal);
begin
  wr(nombre, valor, 12, 3);
end;

procedure TArchiTexto.wr(const nombre: string; const valor: TDAOfNInt);
var
  k: integer;
begin
  Write(f, indentstr + nombre, '= [', length(valor), '| ');
  if length(valor) > 0 then
  begin
    Write(f, valor[0]);
    for k := 1 to high(valor) do
    begin
      if ((k mod 12) = 0) then
        writeline('  ');
      Write(f, ', ', valor[k]);
    end;
  end;
  writeln(f, '];');
end;

procedure TArchiTexto.wr(const nombre: string; const valor: TDAOfBoolean); overload;
var
  k: integer;
begin
  Write(f, indentstr + nombre, '= [', length(valor), '| ');
  if length(valor) > 0 then
  begin
    Write(f, valor[0]);
    for k := 1 to high(valor) do
    begin
      if ((k mod 12) = 0) then
        writeline('  ');
      if Valor[k] then
        Write(f, ', ', valor[k])
      else
        Write(f, ', ', valor[k]);
    end;
  end;
  writeln(f, '];');
end;


procedure TArchiTexto.wr(const nombre: string; const valor: TStringList);
var
  k: integer;
begin
  Write(f, indentstr + nombre, '= [', valor.Count, '| ');
  if valor.Count > 0 then
  begin
    Write(f, valor[0]);
    for k := 1 to valor.Count - 1 do
    begin
      if ((k mod 12) = 0) then
        writeline('  ');
      Write(f, ', ', valor[k]);
    end;
  end;
  writeln(f, '];');
end;


procedure TArchiTexto.wr(const nombre: string; const valor: TDAOfString);
var
  k: integer;
begin
  Write(f, indentstr + nombre, '= [', length(valor), '| ');
  if length(valor) > 0 then
  begin
    Write(f, valor[0]);
    for k := 1 to high(valor) do
    begin
      if ((k mod 12) = 0) then
        writeline('  ');
      Write(f, ', ', valor[k]);
    end;
  end;
  writeln(f, '];');
end;



procedure TArchiTexto.wrReferencia(const nombre: string; const valor: TCosa);
begin
  writeline(nombre + '= ' + referenciaACosa(valor) + ';');
{  if assigned(valor) and (valor is TCosaConNombre) then
    writeline(nombre + '= <' + valor.ClassName + '.' +
      TCosaConNombre(valor).nombre + '>;')
(* rch0707121651 comento esto para que si está mal escriba referencia vacía y listo
    else
      raise Exception.Create('TArchiTexto.wrReferencia: intenta escribir una referencia a algo que no es TCosaConNombre');
  end
*)
  else
    writeline(nombre + '= <?.?>;');}
end;

procedure TArchiTexto.rd(const nombre: string; var valor: NReal);
begin
  ListaPorLeer.Esperar_NReal(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: string);
var
  a: integer;
begin
  ListaPorLeer.Esperar_String(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rdArchRef(const nombre: string; var valor: TArchiRef);
begin
  ListaPorLeer.Esperar_ArchRef(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: NInt);
begin
  ListaPorLeer.Esperar_NInt(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: cardinal);
begin
  ListaPorLeer.Esperar_NCardinal(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: TFecha);
begin
  ListaPorLeer.Esperar_Fecha(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

// flgError= 0 lo mismo que sin flg. flgError= 1 no avisa del error
procedure TArchiTexto.rd(const nombre: string; var valor: TFecha; flgError: integer);
begin
  ListaPorLeer.Esperar_Fecha(nombre, valor, nil, flgError);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;


procedure TArchiTexto.rd(const nombre: string; var valor: boolean);
begin
  ListaPorLeer.Esperar_Boolean(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: TDAOfNReal);
begin
  ListaPorLeer.Esperar_DAOfNReal(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: TDAOfBoolean); overload;
begin
  ListaPorLeer.Esperar_DAOfBoolean(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

{$IFDEF MIGRACION_PERSISTENCIA}
procedure TArchiTexto.rd(const nombre: string; var valor: TDataRowOfCosa);
begin
  ListaPorLeer.Esperar_TDataRowOfCosa(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;
{$ENDIF}

procedure TArchiTexto.rd(const nombre: string; var valor: TCosa);
begin
  ListaPorLeer.Esperar_Cosa(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: TDAOfNInt);
begin
  ListaPorLeer.Esperar_DAOfNInt(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: TDAOfString);
begin
  ListaPorLeer.Esperar_DAOfString(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.rd(const nombre: string; var valor: TStringList);
begin
  ListaPorLeer.Esperar_StringList(nombre, valor, nil);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;



procedure TArchiTexto.rdReferencia(const nombre: string; var valor: TCosa;
  referente: TCosa);
begin
  ListaPorLeer.Esperar_Referencia(nombre, valor, nil, referente);
  if not ListaPorLeer.lecturaRetrasada then
    EjecutarLectura;
end;

procedure TArchiTexto.IniciarLecturaRetrasada;
begin
  if ListaPorLeer <> nil then
    ListaPorLeer.lecturaRetrasada := True;
end;

procedure TArchiTexto.EjecutarLectura;
var
  s, NombreClaseSiguienteObjeto: string;
  resObjetoHijo, resObjetoPadre: TResultadoProcesoLinea;
  lpl: TListCamposParaLeer;
{$IFDEF DBG}
  dbgAux: TDAofString;
{$ENDIF}
  indiceDeLinea: integer;
  posIgual: integer;
  ColumnName: string;
  i: integer;
  a: integer;
begin
  ColumnName := '';
  NombreClaseSiguienteObjeto := '';
  try
    lpl := ListaPorLeer;
    resObjetoPadre := CTPLR_CONTINUAR;
    while (resObjetoPadre <> CTPLR_FIN_OBJETO) and (lpl <> nil) and
      (lpl.Count > 0) and not EOF do
    begin
      {$IFDEF DBG}
      dbgAux := lpl.nombresCamposPorLeer;
      {$ENDIF}

      indiceDeLinea := klinea;
      s := nextline;
 {$IFDEF LECTURA_VERBOSA}
      system.writeln(indiceDeLinea: 6, ': ', s);
 {$ENDIF}
      try
        resObjetoPadre := lpl.ProcesarLinea(klinea, s, NombreClaseSiguienteObjeto,
          ColumnName);
      except
        on E: ExcepcionNoSePidioElCampo do
        begin
          lpl.Ejecutar_vpds(klinea);
          if quedanCosasDelObjetoActualPorLeer(lpl.IdentificadorDeLaClaseDelObjeto) then
          begin
            resObjetoPadre := CTPLR_CONTINUAR;
            klinea := indiceDeLinea;
          end
          else
            resObjetoPadre := CTPLR_FIN_OBJETO;
        end;
        on E: ExcepcionResolverCampoValor do
        begin
          posIgual := pos('=', s);
          lpl.Ejecutar_vpd(klinea, copy(s, 0, posIgual - 1));
          if quedanCosasDelObjetoActualPorLeer(lpl.IdentificadorDeLaClaseDelObjeto) then
          begin
            resObjetoPadre := CTPLR_CONTINUAR;
            klinea := indiceDeLinea + 1;
            //El +1 debería ser el número de lineas que leyo y no le sirvieron
          end
          else
            resObjetoPadre := CTPLR_FIN_OBJETO;
        end
      end;
      if resObjetoPadre = CTPLR_ABRIR_NUEVO_OBJETO then
      begin
        listaObjetosPorLeer.InicioObjeto(NombreClaseSiguienteObjeto);
        lpl := ListaPorLeer;
        try
          {$IFDEF MIGRACION_PERSISTENCIA}
          if Version >= 10113 then
            ReadDataRowFromText(lpl.DataRow, ColumnName, NombreClaseSiguienteObjeto)
          else
            ReadCosaFromText(lpl.cosa, NombreClaseSiguienteObjeto);
          {$ELSE}
          ReadCosaFromText(lpl.cosa, NombreClaseSiguienteObjeto);
          {$ENDIF}

        except
          on E: Exception do
          begin
            {$IFDEF MIGRACION_PERSISTENCIA}
            lpl.DataRow := nil;
            {$ENDIF}
            lpl.cosa := nil;
            if Assigned(procMsgErrorLectura) then
              procMsgErrorLectura(E.Message + '. Leyendo el actor en la linea ' +
                IntToStr(klinea));
            if abortarEnError then
              raise;

            if not EOF then
              repeat
                s := NextLine;
              until (s = '<-' + NombreClaseSiguienteObjeto + '>;') or EOF;
            lpl.leiFinDeObjeto := True;
          end
        end;
        if not lpl.leiFinDeObjeto then
          repeat
            s := nextline;
            resObjetoHijo := lpl.ProcesarLinea(klinea, s,
              NombreClaseSiguienteObjeto, ColumnName);
          until resObjetoHijo = CTPLR_FIN_OBJETO;

        listaObjetosPorLeer.FinalObjeto;
        lpl := ListaPorLeer;
      end;
    end;
    if resObjetoPadre = CTPLR_FIN_OBJETO then
      ListaPorLeer.leiFinDeObjeto := True;
  except
    on E: Exception do
      if pos('kLinea:', E.Message) = 0 then
        raise Exception.Create('kLinea:(' + IntToStr(klinea) + ') ' + E.Message)
      else
        raise;
  end; // try
  //ListaPorLeer.lecturaRetrasada:= false;
end;

function TipoToStr(tipo: TTipoCampo): string;
begin
  case Tipo of
    CTC_NReal: Result := 'NReal';
    CTC_String: Result := 'String';
    CTC_ArchRef: Result := 'Ref. Archivo Externo';
    CTC_NInt: Result := 'Entero';
    CTC_NCardinal: Result := 'Entero Positivo';
    CTC_Boolean: Result := 'Booleano';
    CTC_DAOfNReal: Result := 'Array dinámico de Reales';
    CTC_DAOfNInt: Result := 'Array dinámico de Enteros';
    CTC_DAOfBoolean: Result := 'Array dinámico de Booleanas';
    CTC_Fecha: Result := 'Fecha';
    CTC_TDataRowOfCosa: Result := 'TDataRowOfCosa';
    CTC_Cosa: Result := 'Cosa';
  end; // case
end;

procedure ValorPorDefectoNulo(const nombre: string; pVal: Pointer;
  tipo: TTipoCampo; const klinea: integer);
var
  Msg: string;
begin
  case Tipo of
    CTC_NReal: NReal(pVal^) := 0;
    CTC_String: string(pVal^) := '';
    CTC_ArchRef: TArchiRef(pVal^) := TArchiRef.Create('');
    CTC_NInt: NEntero(pVal^) := 0;
    CTC_NCardinal: cardinal(pVal^) := 0;
    CTC_Boolean: boolean(pVal^) := False;
    CTC_DAOfNReal: setlength(TDAOfNReal(pVal^), 0);
    CTC_DAOfNInt: setlength(TDAOfNInt(pVal^), 0);
    CTC_DAOfBoolean: setlength(TDAOfBoolean(pVal^), 0);
    CTC_Fecha: TFecha(pVal^) := TFecha.Create_dt(0);
    CTC_TDataRowOfCosa: pointer(pVal^) := nil;
    CTC_Cosa: pointer(pVal^) := nil;
    CTC_Referencia: Pointer(pVal^) := nil;
  end; // case


  if global_flgError = 0 then
    if assigned(procMsgValorPorDefecto) then
    begin
      msg := 'Asignando valor por defecto al campo ' + nombre;
      if klinea <> -1 then
        msg := msg + ' leyendo el final del objeto en la linea ' + IntToStr(klinea)
      else
        msg := msg + ' en el final del archivo.';
      procMsgValorPorDefecto(msg);
    end;
end;

procedure ValorPorDefecto_ERROR(const nombre: string; pVal: Pointer;
  tipo: TTipoCampo; const klinea: integer);
var
  msg: string;
begin
  msg := 'No se encontró una definición para el campo: ' + nombre +
    ' tipo: ' + TipoToStr(tipo);
  if klinea <> -1 then
    msg := msg + ' leyendo el final del objeto en la linea ' + IntToStr(klinea)
  else
    msg := msg + ' en el final del archivo.';

  //  raise Exception.Create(msg);

{$IFDEF FPC}
  writeln(msg);
  Write('... presione ENTER para continuar.');
  system.readln;
{$ELSE}
  ShowMessage(msg);
{$ENDIF}
end;

constructor TCampoParaLeer.Create(const Nombre: string; pVal: Pointer;
  Tipo: TTIpoCampo; proc_vpd: TProcFijarValorPorDefecto);
begin
  Create(Nombre, pVal, Tipo, proc_vpd, nil);
end;

constructor TCampoParaLeer.Create(const Nombre: string; pVal: Pointer;
  Tipo: TTIpoCampo; proc_vpd: TProcFijarValorPorDefecto; flgError: integer);
begin
  Create(Nombre, pVal, Tipo, proc_vpd, nil);
  self.flgError := flgError;
end;


constructor TCampoParaLeer.Create(const Nombre: string; pVal: Pointer;
  Tipo: TTipoCampo; proc_vpd: TProcFijarValorPorDefecto; referente: TCosa);
begin
  inherited Create;
  self.flgError := 0;
  self.nombre := nombre;
  self.tipo := tipo;
  self.pVal := pVal;
  self.referente := referente;

{$IFDEF PORDEFECTO_NULO}
  if not assigned(proc_vpd) then
    self.proc_vpd := ValorPorDefectoNulo
  else
    self.proc_vpd := proc_vpd;
{$ELSE}
  self.proc_vpd := proc_vpd;
{$ENDIF}
end;

procedure TCampoParaLeer.Ejecutar_vpd(const klinea: integer);
begin
  if assigned(proc_vpd) then
  begin
    global_flgError := Self.flgError;
    proc_vpd(nombre, pVal, tipo, klinea);
  end
  else
  if flgError = 0 then
    ValorPorDefecto_ERROR(nombre, pVal, tipo, klinea);
end;

constructor TListCamposParaLeer.Create(IdentificadorDeLaClaseDelObjeto: string);
begin
  inherited Create;
  self.IdentificadorDeLaClaseDelObjeto := IdentificadorDeLaClaseDelObjeto;
  lineaAbierta := False;
  xNombre := '';
  xval := '';
  lecturaRetrasada := False;
  leiFinDeObjeto := False;

  {$IFDEF MIGRACION_PERSISTENCIA}
  self.DataRow := nil;
  {$ENDIF}
  self.cosa := nil;
end;

procedure TListCamposParaLeer.Esperar_NReal(const Nombre: string;
  var Variable: NReal; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_NReal, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_String(const Nombre: string;
  var Variable: string; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_String, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_ArchRef(const Nombre: string;
  var Variable: TArchiRef; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_ArchRef, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_NInt(const Nombre: string;
  var Variable: integer; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_NInt, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_NCardinal(const Nombre: string;
  var Variable: cardinal; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_NCardinal, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_Boolean(const Nombre: string;
  var Variable: boolean; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_Boolean, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_DAOfNReal(const Nombre: string;
  var Variable: TDAOfNReal; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_DAOfNReal, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_DAOfNInt(const Nombre: string;
  var Variable: TDAOfNInt; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_DAOfNInt, proc_vpd);
  add(a);
end;


procedure TListCamposParaLeer.Esperar_DAOfBoolean(const Nombre: string;
  var Variable: TDAOfBoolean; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_DAOfBoolean, proc_vpd);
  add(a);
end;


procedure TListCamposParaLeer.Esperar_DAOfString(const Nombre: string;
  var Variable: TDAOfString; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_DAOfString, proc_vpd);
  add(a);
end;



procedure TListCamposParaLeer.Esperar_StringList(const Nombre: string;
  var Variable: TStringList; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_StringList, proc_vpd);
  add(a);
end;

{$IFDEF MIGRACION_PERSISTENCIA}
procedure TListCamposParaLeer.Esperar_TDataRowOfCosa(const nombre: string;
  var variable: TDataRowOfCosa; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_TDataRowOfCosa, proc_vpd);
  add(a);
end;
{$ENDIF}

procedure TListCamposParaLeer.Esperar_Cosa(const Nombre: string;
  var Variable: TCosa; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_Cosa, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_Referencia(const Nombre: string;
  var Variable: TCosa; proc_vpd: TProcFijarValorPorDefecto; referente: TCosa);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_Referencia, proc_vpd, referente);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_Fecha(const Nombre: string;
  var Variable: TFecha; proc_vpd: TProcFijarValorPorDefecto);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_Fecha, proc_vpd);
  add(a);
end;

procedure TListCamposParaLeer.Esperar_Fecha(const Nombre: string;
  var Variable: TFecha; proc_vpd: TProcFijarValorPorDefecto; flgError: integer);
var
  a: TCampoParaLeer;
begin
  a := TCampoParaLeer.Create(Nombre, @Variable, CTC_Fecha, proc_vpd, flgError);
  add(a);
end;

(*
    CTPLR_CONTINUAR,
    CTPLR_FIN_OBJETO,
    CTPLR_ABRIR_NUEVO_OBJETO );
*)
function TListCamposParaLeer.ProcesarLinea(klinea: integer; linea: string;
  var extres: string; var ColumnName: string): TResultadoProcesoLinea;
var
  NombreClase: string;
begin
  ColumnName:= '';
  Result := CTPLR_CONTINUAR;
  if pos('<-', linea) = 1 then
  begin
    if IdentificadorDeLaClaseDelObjeto <> '' then
      if pos(IdentificadorDeLaClaseDelObjeto, linea) <> 3 then
        raise Exception.Create('Error, cierre de objeto : ' + linea +
          ' Se esperaba: ' + '<-' + IdentificadorDeLaClaseDelObjeto +
          '>, en línea: ' + IntToStr(klinea));

    Ejecutar_vpds(klinea);      // procesamos los no resueltos
    Result := CTPLR_FIN_OBJETO; // cierro esta lista FIN DEL OBJETO
  end
  else
  begin
    if LineaAbierta then
    begin
      xval := xval + ' ' + linea;
      if xval[length(xval)] = ';' then
      begin
        xval := copy(xval, 0, Length(xval) - 1);//Saco el ;
        ResolverCampoValor(xNombre, xVal);
        LineaAbierta := False;
      end;
    end
    else
    begin
      if getPalHastaSep(xnombre, linea, '=') then
      begin
        xval := trim(linea);
        if length(xval) > 0 then
        begin
          if pos('<+', xval) = 1 then
          begin
            system.Delete(xval, 1, 2);
            if xval[length(xval)] <> '>' then
              raise Exception.Create('Error, falta el > de cierre en fin de clase : '
                + xval);
            NombreClase := copy(xval, 1, length(xval) - 1);
            ColumnName := xNombre;
            extres := NombreClase;
            Result := CTPLR_ABRIR_NUEVO_OBJETO;
          end
          else
          if xval[length(xval)] = ';' then
            //if getPalHastaSep(xval, linea, ';') then
          begin
            xval := copy(xval, 0, Length(xval) - 1);//Saco el ;
            try
              ResolverCampoValor(xNombre, xVal)
            except
              on E: ExcepcionNoSePidioElCampo do
              begin
                if Assigned(procMsgAdvertenciaLectura) then
                  procMsgAdvertenciaLectura(E.Message);
              end;
              on Exception do
              begin
                raise;
              end;
            end;
          end
          else
            LineaAbierta := True;
        end
        else
          LineaAbierta := True;
      end
      else
        Exception.Create('Error, no encontré el (=) en la línea: [' +
          IntToStr(klinea) + '] :' + linea);
    end;
  end;
end;

procedure TListCamposParaLeer.Free;
var
  k: integer;
begin
  Ejecutar_vpds(-1);
  //PAlfaro@101104 0307
  //Ojo, estaba comentado, pero me parece que tiene que ir. No se tranca y sin esto
  //quedarían colgados todos los campos para leer de cada TCosa
  for k := 0 to Count - 1 do
    TCampoParaLeer(items[k]).Free;
  inherited Free;
end;

procedure TListCamposParaLeer.ResolverCampoValor(const xNombre, xVal: string);
var
  buscando: boolean;
  k: integer;
  clase, nombreRef: string;
  lc_xNombre, lct: string;
  ts: string;
begin
  buscando := True;
  k := 0;
  lc_xNombre := lowercase(xNombre);
  while buscando and (k < Count) do
  begin
    lct := lowercase(TCampoParaLeer(items[k]).nombre);
    if lct = lc_xNombre then
      buscando := False
    else
      Inc(k);
  end;

  if buscando then
    raise ExcepcionNoSePidioElCampo.Create('Se leyó el campo: ' +
      xnombre + ' pero no se esperaba en el objeto: ' +
      IdentificadorDeLaClaseDelObjeto +
      '. Pruebe guardar y cargar nuevamente la sala. Intentaré solucionar el problema.')
  else
  begin
    try
      with  TCampoParaLeer(items[k]) do
        case Tipo of
          CTC_NReal: NReal(pVal^) := StrToFloat(xval);
          CTC_String: string(pVal^) := xval;
          CTC_ArchRef:
          begin
            ts := xval;
             {$IFDEF LINUX}
            toLinuxDir(ts);
             {$ENDIF}
            TArchiRef(pVal^) := TArchiRef.Create(ts);
          end;
          CTC_NInt: NEntero(pVal^) := StrToInt(xval);
          CTC_NCardinal: cardinal(pVal^) := StrToInt(xval);
          CTC_Boolean: boolean(pVal^) := (xval <> '') and (xval[1] = '1');
          CTC_DAOfNReal: parseDAOfNreal(TDAOfNReal(pVal^), xval);
          CTC_DAOfNInt: parseDAOfNInt(TDAOfNInt(pVal^), xval);
          CTC_DAOfBoolean: parseDAOfBoolean(TDAOfBoolean(pVal^), xval);
          CTC_DAOfString: parseDAOfString(TDAOfString(pVal^), xval);
          CTC_StringList: parseStringList(TStringList(pVal^), xval);
          CTC_Fecha: TFecha(pVal^) := TFecha.Create_str(xval);
          CTC_Referencia:
          begin
            nombreRef := '';
            clase := '';
            parsearReferencia(xval, clase, nombreRef);
            if (clase <> '?') then
              uCosaConNombre.registrar_referencia(referente,
                clase, nombreRef, pVal^)
            else
              TCosa(pVal^) := nil;
          end;
          {$IFDEF MIGRACION_PERSISTENCIA}
          CTC_TDataRowOfCosa:
          begin
            TDataRowOfCosa(pval^) := nil;
          end;
          {$ENDIF}
          CTC_Cosa:
          begin
            TCosa(pval^) := nil;
          end;
        end; // case
    except
      on E: Exception do
      begin
        raise ExcepcionResolverCampoValor.Create('ResolverCampoValor(' +
          xnombre + ',' + xval + ') ' + E.Message);
      end;
    end; // try
    // eliminamos la ficha de lectura pendiente pues fué resuelta
    TCampoParaLeer(items[k]).Free;
    Delete(k);
  end;
end;

// recorremos las fichas que hallan quedado sin resolver y les asignamos los
// valores por defecto (o tiramos excepciones si no se definió una función
// alternativa.
procedure TListCamposParaLeer.Ejecutar_vpds(klinea: integer);
var
  k: integer;
begin
  for k := 0 to Count - 1 do
  begin
    TCampoParaLeer(items[k]).Ejecutar_vpd(klinea);
    TCampoParaLeer(items[k]).Free;
  end;
  self.Clear;
end;

procedure TListCamposParaLeer.Ejecutar_vpd(klinea: integer; nombreCampo: string);
var
  k: integer;
begin
  for k := 0 to Count - 1 do
  begin
    if TCampoParaLeer(items[k]).nombre = nombreCampo then
    begin
      TCampoParaLeer(items[k]).Ejecutar_vpd(klinea);
      TCampoParaLeer(items[k]).Free;
      Delete(k);
      Pack;
      break;
    end;
  end;
end;

{$IFDEF DBG}
function TListCamposParaLeer.nombresCamposPorLeer: TDAofString;
var
  i: integer;
  res: TDAofString;
begin
  SetLength(res, self.Count);
  for i := 0 to Self.Count - 1 do
    res[i] := TCampoParaLeer(items[i]).nombre;
  Result := res;
end;

{$ENDIF}


begin
  ReferenciasAArchivosExternos := nil;
  lista_caminos := nil;
end.
