unit ucurvasaeros;

{$mode delphi}

interface

uses
  Classes, SysUtils, xmatdefs, matreal,uro_aire,
  {$IFNDEF SIN_DBCON}
  {$IFDEF pq_direct}
  udbconpg,
  {$ELSE}
  urosx,
  {$ENDIF}
  {$ENDIF}
  uauxiliares;

type
  {$IFNDEF SIN_DBCON}
  {$IFNDEF pq_direct}
  TDBPQCon = TDBrosxCon;
  {$ENDIF}
  {$ENDIF}

  TCurvaAerogen = class
    nid: integer;
    modelo: string;
    Diametro: NReal;
    NPuntos: integer;
    curva_v, curva_P: TVectR;

    curva_pv3: TVectR;
    pv3min, pv3max: NReal;
    PMax: NReal; // máximo de la curva
    k_Pmax: integer; // indice de la curva en el que alcanza el PMax.
    v_Pmax, pv3_Pmax: NReal; // velocidad y potencia de viento para el máximo de Potencia.
    k_InicioGen: Integer; // indice en la curva donde comienza la generación.
    v_InicioGen, pv3_InicioGen: NReal; // valores a los cuales la generación es >= 1% de PMax.

    AreaRotor: NReal;

    // coeficientes de ajuste de una curva S del tipo P = Pmax / (1+ exp( -alfa * (v -vm ))
    // La curva se ajusta para valores entre pv3_min y la pv de la potencia máxima
    // o sea que la eventa
    curva_s_alfa, curva_s_vm: NReal;
    curva_s_errc2: NReal; // error cuadrático medio
    curva_s_cntMuestras: integer;

    constructor CreateLoadFromArchi(modelo, archi_curva: string);
    constructor CreateFromStr(strDef: string);

    constructor CreateLoadFromFile(var f: textfile);
    constructor Create_Clone( aCurvaAerogen: TCurvaAeroGen );

    procedure StoreInFile(var f: textfile);

    function P_MW(pv3: NReal): NReal;
    function pv3_of_P_MW( P_MW: NReal): NReal;

    function P_MW_Of_v(v: NReal): NReal;
    function v_of_P_MW( P_MW: NReal): NReal;



    procedure Free;

    {$IFNDEF SIN_DBCON}    class procedure CrearTablaDB(db: TDBPQCon);
    constructor CreateLoadFromDB(r: TSQLQuery);
    constructor CreateLoadFromDB_nid(db: TDBPQCon; nidcurva: integer);
    function InsertIntoDB(db: TDBPQCon): integer;
    procedure UpdateDB(db: TDBPQCon);
    {$ENDIF}

    // sirve para crear la curva de un parque por suma de curvas
    procedure SumarCurvas( Curvas: array of TCurvaAerogen );

    // Calcula los índices k_PMax y k_INicioGen y las variables asociadas.
    procedure Calc_Auxiliares;

    // Calcula los parámetros de la curva_s representativa de la generación
    // entre 0 y PMax.
    procedure Calc_curva_s;

  private
    // lo deven llamar los constructores para calcular le vector pv3
    procedure Prepararse;
  end;


implementation


procedure TCurvaAerogen.Free;
begin
  curva_P.Free;
  curva_v.Free;
  curva_pv3.Free;
  inherited Free;
end;

procedure TCurvaAerogen.Prepararse;
var
  i: integer;
begin
  AreaRotor := pi * sqr(Diametro / 2);
  NPuntos:= self.curva_v.n;
  curva_pv3 := TVectR.Create_Init(NPuntos);
  //bucle
  for i := 1 to NPuntos do
    curva_pv3.pon_e(i, pv3Ofv(curva_v.e(i), ro_BASE));
  curva_pv3.MinMax(i, i, pv3min, pv3max);
  calc_auxiliares;
end;

constructor TCurvaAerogen.CreateLoadFromArchi(modelo, archi_curva: string);
var
  f: textfile;
  i: integer;
  p, v: NReal;

begin
  inherited Create;
  setSeparadoresGlobales;

  nid := -1; // indicamos que todavía no tiene nid

  self.modelo := modelo;

  assignfile(f, archi_curva);
  reset(f);
  readln(f, Diametro);
  readln(f, NPuntos);

  curva_P := TVectR.Create_Init(NPuntos);
  curva_v := TVectR.Create_Init(NPuntos);
  //bucle
  for i := 1 to NPuntos do
  begin
    readln(f, v, p);
    curva_v.pon_e(i, v);
    curva_P.pon_e(i, p / 1000.0);
  end;
  closefile(f);
  setSeparadoresLocales;
  Prepararse;
end;



// | 1| Vestas V112| 112| [ 26| 1; 2; 3; 4; 5; 6; 7; 8; 9; 10; 11; 12; 13; 14; 15; 16; 17; 18; 19; 20; 21; 22; 23; 24; 25; 26]| [ 26| 0; 0; 0.055598456; 0.12355212; 0.29034749; 0.52509653; 0.88957529; 1.3158301; 1.8965251; 2.5266409; 2.9220077; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 3.05; 0]

constructor TCurvaAerogen.CreateFromStr( strDef: string );
var
  pal: string;
  k: integer;
begin
  inherited Create;
  setSeparadoresGlobales;

  getPalHastaSep( pal, strDef, '|' );

  getPalHastaSep( pal, strDef, '|' );
  nid:= StrToInt( pal );

  getPalHastaSep( pal, strDef, '|' );
  modelo:= pal;

  getPalHastaSep( pal, strDef, '|' );
  Diametro:= StrToFloat( pal );

  getPalHastaSep( pal, strDef, '[' );
  getPalHastaSep( pal, strDef, '|' );
  NPuntos:= StrToInt( pal );

  curva_P := TVectR.Create_Init(NPuntos);
  curva_v := TVectR.Create_Init(NPuntos);

  for k:= 1 to NPuntos-1 do
    curva_v.pon_e( k, NextFloat( strDef ) );
  getPalHastaSep( pal, strDef, ']' );
  curva_v.pon_e( NPuntos, StrToFloat( pal ) );

  getPalHastaSep( pal, strDef, '|' );
  getPalHastaSep( pal, strDef, '[' );
  getPalHastaSep( pal, strDef, '|' );

  for k:= 1 to NPuntos-1 do
    curva_p.pon_e( k, NextFloat( strDef ) );
  getPalHastaSep( pal, strDef, ']' );
  curva_p.pon_e( NPuntos, StrToFloat( pal ) );

  setSeparadoresLocales;

  self.curva_P:= curva_P;
  self.curva_v:= curva_v;
  prepararse;
end;


constructor TCurvaAerogen.CreateLoadFromFile(var f: textfile);
var
  i: integer;
begin
  inherited Create;
  setSeparadoresGlobales;
  readln(f, nid );
  readln(f, modelo);
  readln(f, diametro);
  curva_v := TVectR.CreateLoadFromFile(f);
  curva_P := TVectR.CreateLoadFromFile(f);
  setSeparadoresLocales;
  Prepararse;
end;

constructor TCurvaAerogen.Create_Clone( aCurvaAerogen: TCurvaAeroGen );
begin
  inherited Create;
  nid:= aCurvaAerogen.nid;
  modelo:= aCurvaAerogen.modelo;
  diametro:= aCurvaAerogen.Diametro;
  curva_v:= TVectR.Create_Clone( aCurvaAerogen.curva_v );
  curva_P:= TVectR.Create_Clone( aCurvaAerogen.curva_P );
  Prepararse;
end;

{$IFNDEF SIN_DBCON}
constructor TCurvaAerogen.CreateLoadFromDB(r: TSQLQuery);
begin
  inherited Create;
  nid := r.FieldByName('nid').AsInteger;
  modelo := r.FieldByName('modelo').AsString;
  diametro := r.FieldByName('diametro').AsFloat;
  curva_v := TVectR.Create_unserialize(r.FieldByName('curva_v').AsString);
  curva_P := TVectR.Create_unserialize(r.FieldByName('curva_P').AsString);
  prepararse;
end;

constructor TCurvaAerogen.CreateLoadFromDB_nid(db: TDBPQCon; nidcurva: integer);
var
  r: TSQLQuery;
begin
  r := DB.query('SELECT * FROM curvas_aerogen WHERE nid =' + IntToStr(nidcurva));
  CreateLoadFromDB(r);
  r.Free;
end;

function TCurvaAerogen.InsertIntoDB(db: TDBPQCon): integer;
var
  nid: integer;
  s: string;
begin
  nid := DB.nextval('curvas_aerogen');
  s := 'INSERT INTO curvas_aerogen ( nid, modelo, diametro, curva_v, curva_P ) VALUES (';
  s := s + IntToStr(nid) + ', ''' + modelo + ''', ''' + FloatToStr(
    diametro) + ''', ''' + curva_v.serialize + ''', ''' + curva_P.serialize + ''' );';
  DB.exec(s);
  Result := nid;
end;

procedure TCurvaAerogen.UpdateDB(db: TDBPQCon);
var
  s: string;
begin
  s := 'UPDATE curvas_aerogen SET modelo= ''';
  s := s + modelo + ''', diametro= ''' + FloatToStr(diametro) +
    ''', curva_v=''' + curva_v.serialize + ''', curva_P=''' +
    curva_P.serialize + ''' WHERE nid =' + IntToStr(nid) + ';';
  DB.exec(s);
end;

class procedure TCurvaAerogen.CrearTablaDB(db: TDBPQCon);
begin
  DB.exec('DROP TABLE IF EXISTS curvas_aerogen');
  DB.exec('CREATE TABLE curvas_aerogen ( nid serial PRIMARY KEY, ' +
    'modelo VARCHAR(100), diametro double precision, curva_v text, curva_P text ); ');
  DB.exec('CREATE UNIQUE INDEX modelo ON curvas_aerogen ( modelo ); ');
end;
{$ENDIF}

procedure TCurvaAerogen.StoreInFile(var f: textfile);
begin
  setSeparadoresGlobales;
  writeln(f, nid);
  writeln(f, modelo);
  writeln(f, diametro);
  curva_v.StoreInFile(f);
  curva_P.StoreInFile(f);
  setSeparadoresLocales;
end;


function TCurvaAerogen.P_MW(pv3: NReal): NReal;
var
  ind: NReal;
begin
  if (pv3 < pv3min) or (pv3 > pv3max) then
  begin
    Result := 0;
    exit;
  end;
  ind := curva_pv3.inv_interpol(pv3);
  Result := curva_P.interpol(ind);
end;

function TCurvaAerogen.pv3_of_P_MW( P_MW: NReal): NReal;
var
  ind, pv3: NReal;
begin
  if  P_MW < 0 then
  begin
    result:= 0;
    exit;
  end;

  if P_MW > PMax then
  begin
    result:= pv3max;
    exit;
  end;
  ind := curva_P.inv_interpol( P_MW );
   pv3 := curva_pv3.interpol(ind);
   result:= pv3;
end;




function TCurvaAerogen.P_MW_Of_v(v: NReal): NReal;
var
  pv3: NReal;
begin
  pv3:= pv3Ofv( v, ro_base );
  result:= P_MW( pv3 );
end;

function TCurvaAerogen.v_of_P_MW( P_MW: NReal): NReal;
var
  pv3: NReal;
begin
  pv3:= pv3_of_P_MW( P_MW );
  result:= vofpv3( pv3, ro_base );
end;

// sirve para crear la curva de un parque por suma de curvas
procedure TCurvaAerogen.SumarCurvas( Curvas: array of TCurvaAerogen );
var
  k, j: integer;
  aCurva: TCurvaAerogen;
begin
  for j:= 0 to high( Curvas ) do
  begin
    aCurva:= Curvas[j];
    for k:= 1 to curva_pv3.n do
      curva_P.acum_e( k, aCurva.P_MW( curva_pv3.e( k )) );
  end;
  Calc_Auxiliares;
end;

procedure TCurvaAerogen.Calc_Auxiliares;
var
  ind: integer;
  PUmbral: NReal;

begin
  PMax:= curva_P.maxVal;
  ind:= 1;
  while abs(curva_P.e( ind )- PMax)> 1e-10 do
   inc( ind );
  k_Pmax:= ind;
  v_Pmax:= curva_v.e( ind );
  pv3_Pmax:= curva_pv3.e( ind );


  PUmbral:= PMax * 0.01;
  ind:= 1;
  while curva_P.e( ind )< PUmbral do
   inc( ind );
  k_InicioGen:= ind;
  v_InicioGen:= curva_v.e( ind );
  pv3_InicioGen:= curva_pv3.e( ind );

  Calc_curva_s;

end;

procedure TCurvaAerogen.Calc_curva_s;
var
  y: TVectR;
  x: NReal;
  v01, v09: NReal;
  a, b: NReal;
  k: integer;
begin
  y:= TVectR.Create_INit( curva_P.n );
  v01:= v_InicioGen + 1.1;
  v09:= v_Pmax * 0.9;
  for k:= k_InicioGen to k_Pmax do
  begin
    x:= curva_v.e( k );
    if ( v01 <= x ) and ( x < v09 ) then
      y.pon_e( k, ln( PMax/ curva_P.e( k ) - 1 ) )
    else
      y.pon_e( k, 0 );
  end;
  curva_s_errc2:= y.AproximacionLinealFiltrada( a, b, curva_v, v01, v09, curva_s_cntMuestras);
  curva_s_alfa:= -a;
  curva_s_vm:= b/curva_s_alfa;
  y.Free;
end;

end.

